
#include "System.h"
#include "CairoUtil.h"


/****************************************************************************/


CairoOutputSurface::CairoOutputSurface(int image_size_x_, int image_size_y_, 
                                       bool keep_aspect_)
  : keep_aspect(keep_aspect_),
    image_size_x(image_size_x_), image_size_y(image_size_y_)
{
  cairo = NULL;
}

cairo_t *CairoOutputSurface::Init(double view_size_x, double view_size_y)
{
  scale_x = double(image_size_x) / view_size_x;
  scale_y = double(image_size_y) / view_size_y;
  if (keep_aspect)
  {
    scale_x = scale_y = min(scale_x, scale_y);
    image_size_x = int(view_size_x * scale_x + 0.5);
    image_size_y = int(view_size_y * scale_y + 0.5);
  }
  
  surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, image_size_x, image_size_y);
  Assert(cairo_surface_status(surface) == CAIRO_STATUS_SUCCESS);
  
  cairo = cairo_create(surface);
  Assert(cairo_status(cairo) == CAIRO_STATUS_SUCCESS);
  
  cairo_set_antialias(cairo, CAIRO_ANTIALIAS_GRAY);
  cairo_scale(cairo, scale_x, scale_y);
  cairo_save(cairo);
  
  return cairo;
}

void CairoOutputSurface::Finalize(cairo_t *cairo)
{
  cairo_restore(cairo);
  cairo_destroy(cairo);
  cairo_surface_destroy(surface);
}


/****************************************************************************/


CairoOutputPNG::CairoOutputPNG(const string &filename_, 
                               int image_size_x_, int image_size_y_, 
                               bool keep_aspect_)
  : filename(filename_), keep_aspect(keep_aspect_),
    image_size_x(image_size_x_), image_size_y(image_size_y_)
{
  cairo = NULL;
}

cairo_t *CairoOutputPNG::Init(double view_size_x, double view_size_y)
{
  scale_x = double(image_size_x) / view_size_x;
  scale_y = double(image_size_y) / view_size_y;
  if (keep_aspect)
  {
    scale_x = scale_y = min(scale_x, scale_y);
    image_size_x = int(view_size_x * scale_x + 0.5);
    image_size_y = int(view_size_y * scale_y + 0.5);
  }
  
  surface = cairo_image_surface_create(CAIRO_FORMAT_ARGB32, image_size_x, image_size_y);
  Assert(cairo_surface_status(surface) == CAIRO_STATUS_SUCCESS);
  
  cairo = cairo_create(surface);
  Assert(cairo_status(cairo) == CAIRO_STATUS_SUCCESS);
  
  cairo_set_antialias(cairo, CAIRO_ANTIALIAS_GRAY);
  cairo_scale(cairo, scale_x, scale_y);
  cairo_save(cairo);
  
  return cairo;
}

void CairoOutputPNG::Finalize(cairo_t *cairo)
{
  cairo_restore(cairo);
  Assert(cairo_surface_write_to_png(surface, filename.c_str()) == CAIRO_STATUS_SUCCESS);
  
  cairo_destroy(cairo);
  cairo_surface_destroy(surface);
}


/****************************************************************************/


CairoOutputPS::CairoOutputPS(const string &filename_, 
                             double image_size_x_, double image_size_y_, 
                             bool keep_aspect_)
  : filename(filename_), keep_aspect(keep_aspect_),
    image_size_x(image_size_x_), image_size_y(image_size_y_)
{
  cairo = NULL;
}

cairo_t *CairoOutputPS::Init(double view_size_x, double view_size_y)
{
  scale_x = image_size_x / view_size_x;
  scale_y = image_size_y / view_size_y;
  if (keep_aspect)
  {
    scale_x = scale_y = min(scale_x, scale_y);
    image_size_x = view_size_x * scale_x;
    image_size_y = view_size_y * scale_y;
  }

  surface = cairo_ps_surface_create(filename.c_str(), 
                                    image_size_x, image_size_y);
  Assert(cairo_surface_status(surface) == CAIRO_STATUS_SUCCESS);

  cairo = cairo_create(surface);
  Assert(cairo_status(cairo) == CAIRO_STATUS_SUCCESS);

  cairo_set_antialias(cairo, CAIRO_ANTIALIAS_GRAY);
  cairo_surface_set_fallback_resolution(surface, 600, 600);

  cairo_scale(cairo, scale_x, scale_y);
  cairo_save(cairo);
  
  return cairo;
}

void CairoOutputPS::Finalize(cairo_t *cairo)
{
  cairo_restore(cairo);
  cairo_show_page(cairo);
  cairo_destroy(cairo);
  cairo_surface_destroy(surface);
}


/****************************************************************************/


CairoOutputPDF::CairoOutputPDF(const string &filename_, 
                               double image_size_x_, double image_size_y_, 
                               bool keep_aspect_)
  : filename(filename_), keep_aspect(keep_aspect_),
    image_size_x(image_size_x_), image_size_y(image_size_y_)
{
  cairo = NULL;
}

cairo_t *CairoOutputPDF::Init(double view_size_x, double view_size_y)
{
  scale_x = image_size_x / view_size_x;
  scale_y = image_size_y / view_size_y;
  if (keep_aspect)
  {
    scale_x = scale_y = min(scale_x, scale_y);
    image_size_x = view_size_x * scale_x;
    image_size_y = view_size_y * scale_y;
  }

  surface = cairo_ps_surface_create(filename.c_str(), 
                                    image_size_x, image_size_y);
  Assert(cairo_surface_status(surface) == CAIRO_STATUS_SUCCESS);

  cairo = cairo_create(surface);
  Assert(cairo_status(cairo) == CAIRO_STATUS_SUCCESS);

  cairo_set_antialias(cairo, CAIRO_ANTIALIAS_GRAY);
  cairo_surface_set_fallback_resolution(surface, 600, 600);

  cairo_scale(cairo, scale_x, scale_y);
  cairo_save(cairo);
  
  return cairo;
}

void CairoOutputPDF::Finalize(cairo_t *cairo)
{
  cairo_restore(cairo);
  cairo_show_page(cairo);
  cairo_destroy(cairo);
  cairo_surface_destroy(surface);
}


/****************************************************************************/


Cairo::Cairo(double view_size_x_, double view_size_y_, 
             const CairoOutputType &output_)
{
  Init(view_size_x_, view_size_y_, output_);
}

Cairo::~Cairo()
{
  Done();
}


void Cairo::Init(double view_size_x_, double view_size_y_,
                 const CairoOutputType &output_)
{
  view_size_x = view_size_x_;
  view_size_y = view_size_y_;
  output = output_.clone();
  
  cairo = output->Init(view_size_x, view_size_y);
}

void Cairo::Done()
{
  if (cairo)
  {
    output->Finalize(cairo);
    delete output;
    cairo = NULL;
  }
}


void Cairo::RenderText(const string &font_face, 
                       double font_size, int x_align, int y_align, 
                       const string &text)
{
  if (text.empty())
    return;
  
  cairo_select_font_face(cairo, font_face.c_str(), CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
  cairo_set_font_size(cairo, font_size);
  cairo_font_extents_t font_extents;
  cairo_font_extents(cairo, &font_extents);
  
  double font_height = font_extents.ascent - font_extents.descent;
  
  vector<RenderTextParams> params;
  vector<string> texts;
  ParseRenderText(text, params, texts, font_size, font_height);
  
  double text_left, text_right, text_x_offset = 0.0;
  for (int i = 0; i < params.size(); ++i)
  {
    cairo_select_font_face(cairo, font_face.c_str(), 
                           params[i].slant, params[i].weight);
    cairo_set_font_size(cairo, params[i].size);
    
    cairo_text_extents_t extents;
    cairo_text_extents(cairo, texts[i].c_str(), &extents);
    
    if (i == 0)
      text_left = extents.x_bearing;
    if (i == (params.size() - 1))
      text_right = text_x_offset + extents.x_bearing + extents.width;
    
    text_x_offset += extents.x_advance;
  }
  
  if (x_align > 0)
    cairo_rel_move_to(cairo, -text_left, 0.0);
  else if (x_align < 0)
    cairo_rel_move_to(cairo, -text_right, 0.0);
  else
    cairo_rel_move_to(cairo, -0.5 * (text_left + text_right), 0.0);
  
  if (y_align > 0)
    cairo_rel_move_to(cairo, 0.0, font_height);
  else if (y_align < 0)
    cairo_rel_move_to(cairo, 0.0, 0.0);
  else
    cairo_rel_move_to(cairo, 0.0, 0.5 * font_height);
  
  for (int i = 0; i < params.size(); ++i)
  {
    cairo_select_font_face(cairo, font_face.c_str(),
                           params[i].slant, params[i].weight);
    cairo_set_font_size(cairo, params[i].size);
    
    cairo_rel_move_to(cairo, 0.0, params[i].y_offset);
    cairo_show_text(cairo, texts[i].c_str());
    cairo_rel_move_to(cairo, 0.0, -params[i].y_offset);
  }
}


double Cairo::RenderTextWidth(const string &font_face, 
                              double font_size, const string &text)
{
  cairo_select_font_face(cairo, font_face.c_str(), CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
  cairo_set_font_size(cairo, font_size);
  cairo_font_extents_t font_extents;
  cairo_font_extents(cairo, &font_extents);
  
  double font_height = font_extents.ascent - font_extents.descent;
  
  vector<RenderTextParams> params;
  vector<string> texts;
  ParseRenderText(text, params, texts, font_size, font_height);
  if (params.empty())
    return 0.0;
  
  double text_left, text_right, text_x_offset = 0.0;
  for (int i = 0; i < params.size(); ++i)
  {
    cairo_select_font_face(cairo, font_face.c_str(), 
                           params[i].slant, params[i].weight);
    cairo_set_font_size(cairo, params[i].size);
    
    cairo_text_extents_t extents;
    cairo_text_extents(cairo, texts[i].c_str(), &extents);
    
    if (i == 0)
      text_left = extents.x_bearing;
    if (i == (params.size() - 1))
      text_right = text_x_offset + extents.x_bearing + extents.width;
    
    text_x_offset += extents.x_advance;
  }
  
  return (text_right - text_left);
}


double Cairo::RenderTextHeight(const string &font_face, 
                               double font_size)
{
  cairo_select_font_face(cairo, font_face.c_str(), CAIRO_FONT_SLANT_NORMAL, CAIRO_FONT_WEIGHT_NORMAL);
  cairo_set_font_size(cairo, font_size);
  cairo_font_extents_t font_extents;
  cairo_font_extents(cairo, &font_extents);
  
  double font_height = font_extents.ascent - font_extents.descent;
  return font_height;
}


void Cairo::RenderTextAddChar(vector<RenderTextParams> &params, 
                              vector<string> &texts,
                              double size, double y_offset, 
                              cairo_font_slant_t slant,
                              cairo_font_weight_t weight, char c)
{
  if (params.empty() || (size != params.back().size) ||
      (y_offset != params.back().y_offset) ||
      (slant != params.back().slant) || (weight != params.back().weight))
  {
    params.push_back(RenderTextParams(size, y_offset, slant, weight));
    texts.push_back("");
  }
  texts.back() += c;
}

void Cairo::ParseRenderText(const string &text, 
                            vector<RenderTextParams> &params, 
                            vector<string> &texts, 
                            double font_size, double font_height)
{
  double cur_size = font_size;
  double cur_y_offset = 0.0;
  cairo_font_slant_t cur_slant = CAIRO_FONT_SLANT_NORMAL;
  cairo_font_weight_t cur_weight = CAIRO_FONT_WEIGHT_NORMAL;
  
  for (int i = 0; i < text.length(); ++i)
  {
    if (text[i] != '~')
      RenderTextAddChar(params, texts, cur_size, cur_y_offset, 
                        cur_slant, cur_weight, text[i]);
    else
    {
      Assert(++i < text.length());
      if (text[i] == '~')
      {
        RenderTextAddChar(params, texts, cur_size, cur_y_offset, 
                          cur_slant, cur_weight, '~');
        continue;
      }
      else if (text[i] == '0')
      {
        cur_size = font_size;
        cur_y_offset = 0.0;
        cur_slant = CAIRO_FONT_SLANT_NORMAL;
        cur_weight = CAIRO_FONT_WEIGHT_NORMAL;
      }
      else if (toupper(text[i]) == 'B')
      {
        Assert(++i < text.length());
        switch (text[i])
        {
        case '0':
          cur_weight = CAIRO_FONT_WEIGHT_NORMAL;
          break;
        case '1':
          cur_weight = CAIRO_FONT_WEIGHT_BOLD;
          break;
        default:
          Assert(0);
        }
      }
      else if (toupper(text[i]) == 'I')
      {
        Assert(++i < text.length());
        switch (text[i])
        {
        case '0':
          cur_slant = CAIRO_FONT_SLANT_NORMAL;
          break;
        case '1':
          cur_slant = CAIRO_FONT_SLANT_ITALIC;
          break;
        case '2':
          cur_slant = CAIRO_FONT_SLANT_OBLIQUE;
          break;
        default:
          Assert(0);
        }
      }
      else if (toupper(text[i]) == 'S')
      {
        Assert(++i < text.length());
        switch (text[i])
        {
        case '0':
          cur_size = font_size;
          cur_y_offset = 0.0;
          break;
        case '1':
          cur_size = 0.6 * font_size;
          cur_y_offset = -0.7 * font_height; 
          break;
        case '2':
          cur_size = 0.6 * font_size;
          cur_y_offset = 0.3 * font_height;
          break;
        default:
          Assert(0);
        }
      }
    }
  }
  Assert(params.size() == texts.size());
}
