
#ifndef CAIROUTIL_H
#define CAIROUTIL_H

#include "System.h"
#ifdef CAIRO
#include <cairo.h>
#include <cairo-ps.h>
#include <cairo-pdf.h>
#else
#error Must compile with cairo!
#endif


inline double color_r(uint32 color) { return double((color >> 16) & 0xff) / 255.0; }
inline double color_g(uint32 color) { return double((color >> 8) & 0xff) / 255.0; }
inline double color_b(uint32 color) { return double(color & 0xff) / 255.0; }

#define color_rgb(color) color_r(color), color_g(color), color_b(color)

inline double clamp(double x, double lower, double upper) { return ((x < lower) ? lower : (x > upper) ? upper : x); }

inline uint32 make_rgba(double r, double g, double b, double a)
{
  a = clamp(a, 0.0, 1.0);
  return (uint32(255.0 * a) << 24) |
    (uint32(255.0 * clamp(r * a, 0.0, 1.0)) << 16) |
    (uint32(255.0 * clamp(g * a, 0.0, 1.0)) << 8) |
    (uint32(255.0 * clamp(b * a, 0.0, 1.0)));
}

inline double interpolate(double x0, double x1, double s) {return x0 + s * (x1 - x0);}

#define color_interpolate_rgb(color0, color1, s) interpolate(color_r(color0), color_r(color1), s), interpolate(color_g(color0), color_g(color1), s), interpolate(color_b(color0), color_b(color1), s)


class CairoOutputType
{
public:
  virtual cairo_t *Init(double view_size_x, double view_size_y) = 0;
  virtual void Finalize(cairo_t *cairo) = 0;
  virtual CairoOutputType *clone() const = 0;
  virtual cairo_surface_t *GetSurface() = 0;
};


class CairoOutputSurface : public CairoOutputType
{
public:
  int image_size_x, image_size_y;
  bool keep_aspect;
  double scale_x, scale_y;
  cairo_surface_t *surface;
  cairo_t *cairo;

  CairoOutputSurface(int image_size_x_, int image_size_y_,
                     bool keep_aspect_ = false);

  virtual cairo_t *Init(double view_size_x, double view_size_y);
  virtual void Finalize(cairo_t *cairo);
  virtual CairoOutputSurface *clone() const { return new CairoOutputSurface(*this); }
  virtual cairo_surface_t *GetSurface() { return surface; }
};


class CairoOutputPNG : public CairoOutputType
{
public:
  string filename;
  int image_size_x, image_size_y;
  bool keep_aspect;
  double scale_x, scale_y;
  cairo_surface_t *surface;
  cairo_t *cairo;

  CairoOutputPNG(const string &filename_, 
                 int image_size_x_, int image_size_y_, 
                 bool keep_aspect_ = false);

  virtual cairo_t *Init(double view_size_x, double view_size_y);
  virtual void Finalize(cairo_t *cairo);
  virtual CairoOutputPNG *clone() const { return new CairoOutputPNG(*this); }
  virtual cairo_surface_t *GetSurface() { return surface; }
};


class CairoOutputPS : public CairoOutputType
{
public:
  string filename;
  double image_size_x, image_size_y;
  bool keep_aspect;
  double scale_x, scale_y;
  cairo_surface_t *surface;
  cairo_t *cairo;

  CairoOutputPS(const string &filename_,
                double image_size_x_, double image_size_y_,
                bool keep_aspect_ = false);

  virtual cairo_t *Init(double view_size_x, double view_size_y);
  virtual void Finalize(cairo_t *cairo);
  virtual CairoOutputPS *clone() const { return new CairoOutputPS(*this); }
  virtual cairo_surface_t *GetSurface() { return surface; }
};


class CairoOutputPDF : public CairoOutputType
{
public:
  string filename;
  double image_size_x, image_size_y;
  bool keep_aspect;
  double scale_x, scale_y;
  cairo_surface_t *surface;
  cairo_t *cairo;

  CairoOutputPDF(const string &filename_,
                 double image_size_x_, double image_size_y_,
                 bool keep_aspect_ = false);

  virtual cairo_t *Init(double view_size_x, double view_size_y);
  virtual void Finalize(cairo_t *cairo);
  virtual CairoOutputPDF *clone() const { return new CairoOutputPDF(*this); }
  virtual cairo_surface_t *GetSurface() { return surface; }
};


class Cairo
{ 
public:
  cairo_t *cairo;
  double view_size_x, view_size_y;
  CairoOutputType *output;

  Cairo() { }
  Cairo(double view_size_x_, double view_size_y_, 
        const CairoOutputType &output_);
  ~Cairo();

  void Init(double view_size_x_, double view_size_y_,
            const CairoOutputType &output_);
  void Done();

  operator cairo_t *() const { return cairo; }

  operator cairo_surface_t *() const { return output->GetSurface(); }

  
  void RenderText(const string &font_face, double font_size,
                  int x_align, int y_align, const string &text);
  /* ~~ = tilde
     ~b0 = weight normal
     ~b1 = weight bold
     ~i0 = slant normal
     ~i1 = slant italic
     ~i2 = slant oblique
     ~s0 = full script
     ~s1 = super script
     ~s2 = sub script
     ~0 = normal everything
  */
  
  double RenderTextWidth(const string &font_face, 
                         double font_size, const string &text);
  
  double RenderTextHeight(const string &font_face, 
                          double font_size);
  
private:
  struct RenderTextParams
  {
    double size;
    double y_offset;
    cairo_font_slant_t slant;
    cairo_font_weight_t weight;
    
    RenderTextParams() { }
    
    RenderTextParams(double size_, double y_offset_, 
                     cairo_font_slant_t slant_, cairo_font_weight_t weight_)
      : size(size_), y_offset(y_offset_), slant(slant_), weight(weight_)
    {
    }
  };
  
  void RenderTextAddChar(vector<RenderTextParams> &params, 
                         vector<string> &texts,
                         double size, double y_offset, 
                         cairo_font_slant_t slant,
                         cairo_font_weight_t weight, char c);
  void ParseRenderText(const string &text, 
                       vector<RenderTextParams> &params, 
                       vector<string> &texts, 
                       double font_size, double font_height);
};

#endif
