
#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"
#include "Simulate.h"


Params params(false);

int debug = 0;

string input_dir;
string input_file;
vector<string> simulate_input_files;
vector<int> selected;
string output_dir;
string output_file;
bool composite = false;
bool force_admixture = false;
bool do_gender = false;

int default_gen_samples = 100;
map<int, int> gen_samples;
int min_gen = 1;
int max_gen = 20;

vector<double> pop_prob;


string required_param[] =
{
  "input_file",
  "output_file"
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

void GetIntegerSpec(string s, vector<int> &result);

void InitParams()
{
  params.ExpandImports();

  uint32 seed = RandomSeed();
  if (params.Contains("rand_seed"))
    seed = params["rand_seed"].GetInt();
  srand(seed);
  cerr << "- Random seed = " << seed << endl;

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("input_file"))
  {
    input_file = params["input_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(input_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(input_file))) == "snpprofiledata");
  }

  if (params.Contains("simulate_input_files"))
  {
    vector<ParamValue> values = params["simulate_input_files"].GetList();
    for (int i = 0; i < values.size(); ++i)
    {
      string filename = values[i].GetString();
      Assert(GetExtension(RemoveCompressedExtension(filename)) == "bin");
      Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "diplopop");
      simulate_input_files.push_back(filename);
    }
  }

  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (params.Contains("output_file"))
  {
    output_file = params["output_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(output_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(output_file))) == "diplopop");
  }

  if (params.Contains("select"))
  {
    GetIntegerSpec(params["select"].GetString(), selected);
    sort(selected.begin(), selected.end());
    selected.resize(unique(selected.begin(), selected.end()) - selected.begin());
  }

  if (params.Contains("min_gen"))
  {
    min_gen = params["min_gen"].GetInt();
    Assert(min_gen >= 1);
  }
  if (params.Contains("max_gen"))
  {
    max_gen = params["max_gen"].GetInt();
    Assert(max_gen >= min_gen);
  }

  if (params.Contains("gen_samples"))
    default_gen_samples = params["gen_samples"].GetInt();

  for (int gen = min_gen; gen <= max_gen; ++gen)
    gen_samples[gen] = default_gen_samples;

  if (params.Contains("gen_samples_list"))
  {
    vector<ParamValue> values = params["gen_samples_list"].GetList();
    Assert(values.size() <= (max_gen - min_gen + 1));
    for (int i = 0; i < values.size(); ++i)
      gen_samples[min_gen + i] = values[i].GetInt();
  }
  
  if (params.Contains("composite"))
  {
    Assert(!simulate_input_files.empty());
    composite = true;
  }

  if (params.Contains("pop_prob"))
  {
    vector<ParamValue> values = params["pop_prob"].GetList();
    pop_prob.resize(values.size());
    for (int i = 0; i < values.size(); ++i)
      pop_prob[i] = values[i].GetDouble();
  }

  if (params.Contains("force_admixture"))
    force_admixture = true;

  if (params.Contains("do_gender"))
    do_gender = true;

  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;
  cerr << "- Input file = " << input_file << endl;
  if (!simulate_input_files.empty())
  {
    cerr << "- Input simulated files = {";
    for (int i = 0; i < simulate_input_files.size(); ++i)
      cerr << " " << simulate_input_files[i];
    cerr << " }" << endl;
  }
  if (!selected.empty())
    cerr << "- Selected sources = { " << IntegerListString(selected.begin(), selected.end()) << " }" << endl;
  if (!output_dir.empty())
    cerr << "- Output directory = " << output_dir << endl;
  cerr << "- Output file = " << output_file << endl;
  cerr << "- Doing generations [" << min_gen << ", " << max_gen << "]" << endl;
  cerr << "- # samples per generation = {";
  for (int gen = min_gen; gen <= max_gen; ++gen)
    cerr << " " << gen << ":" << gen_samples[gen];
  cerr << " }" << endl;
  if (composite)
    cerr << "- Composite simulations" << endl;
  if (!pop_prob.empty())
  {
    cerr << "- Population probabilities = {";
    for (int pop = 0; pop < pop_prob.size(); ++pop)
      cerr << " " << pop_prob[pop];
    cerr << " }" << endl;
  }
  else
    cerr << "- Equal population probabilities" << endl;
  if (force_admixture)
    cerr << "- Force admixture" << endl;
  if (do_gender)
    cerr << "- Simulate gender-correct mating" << endl;
}

void GetIntegerSpec(string s, vector<int> &result)
{
  int p = 0;
  while (p < s.length())
  {
    if (isdigit(s[p]))
    {
      int i;
      Assert(StringToInt(s, i, p, p));
      if ((p == s.length()) || (s[p] == ','))
      {
        result.push_back(i);
        if (p < s.length())
          ++p;
        continue;
      }

      Assert(s[p++] == '-');
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p == s.length()) || (s[p] == ','));
      if (i <= j)
      {
        for (int k = i; k <= j; ++k)
          result.push_back(k);
      }
      else
      {
        for (int k = i; k >= j; --k)
          result.push_back(k);
      }
      if (p < s.length());
      ++p;
    }
    else if (s[p] == '(')
    {
      int i;
      Assert(StringToInt(s, i, p, ++p));
      Assert((p < s.length()) && (s[p] == ','));
      ++p;
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p < s.length()) && (s[p] == ')'));
      Assert(j >= 0);
      for (int k = 0; k < j; ++k)
        result.push_back(i + k);
      ++p;
      Assert((p == s.length()) || (s[p] == ','));
      if (p < s.length())
        ++p;
    }
    else
      Assert(false);
  }
}

void InitSamplePop()
{
  double total = 0.0;
  for (int pop = 0; pop < pop_prob.size(); ++pop)
    total += pop_prob[pop];
  Assert(total > 0.0);
  
  double cumul = 0.0;
  for (int pop = 0; pop < pop_prob.size(); ++pop)
  {
    cumul += pop_prob[pop] / total;
    pop_prob[pop] = cumul;
  }

  Assert(fabs(pop_prob.back() - 1.0) < 1e-6);
  pop_prob.back() = 1.0;
}

int SamplePop()
{
  double p = drand();
  vector<double>::const_iterator find = lower_bound(pop_prob.begin(), pop_prob.end(), p);
  Assert(find != pop_prob.end());
  return (find - pop_prob.begin());
}


int Main(vector<string> args)
{
  cerr << "CreateAdmixture" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;
  
  SNPProfile profile;
  SNPData data;
  {
    istream *in = NULL;
    if (!input_dir.empty())
      in = InFileStream(RelPath(input_dir, input_file));
    if (!in)
      in = InFileStream(input_file);

    if (!in)
    {
      cerr << "Could not open file \"" << input_file << "\" in directory";
      if (!input_dir.empty())
        cerr << " \"" << input_dir << "\"";
      Exit(1);
    }

    cerr << "* Reading sample data...";
    ReadBinSNPProfileData(*in, profile, data);
    cerr << " done." << endl;
    delete in;

    profile.Validate();
    data.ComputeDerived();
    data.Validate();
  }
  
  vector<Simulate::DiploPop> sources;
  if (!simulate_input_files.empty())
  {
    sources.resize(simulate_input_files.size(), Simulate::DiploPop(data));
    for (int pop = 0; pop < simulate_input_files.size(); ++pop)
    {
      istream *in = InFileStream(RelPath(input_dir, simulate_input_files[pop]));
      if (!in)
      {
        cerr << "Could not open file \"" << simulate_input_files[pop] << "\"";
        if (!input_dir.empty())
          cerr << " in directory \"" << input_dir << "\"";
        cerr << endl;
        Exit(1);
      }

      cerr << "* Reading simulated population \"" << simulate_input_files[pop] << "\"...";
      sources[pop].ReadBin(*in);
      delete in;
      cerr << " done. # diploid individuals = " << sources[pop].diplos.size() << endl;

      if (!selected.empty())
      {
        Simulate::DiploPop temp(data);
        for (int i = 0; i < selected.size(); ++i)
        {
          int k = selected[i];
          Assert((k >= 0) && (k < sources[pop].diplos.size()));
          temp.diplos.push_back(sources[pop].diplos[k]);
        }
        sources[pop].swap(temp);
      }
    }
  }
  else
  {
    sources.resize(data.num_pops, Simulate::DiploPop(data));
    for (int pop = 0; pop < data.num_pops; ++pop)
    {
      if (!selected.empty())
      {
        for (int i = 0; i < selected.size(); ++i)
        {
          int k = selected[i];
          Assert((k >= 0) && (k < data.pops[pop].diplos.size()));
          sources[pop].diplos.push_back(Simulate::Diplo(data, data.pops[pop].diplos[k]));
        }
      }
      else
        for (int i = 0; i < data.pops[pop].diplos.size(); ++i)
          sources[pop].diplos.push_back(Simulate::Diplo(data, data.pops[pop].diplos[i]));
    }
  }

  if (!pop_prob.empty())
    pop_prob.resize(sources.size(), 0.0);
  else
    pop_prob.resize(sources.size(), 1.0);
  InitSamplePop();

  Simulate::DiploPop admixed;
  admixed.SetSNPData(data);
  {
    int num_samples = 0;
    for (int gen = min_gen; gen <= max_gen; ++gen)
      num_samples += gen_samples[gen];
    admixed.diplos.reserve(num_samples);
  }
  set<int> haplo_count;
  for (int gen = min_gen; gen <= max_gen; ++gen)
  {
    cerr << "* Generation " << gen << ": ";
    int num_samples = gen_samples[gen];

    set<int> leave_diplos;
    map<int, int> leaves;
    for (int sample = 0; sample < num_samples; ++sample)
    {
      Simulate::Diplo tree;
      tree.SetFullTree(*data.profile, gen, -1);
      
      set<int> pops_sampled;
      do
      {
        pops_sampled.clear();

        leave_diplos.clear();
        for (int a = 0; a < 2; ++a)
          for (int i = 0; i < tree.haplos[a].sources.size(); ++i)
            leave_diplos.insert(tree.haplos[a].sources[i].second >> 1);
        
        leaves.clear();
        for (set<int>::const_iterator iter = leave_diplos.begin(); 
             iter != leave_diplos.end(); ++iter)
        {
          int gender = (*iter & 1);
          int pop = SamplePop();
          int d;
          do
          {
            d = (irand() % sources[pop].diplos.size());
          } while (do_gender && (sources[pop].diplos[d].gender != gender));
          leaves[(*iter)*2 + 0] = (pop << 16) + d*2 + 0;
          leaves[(*iter)*2 + 1] = (pop << 16) + d*2 + 1;
          pops_sampled.insert(pop);
        }
      } while (force_admixture && (pops_sampled.size() < 2));
        
      tree.MapSources(leaves);
      for (int a = 0; a < 2; ++a)
        for (int i = 0; i < tree.haplos[a].sources.size(); ++i)
          haplo_count.insert(tree.haplos[a].sources[i].second);
      
      Simulate::Diplo result;
      if (composite || simulate_input_files.empty())
        result.Composite(sources, tree);
      else
        result = tree;

      admixed.diplos.push_back(result);
      cerr << ".";
      cerr << "(" << pops_sampled.size() << ")";
    }
    cerr << endl;
  }

  if (do_gender)
  {
    vector<int> gender_count(2, 0);
    //set<int> haplo_count;
    for (int d = 0; d < admixed.diplos.size(); ++d)
    {
      Assert((admixed.diplos[d].gender == 0) || (admixed.diplos[d].gender == 1));
      ++gender_count[admixed.diplos[d].gender];
    }
    
    cerr << "# female = " << gender_count[0] << endl;
    cerr << "# male = " << gender_count[1] << endl;
  }

  cerr << "haplos = { " << IntegerListString(haplo_count.begin(), haplo_count.end()) << " }" << endl;

  {
    ostream *out = OutFileStream(RelPath(output_dir, output_file));
    AssertMsg(out, output_file);
    cerr << "* Writing output...";
    Write(*out, admixed);
    delete out;
    cerr << " done." << endl;
  }

  cerr << "Done!" << endl;
  return 0;
}
