

#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"
#include "Segment.h"
#include "IntervalMap.h"


Params params(false);

int debug = 0;

string input_dir, output_dir;


/* SNPs:
 *   X    par1: [90118, 2688701]
 *   X non-par: [2696752, 154493116]
 *   X    par2: [154532742, 154798782]
 *   Y non-par: [2700537, 20324263]
 *
 * Sequence:
 *   X    par1: [1, 2692881]
 *   X non-par: [2692882, 154494746]
 *   X    par2: [154494747, 154824264]
 *   Y    par1: [1, 2692881]
 *   Y non-par: [2692882, 57372173]
 *   Y    par2: [57372174, 57701691]
 *
 * SRY: [2715030, 2715644] = 2715337
 * Mito: [1, 16571] = 8286
 */


string required_param[] =
{
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

void InitParams()
{
  params.ExpandImports();

  srand(TimerSeconds());
  if (params.Contains("rand_seed"))
    srand(params["rand_seed"].GetInt());

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;

  if (!output_dir.empty())
    cerr << "- Output directory = " << output_dir << endl;
}


string PopStr(int pop)
{
  Assert((pop >= 0) && (pop <= 3));
  if (pop == 0)
    return "CEU";
  else if (pop == 1)
    return "YRI";
  else if (pop == 2)
    return "CHB";
  else
    return "JPT";
}

string CPopStr(int pop)
{
  Assert((pop >= 0) && (pop <= 2));
  if (pop == 0)
    return "CEU";
  else if (pop == 1)
    return "YRI";
  else
    return "JPT+CHB";
}

string ChromStr(int chrom)
{
  Assert((chrom >= 0) && (chrom <= 24));
  if (chrom == 0)
    return "M";
  else if ((chrom >= 1) && (chrom <= 22))
    return ToStr(chrom);
  else if (chrom == 23)
    return "X";
  else
    return "Y";
}

string GenotypesFile(string pop, int chrom)
{
  string filename;
  if (chrom != 0)
    filename = "genotypes_chr" + ChromStr(chrom) + "_" + pop + "_r21a_nr.txt.gz";
  else
    filename = "genotypes_chrM_" + pop + "_nr.txt.gz";
  return RelPath(input_dir, filename);
}

string LegendFile(int cpop, string chrom_str, string chrom_str2)
{
  string filename = "genotypes_chr" + chrom_str + "_" + CPopStr(cpop) + "_r21_nr_fwd_" + chrom_str2 + (chrom_str2.empty() ? "" : "_") + "legend_all.gz";
  return RelPath(input_dir, filename);
}

string PhasedFile(int cpop, string chrom_str, string chrom_str2)
{
  string filename = "genotypes_chr" + chrom_str + "_" + CPopStr(cpop) + "_r21_nr_fwd_" + chrom_str2 + (chrom_str2.empty() ? "" : "_") + "phased_all.gz";
  return RelPath(input_dir, filename);
}

string SampleFile(int cpop)
{
  string filename = "genotypes_chr1_" + CPopStr(cpop) + "_r21_nr_fwd_sample.txt.gz";
  return RelPath(input_dir, filename);
}


string GeneticMapFile(string chrom)
{
  string filename = "genetic_map_chr" + chrom + ".txt.gz";
  return RelPath(input_dir, filename);
}


string Restrict550File(int chrom)
{
  string filename = "Hap550v3_561466_SNPs_chr" + ChromStr(chrom) + ".txt";
  return RelPath(input_dir, filename);
}

string Restrict650File(int chrom)
{
  string filename = "Hap650Yv3_660918_SNPs_chr" + ChromStr(chrom) + ".txt";
  return RelPath(input_dir, filename);
}


bool SkipWhitespace(const string &s, int &pos)
{
  bool skipped = false;
  while (pos < s.length())
    if (isspace(s[pos]))
    {
      ++pos;
      skipped = true;
    }
    else
      break;
  return skipped;
}

bool SkipNonWhitespace(const string &s, int &pos)
{
  bool skipped = false;
  while (pos < s.length())
    if (!isspace(s[pos]))
    {
      ++pos;
      skipped = true;
    }
    else
      break;
  return skipped;
}

struct SNPInfo
{
  int64 pos;
  string label;
  int8 strand;

  SNPInfo() { }
  SNPInfo(int64 pos_, const string &label_, int8 strand_)
    : pos(pos_), label(label_), strand(strand_) { }
};

bool operator<(const SNPInfo &x, const SNPInfo &y)
{
  return x.pos < y.pos;
}

bool operator==(const SNPInfo &x, const SNPInfo &y)
{
  return x.pos == y.pos;
}

void ScanAllSNPLoci(vector<int64> &snp_loci, vector<string> &snp_labels,
                    vector<int8> &snp_strands)
{
  vector<SNPInfo> temp;
  temp.reserve(20000000);
  cerr << "Scanning files for SNP loci:" << endl;
  for (int pop = 0; pop < 4; ++pop)
  //for (int pop = 0; pop < 1; ++pop)
  {
    cerr << "- " << PopStr(pop);
    for (int chrom = 0; chrom <= 24; ++chrom)
    {
      string filename = GenotypesFile(PopStr(pop), chrom);
      istream *in = InFileStream(filename);
      AssertMsg(in, filename);

      string line;
      Assert(getline(*in, line));

      vector<string> fields;
      while (getline(*in, line))
      {
        int p = 0;
        SkipWhitespace(line, p);
        if (p == line.length())
          continue;

        int q = p;
        Assert(SkipNonWhitespace(line, p));
        string label = line.substr(q, p - q);

        Assert(SkipWhitespace(line, p));
        Assert(SkipNonWhitespace(line, p));
        Assert(SkipWhitespace(line, p));
        Assert(SkipNonWhitespace(line, p));
        Assert(SkipWhitespace(line, p));
        
        int pos;
        q = p;
        Assert(SkipNonWhitespace(line, p));
        Assert(StringToInt(line.substr(q, p - q), pos));

        Assert(SkipWhitespace(line, p));
        q = p;
        Assert(SkipNonWhitespace(line, p));
        Assert(p == (q + 1));
        Assert((line[q] == '+') || (line[q] == '-'));
        int8 strand = (line[q] == '-') ? 1 : 0;

        Assert((p == line.length()) || isspace(line[p]));
        if (chrom == 0)
          label = "mito" + ToStr(pos);
        temp.push_back(SNPInfo((int64(chrom) << 32) + int64(pos), label, strand));
      }

      delete in;
      cerr << ".";
    }
    cerr << endl;
  }
 
  cerr << "- Finalizing...";
  sort(temp.begin(), temp.end());
  temp.resize(unique(temp.begin(), temp.end()) - temp.begin());

  snp_loci.reserve(temp.size());
  snp_labels.reserve(temp.size());
  snp_strands.reserve(temp.size());

  for (int i = 0; i < temp.size(); ++i)
  {
    snp_loci.push_back(temp[i].pos);
    snp_labels.push_back(temp[i].label);
    snp_strands.push_back(temp[i].strand);
  }
  cerr << ". done." << endl;
}


int chrom_sizes[] =
{
  16571, //M
  245522847, //1
  243018229,
  199505740,
  191411218,
  180857866,
  170975699,
  158628139,
  146274826,
  138429268,
  135413628, //10
  134452384,
  132449811,
  114142980,
  106368585,
  100338915,
  88827254,
  78774742,
  76117153,
  63811651,
  62435964, //20
  46944323,
  49554710,
  154824264, //X
  57701691 //Y
};

int par1_size = 2692881;
int par2_size = 329518;


void ReadGeneticMap(IntervalMap<double, int64> &rates)
{
  rates.clear();
  for (int chrom = 0; chrom <= 24; ++chrom)
    rates[TSegment<int64>((int64(chrom) << 32) + 1, chrom_sizes[chrom])] = 0.0;

  cerr << "Reading genetic map";
  for (int chrom = 1; chrom <= 25; ++chrom)
  {
    string chrom_str = ToStr(chrom);
    if (chrom == 23)
      chrom_str = "X_non-par";
    else if (chrom == 24)
      chrom_str = "X_par1";
    else if (chrom == 25)
      chrom_str = "X_par2";
    string filename = GeneticMapFile(chrom_str);
    istream *in = InFileStream(filename);
    AssertMsg(in, filename);

    string line;
    Assert(getline(*in, line));

    int last = -1;
    double last_rate = 0.0;
    while (getline(*in, line))
    {
      int p = 0;
      SkipWhitespace(line, p);
      if (p == line.length())
        continue;

      int pos;
      Assert(StringToInt(line, pos, p, p));
      Assert(isspace(line[p]));
      if (last >= 0)
      {
        int64 coord_base = int64((chrom <= 23) ? chrom : 23) << 32;
        TSegment<int64> range(coord_base + last, coord_base + pos, true);
        rates[range] = last_rate;
      }
      
      last = pos;
      Assert(SkipWhitespace(line, p));
      int q = p;
      Assert(SkipNonWhitespace(line, q));
      StrTo(line.substr(p, q - p), last_rate);
    }

    delete in;
    cerr << ".";
  }
  cerr << " done." << endl;
}


double GeneticDistance(const IntervalMap<double, int64> &rates,
                       TSegment<int64> range)
{
  IntervalMap<double, int64>::const_iterator begin_iter = rates.lower_bound(range);
  IntervalMap<double, int64>::const_iterator end_iter = rates.upper_bound(range);
  double total_cM = 0.0;
  for (IntervalMap<double, int64>::const_iterator iter = begin_iter;
       iter != end_iter; ++iter)
  {
    TSegment<int64> intersection;
    if (Intersect(range, iter->first, intersection))
      total_cM += iter->second * double(intersection.len()) / 1000000.0;
  }

  return total_cM / 100.0;
}


int par2_x_y_delta = 154494747 - 57372174;


string ChromLabel(int chrom)
{
  if (chrom == 0)
    return "mito";
  else if (chrom <= 22)
    return "chr" + ToStr(chrom);
  else
    return "chrXY";
}

void CreateSNPLoci(SNPProfile &profile)
{
  vector<int64> all_genomic_pos;
  vector<string> all_snp_labels;
  vector<int8> all_snp_strands;
  ScanAllSNPLoci(all_genomic_pos, all_snp_labels, all_snp_strands);

  vector< pair< vector<int64>::iterator, vector<int64>::iterator > > chrom_genomic_pos_ranges;
  for (int chrom = 0; chrom <= 24; ++chrom)
  {
    chrom_genomic_pos_ranges.push_back(pair< vector<int64>::iterator, vector<int64>::iterator >(lower_bound(all_genomic_pos.begin(), all_genomic_pos.end(), int64(chrom) << 32), upper_bound(all_genomic_pos.begin(), all_genomic_pos.end(), int64(chrom + 1) << 32)));
    Assert(chrom_genomic_pos_ranges.back().first != chrom_genomic_pos_ranges.back().second);
  }
  pair< vector<int64>::iterator, vector<int64>::iterator > par1_genomic_pos_range(lower_bound(all_genomic_pos.begin(), all_genomic_pos.end(), int64(23) << 32), upper_bound(all_genomic_pos.begin(), all_genomic_pos.end(), (int64(23) << 32) + 2692882));
  Assert(par1_genomic_pos_range.first != par1_genomic_pos_range.second);
  pair< vector<int64>::iterator, vector<int64>::iterator > x_nonpar_genomic_pos_range(lower_bound(all_genomic_pos.begin(), all_genomic_pos.end(), (int64(23) << 32) + 2692882), upper_bound(all_genomic_pos.begin(), all_genomic_pos.end(), (int64(23) << 32) + 154494747));
  Assert(x_nonpar_genomic_pos_range.first != x_nonpar_genomic_pos_range.second);
  pair< vector<int64>::iterator, vector<int64>::iterator > par2_genomic_pos_range(lower_bound(all_genomic_pos.begin(), all_genomic_pos.end(), (int64(23) << 32) + 154494747), upper_bound(all_genomic_pos.begin(), all_genomic_pos.end(), int64(24) << 32));
  Assert(par2_genomic_pos_range.first != par2_genomic_pos_range.second);
  pair< vector<int64>::iterator, vector<int64>::iterator > y_nonpar_genomic_pos_range(lower_bound(all_genomic_pos.begin(), all_genomic_pos.end(), (int64(24) << 32) + 2692882), upper_bound(all_genomic_pos.begin(), all_genomic_pos.end(), (int64(24) << 32) + 57372174));
  Assert(y_nonpar_genomic_pos_range.first != y_nonpar_genomic_pos_range.second);
  IntervalMap<double, int64> rates;
  ReadGeneticMap(rates);

  cerr << "Preparing SNP map...";
  profile.loci.reserve(all_genomic_pos.size());

  for (int chrom = 0; chrom <= 22; ++chrom)
  {
    if (chrom == 0)
      profile.mito_range.SetLeft(profile.loci.size());

    int64 last_pos = -1;
    for (vector<int64>::iterator iter = chrom_genomic_pos_ranges[chrom].first;
         iter != chrom_genomic_pos_ranges[chrom].second; ++iter)
    {
      double genetic_pos;
      double gender_genetic_pos[2];
      if (last_pos < 0)
      {
        if (chrom == 0)
          genetic_pos = gender_genetic_pos[0] = gender_genetic_pos[1] = 0.0;
        else
        {
          genetic_pos = profile.loci.back().genetic_pos + 0.5;
          gender_genetic_pos[0] = profile.loci.back().gender_genetic_pos[0] + 0.5;
          gender_genetic_pos[1] = profile.loci.back().gender_genetic_pos[1] + 0.5;
        }
      }
      else
      {
        double genetic_dist = GeneticDistance(rates, TSegment<int64>(last_pos, *iter, true));
        Assert(genetic_dist >= 0.0);
        genetic_pos = profile.loci.back().genetic_pos + genetic_dist;
        gender_genetic_pos[0] = profile.loci.back().gender_genetic_pos[0] + 1.6*genetic_dist;
        gender_genetic_pos[1] = profile.loci.back().gender_genetic_pos[1] + 0.4*genetic_dist;
      }

      profile.loci.push_back(SNPProfile::Locus(
        all_snp_labels[iter - all_genomic_pos.begin()],
        all_snp_strands[iter - all_genomic_pos.begin()],
        *iter >> 32, *iter & 0xffffffff,
        genetic_pos, gender_genetic_pos[0], gender_genetic_pos[1]));

      last_pos = *iter;
    }

    if (chrom == 0)
      profile.mito_range.SetRight(profile.loci.size());
  }

  double genetic_dist_par1;
  {
    vector<int64>::iterator end = par1_genomic_pos_range.second;
    --end;
    genetic_dist_par1 = GeneticDistance(rates, TSegment<int64>(*(par1_genomic_pos_range.first), *end, true));
    Assert(genetic_dist_par1 >= 0.0);
  }
  
  double genetic_dist_par2;
  {
    vector<int64>::iterator end = par2_genomic_pos_range.second;
    --end;
    genetic_dist_par2 = GeneticDistance(rates, TSegment<int64>(*(par2_genomic_pos_range.first), *end, true));
    Assert(genetic_dist_par2 >= 0.0);
  }

  int64 last_pos = -1;
  profile.par1_range.SetLeft(profile.loci.size());
  for (vector<int64>::iterator iter = par1_genomic_pos_range.first;
       iter != par1_genomic_pos_range.second; ++iter)
  {
    double genetic_pos;
    double gender_genetic_pos[2];
    if (last_pos < 0)
    {
      genetic_pos = profile.loci.back().genetic_pos + 0.5;
      gender_genetic_pos[0] = profile.loci.back().gender_genetic_pos[0] + 0.5;
      gender_genetic_pos[1] = profile.loci.back().gender_genetic_pos[1] + 0.5;
    }
    else
    {
      double genetic_dist = GeneticDistance(rates, TSegment<int64>(last_pos, *iter, true));
      Assert(genetic_dist >= 0.0);
      genetic_pos = profile.loci.back().genetic_pos + genetic_dist;
      //profile.gender_genetic_pos[0].push_back(profile.gender_genetic_pos[0].back() + genetic_dist * (2.0 - 1.0/(genetic_dist_par1 + genetic_dist_par2)));
      gender_genetic_pos[0] = profile.loci.back().gender_genetic_pos[0] + genetic_dist;
      gender_genetic_pos[1] = profile.loci.back().gender_genetic_pos[1] + genetic_dist / (genetic_dist_par1 + genetic_dist_par2);
    }

    profile.loci.push_back(SNPProfile::Locus(
      all_snp_labels[iter - all_genomic_pos.begin()],
      all_snp_strands[iter - all_genomic_pos.begin()],
      23, *iter & 0xffffffff,
      genetic_pos, gender_genetic_pos[0], gender_genetic_pos[1]));

    last_pos = *iter;
  }
  profile.par1_range.SetRight(profile.loci.size());

  profile.y_nonpar_range.SetLeft(profile.loci.size());
  for (vector<int64>::iterator iter = y_nonpar_genomic_pos_range.first;
       iter != y_nonpar_genomic_pos_range.second; ++iter)
  {
    profile.loci.push_back(SNPProfile::Locus(
      all_snp_labels[iter - all_genomic_pos.begin()],
      all_snp_strands[iter - all_genomic_pos.begin()],
      23, *iter & 0xffffffff, profile.loci.back().genetic_pos, 
      profile.loci.back().gender_genetic_pos[0], 
      profile.loci.back().gender_genetic_pos[1]));
  }
  profile.y_nonpar_range.SetRight(profile.loci.size());

  profile.x_nonpar_range.SetLeft(profile.loci.size());
  for (vector<int64>::iterator iter = x_nonpar_genomic_pos_range.first;
       iter != x_nonpar_genomic_pos_range.second; ++iter)
  {
    double genetic_dist = GeneticDistance(rates, TSegment<int64>(last_pos, *iter, true));
    Assert(genetic_dist >= 0.0);

    profile.loci.push_back(SNPProfile::Locus(
      all_snp_labels[iter - all_genomic_pos.begin()],
      all_snp_strands[iter - all_genomic_pos.begin()],
      23, *iter & 0xffffffff, profile.loci.back().genetic_pos + genetic_dist, 
      profile.loci.back().gender_genetic_pos[0] + 4.0/3.0*genetic_dist, 
      profile.loci.back().gender_genetic_pos[1]));

    last_pos = *iter;
  }
  profile.x_nonpar_range.SetRight(profile.loci.size());

  profile.par2_range.SetLeft(profile.loci.size());
  for (vector<int64>::iterator iter = par2_genomic_pos_range.first;
       iter != par2_genomic_pos_range.second; ++iter)
  {
    double genetic_dist = GeneticDistance(rates, TSegment<int64>(last_pos, *iter, true));
    Assert(genetic_dist >= 0.0);

    profile.loci.push_back(SNPProfile::Locus(
      all_snp_labels[iter - all_genomic_pos.begin()],
      all_snp_strands[iter - all_genomic_pos.begin()],
      23, *iter & 0xffffffff, profile.loci.back().genetic_pos + genetic_dist, 
      profile.loci.back().gender_genetic_pos[0] + genetic_dist,
      profile.loci.back().gender_genetic_pos[1] + genetic_dist / (genetic_dist_par1 + genetic_dist_par2)));

    //profile.gender_genetic_pos[0].push_back(profile.gender_genetic_pos[0].back() + genetic_dist * (2.0 - 1.0/(genetic_dist_par1 + genetic_dist_par2)));
    last_pos = *iter;
  }
  profile.par2_range.SetRight(profile.loci.size());

  for (int chrom = 0; chrom <= 22; ++chrom)
  {
    int begin = -1, end = -1;
    for (int i = 0; i < profile.loci.size(); ++i)
      if (profile.loci[i].chrom == chrom)
      {
        begin = i;
        break;
      }
    for (int i = profile.loci.size() - 1; i >= 0; --i)
      if (profile.loci[i].chrom == chrom)
      {
        end = i + 1;
        break;
      }
    AssertMsg((begin >= 0) && (end >= 0), ToStr(chrom));
    profile.chroms.push_back(SNPProfile::Chrom(ChromLabel(chrom), Segment(begin, end, true)));
  }
  profile.chroms.push_back(SNPProfile::Chrom(ChromLabel(23), Segment(profile.par1_range.left(), profile.par2_range.right(), true)));
  
  profile.sry_genomic_pos = 2715337;
  profile.x_y_par2_genomic_pos_delta = par2_x_y_delta;

  profile.ComputeDerived();

  cerr << " done." << endl;
}


void CreatePopData(SNPData &data)
{
  for (int pop = 0; pop < 4; ++pop)
    data.pops.push_back(SNPData::Pop(PopStr(pop)));

  data.num_pops = data.pops.size();
}


void CreateHaploDiploData(const SNPProfile &profile, SNPData &data)
{
  cerr << "Reading HapMap SNP data:" << endl;

  data.num_diplos = 60 + 60 + 45 + 45;
  data.num_haplos = 2 * data.num_diplos;

  data.haplos.resize(data.num_haplos);
  for (int k = 0; k < data.haplos.size(); ++k)
    data.haplos[k].data.resize(profile.seq_len, 4);

  data.diplos.resize(data.num_diplos);

  map<int64, int> genomic_pos_i;
  for (int i = 0; i < profile.loci.size(); ++i)
  {
    int chrom = profile.loci[i].chrom;
    int32 pos = profile.loci[i].genomic_pos;
    if ((i >= profile.par1_range.left()) && (i < profile.par1_range.right()))
      chrom = 25;
    if ((i >= profile.par2_range.left()) && (i < profile.par2_range.right()))
      chrom = 25;
    if ((i >= profile.y_nonpar_range.left()) && (i < profile.y_nonpar_range.right()))
      chrom = 24;
    if ((i >= profile.x_nonpar_range.left()) && (i < profile.x_nonpar_range.right()))
      chrom = 23;
    genomic_pos_i[(int64(chrom) << 32) | pos] = i;
  }

  map<string, int> NA_diplo;
  vector<int> diplo_gender(data.num_diplos, -1);
  {
    int diplo_base = 0;
    for (int cpop = 0; cpop < 3; ++cpop)
    {
      istream *in = InFileStream(SampleFile(cpop));
      AssertMsg(in, SampleFile(cpop));
      string line;
      int pop_num_diplos = (cpop < 2) ? 60 : 90;
      for (int d = 0; d < pop_num_diplos; ++d)
      {
        Assert(getline(*in, line));
        int p = 0;
        SkipWhitespace(line, p);
        int q = p;
        Assert(SkipNonWhitespace(line, p));

        string NA_label = line.substr(q, p - q);
        Assert(SkipWhitespace(line, p));
        
        int gender;
        Assert(StringToInt(line, gender, p, p));
        Assert((p == line.length()) || isspace(line[p]));

        NA_diplo[NA_label] = diplo_base + d;
        diplo_gender[diplo_base + d] = (gender == 1) ? 1 : 0;
      }
      diplo_base += pop_num_diplos;
    }
  }

  {
    int diplo_base = 0;
    int haplo_base = 0;
    for (int pop = 0; pop < 4; ++pop)
    {
      int pop_num_diplos = (pop < 2) ? 60 : 45;
      for (int d = 0; d < pop_num_diplos; ++d)
      {
        data.diplos[diplo_base + d].label = "hapmap_" + PopStr(pop) + "_" + StrFitRight(d, 3, '0');
        data.diplos[diplo_base + d].haplos[0] = haplo_base + 2*d + 0;
        data.diplos[diplo_base + d].haplos[1] = haplo_base + 2*d + 1;

        for (int h = 0; h < 2; ++h)
        {
          int k = haplo_base + 2*d + h;
          data.haplos[k].label = "hapmap_" + PopStr(pop) + "_" + StrFitRight(d, 3, '0') + "." + ToStr(h);
          data.haplos[k].sry = 0;
          data.haplos[k].pop = pop;
        }
      }
      diplo_base += pop_num_diplos;
      haplo_base += 2*pop_num_diplos;
    }
  }

  for (int chrom = 1; chrom <= 25; ++chrom)
  //for (int chrom = 1; chrom <= 23; chrom += 22)
  //for (int chrom = 23; chrom <= 25; ++chrom)
  {
    string chrom_str = ToStr(chrom), chrom_str2;
    if (chrom >= 23)
    {
      chrom_str = "X";
      if (chrom == 23)
        chrom_str2 = "non-par";
      else if (chrom == 24)
        chrom_str2 = "par1";
      else if (chrom == 25)
        chrom_str2 = "par2";
    }

    cerr << "- Chromosome " << chrom_str << (chrom_str2.empty() ? "" : " (") << chrom_str2 << (chrom_str2.empty() ? "" : ")") << ":";
    int haplo_base = 0;
    int diplo_base = 0;

    for (int cpop = 0; cpop < 3; ++cpop)
    {
      cerr << " " << CPopStr(cpop);

      vector<int> legend_i;
      legend_i.reserve(300000);
      vector< vector<int8> > legend_c(2);
      legend_c[0].reserve(300000);
      legend_c[1].reserve(300000);
      
      istream *in = InFileStream(LegendFile(cpop, chrom_str, chrom_str2));
      AssertMsg(in, LegendFile(cpop, chrom_str, chrom_str2));
      string line;
      Assert(getline(*in, line));

      while (getline(*in, line))
      {
        int p = 0;
        SkipWhitespace(line, p);
        if (p == line.length())
          continue;

        Assert(SkipNonWhitespace(line, p));
        Assert(SkipWhitespace(line, p));

        int pos;
        int q = p;
        Assert(SkipNonWhitespace(line, q));
        double fpos;
        StrTo(line.substr(p, q - p), fpos);
        pos = int(fpos);
        p = q;

        Assert(SkipWhitespace(line, p));
        Assert(p != line.length());
        char c0 = line[p++];

        Assert(SkipWhitespace(line, p));
        Assert(p != line.length());
        char c1 = line[p++];
        Assert((p == line.length()) || isspace(line[p]));

        map<int64, int>::iterator find = genomic_pos_i.find((int64((chrom <= 23) ? chrom : 25) << 32) + pos);
        Assert(find != genomic_pos_i.end());
        legend_i.push_back(find->second);

        legend_c[0].push_back((char2base[c0] >= 0) ? char2base[c0] : 4);
        legend_c[1].push_back((char2base[c1] >= 0) ? char2base[c1] : 4);
      }

      delete in;
      cerr << ".";

      in = InFileStream(PhasedFile(cpop, chrom_str, chrom_str2));
      AssertMsg(in, PhasedFile(cpop, chrom_str, chrom_str2));
      
      int pop_num_diplos = (cpop < 2) ? 60 : 90;
      for (int d = 0; d < pop_num_diplos; ++d)
      {
        int pop = (cpop < 2) ? cpop : ((d < 45) ? 2 : 3);
        for (int h = 0; h < 2; ++h)
        {
          int k = haplo_base + 2*d + h;

          Assert(getline(*in, line));
          int p = 0;
          
          for (int j = 0; j < legend_i.size(); ++j)
          {
            SkipWhitespace(line, p);
            Assert(p < line.length());
            if (line[p] == '-')
            {
              Assert(chrom == 23);
              Assert(diplo_gender[diplo_base + d] == 1);
              data.haplos[k].sry = 1;
              //cerr << "chrom=" << chrom << ", cpop=" << cpop << ", d=" << d << ", sry=1" << endl;
              break;
            }

            AssertMsg((line[p] == '0') || (line[p] == '1'), ToStr(line[p]));
            int v = (line[p++] - '0');
            data.haplos[k][legend_i[j]] = legend_c[v][j];

            Assert((p == line.length()) || isspace(line[p]));
          }
        }

      }
      
      while (getline(*in, line))
      {
        int p = 0;
        SkipWhitespace(line, p);
        Assert(p == line.length());
      }

      delete in;
      cerr << ".";

      diplo_base += pop_num_diplos;
      haplo_base += 2*pop_num_diplos;
    }
    cerr << endl;
  }  

  for (int chrom_index = 0; chrom_index < 2; ++chrom_index)
  {
    int chrom = (chrom_index == 0) ? 0 : 24;
    string chrom_str = ChromStr(chrom);
    cerr << "- Chromosome " << chrom_str << ":";
    
    for (int pop = 0; pop < 4; ++pop)
    {
      cerr << " " << PopStr(pop);
      int pop_num_diplos = (pop < 2) ? 60 : 45;
      
      istream *in = InFileStream(GenotypesFile(PopStr(pop), chrom));
      AssertMsg(in, GenotypesFile(PopStr(pop), chrom));
      string line;

      vector<int> diplos(pop_num_diplos, -1);
      {
        Assert(getline(*in, line));
        int p = 0;
        SkipWhitespace(line, p);
        for (int column = 0; column < 11; ++column)
        {
          Assert(SkipNonWhitespace(line, p));
          Assert(SkipWhitespace(line, p));
        }

        for (int d = 0; d < pop_num_diplos; ++d)
        {
          int q = p;
          Assert(SkipNonWhitespace(line, p));
          string NA_label = line.substr(q, p - q);
          map<string, int>::iterator find = NA_diplo.find(NA_label);
          if (find == NA_diplo.end())
            diplos[d] = -1;
          else
            diplos[d] = find->second;
          if (p < line.length())
            Assert(SkipWhitespace(line, p));
        }
      }

      while (getline(*in, line))
      {
        int p = 0;
        SkipWhitespace(line, p);
        if (p == line.length())
          continue;

        for (int column = 0; column < 3; ++column)
        {
          Assert(SkipNonWhitespace(line, p));
          Assert(SkipWhitespace(line, p));
        }

        int pos;
        Assert(StringToInt(line, pos, p, p));
        Assert((p < line.length()) && isspace(line[p]));

        map<int64, int>::iterator find = genomic_pos_i.find((int64((chrom_index == 0) ? 0 : 24) << 32) + pos);
        Assert(find != genomic_pos_i.end());
        int i = find->second;
        bool rc_strand = data.loci[i].strand;

        for (int column = 0; column < 7; ++column)
        {
          Assert(SkipWhitespace(line, p));
          Assert(SkipNonWhitespace(line, p));
        }

        for (int d = 0; d < pop_num_diplos; ++d)
        {
          int diplo = diplos[d];

          Assert(SkipWhitespace(line, p));
          Assert((p + 2) <= line.length());
          
          int b0 = char2base[line[p++]];
          if (b0 < 0)
            b0 = 4;
          else if (rc_strand)
            b0 = 3 - b0;
          int b1 = char2base[line[p++]];
          if (b1 < 0)
            b1 = 4;
          else if (rc_strand)
            b1 = 3 - b1;
          if (b0 != b1)
            b0 = b1 = 4;
          Assert((p == line.length()) || isspace(line[p]));

          if (diplo < 0)
            continue;

          int h0 = data.diplos[diplo].haplos[0];
          int h1 = data.diplos[diplo].haplos[1];
          Assert((h0 >= 0) && (h1 >= 0));

          if (chrom_index == 0)
          {
            data.haplos[h0][i] = b0;
            data.haplos[h1][i] = b1;
          }
          else
          {
            if (!data.haplos[h0].sry && !data.haplos[h1].sry)
            {
              Assert(diplo_gender[diplo] == 0);
              b0 = b1 = 4;
              //AssertMsg((b0 == 4) && (b1 == 4), ToStr(pos) + ": " + ToStr(d) + ": " + ToStr(b0) + ", " + ToStr(b1));
            }
            else
            {
              Assert(diplo_gender[diplo] == 1);
              Assert(!(data.haplos[h0].sry && data.haplos[h1].sry));
              int b = min(b0, b1);
              if (data.haplos[h0].sry)
                data.haplos[h0][i] = b;
              else
                data.haplos[h1][i] = b;
            }
          }
        }
      }
    }
    cerr << endl;
  }
}


double pop_switch_interval[] =
{
  8000000.0,
  3600000.0,
  340000.0,
  340000.0
};

void CreateDiploSwitchData(const SNPProfile &profile, SNPData &data)
{
  cerr << "Computing switch probabilities:" << endl;
  for (int d = 0; d < data.num_diplos; ++d)
    data.diplos[d].switches.resize(profile.seq_len - 1, 0);

  vector< vector<double> > switches(2, vector<double>(profile.seq_len - 1, 0.0));
  for (int pop = 0; pop < data.num_pops; ++pop)
  {
    cerr << "- " << data.pops[pop].label;
    for (int gender = 0; gender < 2; ++gender)
    {
      for (int chrom = 0; chrom <= 22; ++chrom)
      {
        for (int i = profile.chroms[chrom].range.left();
             i < (profile.chroms[chrom].range.right() - 1); ++i)
        {
          int32 separation = profile.locus_genomic_pos[i + 1] - profile.locus_genomic_pos[i];
          Assert(separation >= 0);
          switches[gender][i] = double(separation) / pop_switch_interval[pop];
        }
        switches[gender][profile.chroms[chrom].range.right() - 1] = 0.5;
      }

      for (int i = profile.par1_range.left(); i < (profile.par1_range.right() - 1); ++i)
      {
        int32 separation = profile.locus_genomic_pos[i + 1] - profile.locus_genomic_pos[i];
        switches[gender][i] = double(separation) / pop_switch_interval[pop];
      }

      if (gender == 0)
      {
        int32 separation = profile.locus_genomic_pos[profile.x_nonpar_range.left()] - profile.locus_genomic_pos[profile.par1_range.right() - 1];
        switches[gender][profile.par1_range.right() - 1] = double(separation) / pop_switch_interval[pop];

        for (int i = profile.y_nonpar_range.left(); i < profile.y_nonpar_range.right(); ++i)
          switches[gender][i] = 0.0;

        for (int i = profile.x_nonpar_range.left(); i < (profile.x_nonpar_range.right() - 1); ++i)
        {
          int32 separation = profile.locus_genomic_pos[i + 1] - profile.locus_genomic_pos[i];
          Assert(separation >= 0);
          switches[gender][i] = double(separation) / pop_switch_interval[pop];
        }
        
        separation = profile.locus_genomic_pos[profile.par2_range.left()] - profile.locus_genomic_pos[profile.x_nonpar_range.right() - 1];
        switches[gender][profile.x_nonpar_range.right() - 1] = double(separation) / pop_switch_interval[pop];
      }
      else
      {
        int32 separation = profile.locus_genomic_pos[profile.y_nonpar_range.left()] - profile.locus_genomic_pos[profile.par1_range.right() - 1];
        switches[gender][profile.par1_range.right() - 1] = double(separation) / pop_switch_interval[pop];

        for (int i = profile.y_nonpar_range.left(); i < (profile.y_nonpar_range.right() - 1); ++i)
        {
          int32 separation = profile.locus_genomic_pos[i + 1] - profile.locus_genomic_pos[i];
          Assert(separation >= 0);
          switches[gender][i] = double(separation) / pop_switch_interval[pop];
        }
        switches[gender][profile.y_nonpar_range.right() - 1] = 0.0;

        for (int i = profile.x_nonpar_range.left(); i < (profile.x_nonpar_range.right() - 1); ++i)
          switches[gender][i] = 0.0;
        
        separation = (profile.locus_genomic_pos[profile.par2_range.left()] - par2_x_y_delta) - profile.locus_genomic_pos[profile.y_nonpar_range.right() - 1];
        switches[gender][profile.x_nonpar_range.right() - 1] = double(separation) / pop_switch_interval[pop];
      }        

      for (int i = profile.par2_range.left(); i < (profile.par2_range.right() - 1); ++i)
      {
        int32 separation = profile.locus_genomic_pos[i + 1] - profile.locus_genomic_pos[i];
        switches[gender][i] = double(separation) / pop_switch_interval[pop];
      }
    }

    for (int diplo = 0; diplo < data.num_diplos; ++diplo)
    {
      int h0 = data.diplos[diplo].haplos[0];
      int h1 = data.diplos[diplo].haplos[1];
      AssertMsg(data.haplos[h0].pop == data.haplos[h1].pop, ToStr(diplo) + "=<" + ToStr(h0) + ", " + ToStr(h1) + ">");
      if (data.haplos[h0].pop != pop)
        continue;

      Assert(!(data.haplos[h0].sry && data.haplos[h1].sry));
      int gender = (!data.haplos[h0].sry && !data.haplos[h1].sry) ? 0 : 1;

      double total_switch = 0.0;
      for (int i = (profile.seq_len - 1); i >= 0; --i)
      {
        if (switches[gender][i] >= 0.5)
        {
          data.diplos[diplo].switches[i] = SNPData::switch_encode(0.5);
          total_switch = 0.0;
        }
        else
        {
          total_switch += switches[gender][i];
          if (total_switch > 0.5)
            total_switch = 0.5;
          
          if ((data.haplos[h0][i] == data.haplos[h1][i]) &&
              (data.haplos[h0][i] != 4))
            data.diplos[diplo].switches[i] = SNPData::switch_encode(0.0);
          else
          {
            data.diplos[diplo].switches[i] = SNPData::switch_encode(total_switch);
            total_switch = 0.0;
          }
        }
      }
      cerr << ".";
    }
    cerr << endl;
  }
}


void WriteDataChroms(const SNPProfile &profile, const SNPData &data, 
                     string filename_prefix)
{
  string pathname;
  RelPath(output_dir, filename_prefix + ".snpprofiledata.bin.gz", pathname);
  cerr << "Writing " << pathname << " ...";
  WriteBinSNPProfileData(pathname, data);
  cerr << " done." << endl;

  for (int chrom_index = 0; chrom_index <= 23; ++chrom_index)
  {
    int chrom;
    for (chrom = 0; chrom < profile.chroms.size(); ++chrom)
      if (profile.chroms[chrom].label == ChromLabel(chrom_index))
        break;
    if (chrom >= profile.chroms.size())
      continue;

    SNPProfile chrom_profile;
    SNPData chrom_data;
    chrom_profile.CopyRestrictChrom(profile, chrom);
    if (chrom_profile.seq_len > 0)
    {
      chrom_data.SetProfile(chrom_profile);
      chrom_data.CopyRestrictChrom(data, chrom);
      chrom_profile.Validate();
      chrom_data.Validate();
      
      RelPath(output_dir, filename_prefix + "_" + profile.chroms[chrom].label + ".snpprofiledata.bin.gz", pathname);
      cerr << "Writing " << pathname << " ...";
      WriteBinSNPProfileData(pathname, chrom_data);
      cerr << " done." << endl;
    }
  }
}


int Main(vector<string> args)
{
  cerr << "CreateHapMapSNPData" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;

  SNPProfile profile;
  SNPData data;

  CreateSNPLoci(profile);
  profile.ComputeDerived();
  profile.Validate();
  
  data.SetProfile(profile);
  CreatePopData(data);
  CreateHaploDiploData(profile, data);
  CreateDiploSwitchData(profile, data);
  data.ComputeDerived();
  data.Validate();
  
  WriteDataChroms(profile, data, "hapmap");
  
  cerr << "Done!" << endl;
  return 0;
}
