

#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"
#include "Segment.h"


Params params(false);

int debug = 0;
string input_dir, output_dir;
string input_file, output_file;


string required_param[] =
{
  "input_file",
  "output_file"
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

void InitParams()
{
  params.ExpandImports();

  srand(TimerSeconds());
  if (params.Contains("rand_seed"))
    srand(params["rand_seed"].GetInt());

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("input_file"))
    input_file = params["input_file"].GetString();

  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (params.Contains("output_file"))
  {
    output_file = params["output_file"].GetString();
    string filetype = RemoveCompressedExtension(output_file);
    Assert(GetExtension(filetype) == "bin");
    filetype = GetExtension(RemoveExtension(filetype));
    Assert(filetype == "snpprofile");
  }

  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;

  cerr << "- Input file = " << input_file << endl;

  if (!output_dir.empty())
    cerr << "- Output directory = " << output_dir << endl;

  cerr << "- Output file = " << output_file << endl;
}



struct LocusLess
{
  bool operator()(const SNPProfile::Locus &x, const SNPProfile::Locus &y) const
  {
    if (x.chrom != y.chrom)
      return x.chrom < y.chrom;
    else if (x.genomic_pos != y.genomic_pos)
      return x.genomic_pos < y.genomic_pos;
    else
      return x.genetic_pos < y.genetic_pos;
  }
};



int Main(vector<string> args)
{
  cerr << "CreateSNPProfile" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;

  SNPProfile profile;
  map<string, int> chrom_names;

  string pathname;
  RelPath(input_dir, input_file, pathname);
  cerr << "* Reading input file " << pathname << " ...";

  istream *in = InFileStream(pathname);
  AssertMsg(in, pathname);
  
  string line;
  vector<string> fields;
  while (getline(*in, line))
  {
    SplitFields(line, fields, " \t,");
    if (fields.empty() || (fields[0][0] == ';') || (fields[0][0] == '#'))
      continue;
    
    AssertMsg(fields.size() >= 3, line);
    string chrom = fields[0];
    int chrom_index;
    map<string, int>::const_iterator find = chrom_names.find(chrom);
    if (find == chrom_names.end())
    {
      chrom_index = chrom_names[chrom] = profile.chroms.size();
      profile.chroms.push_back(SNPProfile::Chrom(chrom, Segment()));
    }
    else
      chrom_index = find->second;

    int genomic_pos;
    AssertMsg(StringToInt(fields[1], genomic_pos), fields[1]);

    double genetic_pos;
    StrTo(fields[2], genetic_pos);

    string label;
    if (fields.size() >= 4)
      label = fields[3];

    profile.loci.push_back(SNPProfile::Locus(label, false, chrom_index,
                                             genomic_pos, genetic_pos,
                                             genetic_pos, genetic_pos));
  }

  delete in;
  cerr << " done." << endl;

  sort(profile.loci.begin(), profile.loci.end(), LocusLess());

  double genetic_offset = 0.0;
  for (int i = 0; i < profile.loci.size(); )
  {
    int j = i + 1;
    while ((j < profile.loci.size()) && 
           (profile.loci[i].chrom == profile.loci[j].chrom))
      ++j;
    
    for (int k = i; k < j; ++k)
    {
      profile.loci[k].genetic_pos += genetic_offset;
      profile.loci[k].gender_genetic_pos[0] += genetic_offset;
      profile.loci[k].gender_genetic_pos[1] += genetic_offset;
    }

    int chrom = profile.loci[i].chrom;
    profile.chroms[chrom].range = Segment(i, j, true);

    i = j;
    if (i < profile.loci.size())
      genetic_offset = (profile.loci[i - 1].genetic_pos + 0.5) - profile.loci[i].genetic_pos;
  }

  profile.sry_genomic_pos = profile.loci.size();
  profile.x_y_par2_genomic_pos_delta = 0;
  profile.par1_range = Segment(0, 0);
  profile.par2_range = Segment(0, 0);
  profile.x_nonpar_range = Segment(0, 0);
  profile.y_nonpar_range = Segment(0, 0);
  profile.mito_range = Segment(0, 0);

  profile.ComputeDerived();
  profile.Validate();

  RelPath(output_dir, output_file, pathname);
  cerr << "* Writing file " << pathname << " ...";
  ostream *out = OutFileStream(pathname);
  AssertMsg(out, pathname);

  profile.WriteBin(*out);

  delete out;
  cerr << " done." << endl;
  
  cerr << "Done!" << endl;
  return 0;
}
