
#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "HaploModelGraph.h"
#include "CairoUtil.h"
#include "SNPData.h"


Params params(false);

int debug = 0;

string input_dir;
vector<string> input_files;
string output_dir;
string output_file;
bool ignore_missing = false;
vector<int> select_graphs;

int image_size_x = 1024;
int image_size_y = 768;

double line_width = 1.0;
double no_grid = false;


string required_param[] =
{
  "output_file"
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

void GetIntegerSpec(string s, vector<int> &result);

void InitParams()
{
  params.ExpandImports();

  uint32 seed = RandomSeed();
  if (params.Contains("rand_seed"))
    seed = params["rand_seed"].GetInt();
  srand(seed);
  cerr << "- Random seed = " << seed << endl;

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (params.Contains("input_file"))
  {
    string input_file = params["input_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(input_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(input_file))) == "haplomodelgraphs");
    input_files.push_back(input_file);
  }

  if (params.Contains("input_files"))
  {
    vector<ParamValue> values = params["input_files"].GetList();
    for (int i = 0; i < values.size(); ++i)
    {
      string filename = values[i].GetString();
      if (GetExtension(RemoveCompressedExtension(filename)) == "bin")
        AssertMsg((GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "haplopopdata") || (GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "diplopopdata") || (GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "haploposterior") || (GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "diploposterior") || (GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "snpprofile") || (GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "snpprofiledata") || (GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "phasingposterior"), GetExtension(RemoveExtension(RemoveCompressedExtension(filename))));
      else if (GetExtension(RemoveCompressedExtension(filename)) == "lst")
        AssertMsg((GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "saberviterbi") || (GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "saberposterior"), GetExtension(RemoveExtension(RemoveCompressedExtension(filename))));
      else
        AssertMsg(false, GetExtension(RemoveCompressedExtension(filename)));
      input_files.push_back(filename);
    }
  }

  if (params.Contains("output_file"))
  {
    output_file = params["output_file"].GetString();
    Assert(GetExtension(output_file) == "png");
  }

  if (params.Contains("image_size_x"))
    image_size_x = params["image_size_x"].GetInt();

  if (params.Contains("image_size_y"))
    image_size_y = params["image_size_Y"].GetInt();

  if (params.Contains("ignore_missing"))
    ignore_missing = true;

  if (params.Contains("select_graphs"))
    GetIntegerSpec(params["select_graphs"].GetString(), select_graphs);

  if (params.Contains("line_width"))
    line_width = params["line_width"].GetDouble();

  if (params.Contains("no_grid"))
    no_grid = true;

  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;

  cerr << "- Input files = {";
  for (int i = 0; i < input_files.size(); ++i)
    cerr << " " << input_files[i];
  cerr << " }" << endl;

  if (!output_dir.empty())
    cerr << "- Output directory = " << output_dir << endl;

  cerr << "- Output file = " << output_file << endl;

  cerr << "- Image size = " << image_size_x << " x " << image_size_y << endl;

  if (ignore_missing)
    cerr << "- Ignore missing data" << endl;

  if (!select_graphs.empty())
    cerr << "- Selected graphs { " << IntegerListString(select_graphs.begin(), select_graphs.end()) << " }" << endl;
}


void GetIntegerSpec(string s, vector<int> &result)
{
  int p = 0;
  while (p < s.length())
  {
    if (isdigit(s[p]))
    {
      int i;
      Assert(StringToInt(s, i, p, p));
      if ((p == s.length()) || (s[p] == ','))
      {
        result.push_back(i);
        if (p < s.length())
          ++p;
        continue;
      }

      Assert(s[p++] == '-');
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p == s.length()) || (s[p] == ','));
      if (i <= j)
      {
        for (int k = i; k <= j; ++k)
          result.push_back(k);
      }
      else
      {
        for (int k = i; k >= j; --k)
          result.push_back(k);
      }
      if (p < s.length());
      ++p;
    }
    else if (s[p] == '(')
    {
      int i;
      Assert(StringToInt(s, i, p, ++p));
      Assert((p < s.length()) && (s[p] == ','));
      ++p;
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p < s.length()) && (s[p] == ')'));
      Assert(j >= 0);
      for (int k = 0; k < j; ++k)
        result.push_back(i + k);
      ++p;
      Assert((p == s.length()) || (s[p] == ','));
      if (p < s.length())
        ++p;
    }
    else
      Assert(false);
  }
}


template <typename T, typename XT>
void GraphCode(cairo_t *cairo, const vector<T> &code,
               const vector<XT> &x_pos, double scale_x, double offset_x,
               double y0, double y1, const vector<uint32> &colors,
               int start = 0, int end = INT_MAX)
{
  if (end > int(code.size()))
    end = code.size();
  for (int i = start; i < end; )
  {
    int c = code[i];

    int j = i + 1;
    while ((j < end) && (code[j] == c))
      ++j;

    double x0 = (i > 0) ? 0.5*(double(x_pos[i - 1] + x_pos[i])) : double(x_pos[i]);
    double x1 = (j < int(x_pos.size())) ? 0.5*(double(x_pos[j - 1]) + double(x_pos[j])) : double(x_pos[j - 1]); 

    cairo_set_source_rgba(cairo, color_rgb(colors[c]), 1.0);
    cairo_rectangle(cairo, y0, x0 * scale_x + offset_x, y1 - y0, (x1 - x0) * scale_x);
    cairo_fill(cairo);
    i = j;
  }

  double x0 = (start > 0) ? 0.5*(double(x_pos[start - 1] + x_pos[start])) : double(x_pos[start]);
  double x1 = (end < int(x_pos.size())) ? 0.5*(double(x_pos[end - 1]) + double(x_pos[end])) : double(x_pos[end - 1]); 
  cairo_set_source_rgba(cairo, color_rgb(0x000000), 1.0);
  cairo_set_line_width(cairo, 1.0 * line_width);
  cairo_move_to(cairo, y0, x0 * scale_x + offset_x);
  cairo_rel_line_to(cairo, 0.0, (x1 - x0) * scale_x);
  cairo_move_to(cairo, y1, x0 * scale_x + offset_x);
  cairo_rel_line_to(cairo, 0.0, (x1 - x0) * scale_x);
  cairo_stroke(cairo);
}


template <typename T, typename XT>
void GraphProportion(cairo_t *cairo, 
                     const vector< vector<T> > &values, T top_value, T bottom_value,
                     const vector<XT> &x_pos, double scale_x, double offset_x,
                     double y0, double y1,
                     const vector<uint32> &colors, bool outline = true,
                     int start = 0, int end = INT_MAX)
{
  Assert(!values.empty());
  Assert(!values[0].empty());
  if (end > int(values[0].size()))
    end = values[0].size();

  double yheight = y1 - y0;
  double base_y = (top_value > bottom_value) ? y1 : y0;

  vector<T> row_p;
  row_p.resize(end - start, T(0));
  for (int t = 0; t < values.size(); ++t)
  {
    cairo_move_to(cairo, y0 + yheight*row_p[0], double(x_pos[start])*scale_x + offset_x);
    for (int i = start; i < end; ++i)
      cairo_line_to(cairo, y0 + yheight*row_p[i - start], double(x_pos[i])*scale_x + offset_x);
    cairo_line_to(cairo, y0 + yheight*row_p[end - 1 - start], double(x_pos[end - 1])*scale_x + offset_x);

    for (int i = start; i < end; ++i)
      row_p[i - start] += values[t][i];
    
    cairo_line_to(cairo, y0 + yheight*row_p[end - 1 - start], double(x_pos[end - 1])*scale_x + offset_x);
    for (int i = end - 1; i >= start; --i)
      cairo_line_to(cairo, y0 + yheight*row_p[i - start], double(x_pos[i])*scale_x + offset_x);
    cairo_line_to(cairo, y0 + yheight*row_p[0], double(x_pos[start])*scale_x + offset_x);
    
    cairo_set_source_rgba(cairo, color_rgb(colors[t]), 1.0);
    cairo_fill(cairo);
  }

  if (outline)
  {
    row_p.clear();
    row_p.resize(end - start, T(0));
    for (int t = 0; t < (values.size() - 1); ++t)
    {
      for (int i = start; i < end; ++i)
        row_p[i - start] += values[t][i];
      
      cairo_move_to(cairo, y0 + yheight*row_p[0], double(x_pos[start])*scale_x + offset_x);
      for (int i = start; i < end; ++i)
        cairo_line_to(cairo, y0 + yheight*row_p[i - start], double(x_pos[i])*scale_x + offset_x);
      cairo_line_to(cairo, y0 + yheight*row_p[end - 1 - start], double(x_pos[end - 1])*scale_x + offset_x);
      
      cairo_set_source_rgba(cairo, color_rgb(0x000000), 1.0);
      cairo_set_line_width(cairo, 1.0 * line_width);
      cairo_stroke(cairo);
    }
  }

  cairo_set_source_rgba(cairo, color_rgb(0x000000), 1.0);
  cairo_set_line_width(cairo, 1.0 * line_width);
  cairo_move_to(cairo, y0, double(x_pos[start]) * scale_x + offset_x);
  cairo_rel_line_to(cairo, 0.0, double(x_pos[end - 1] - x_pos[start]) * scale_x);
  cairo_move_to(cairo, y1, double(x_pos[start]) * scale_x + offset_x);
  cairo_rel_line_to(cairo, 0.0, double(x_pos[end - 1] - x_pos[start]) * scale_x);
  cairo_stroke(cairo);
}


template <typename T, typename XT>
void GraphGraph(cairo_t *cairo, 
                const vector<T> &values, T top_value, T bottom_value,
                const vector<XT> &x_pos, double scale_x, double offset_x,
                double y0, double y1,
                uint32 background, uint32 foreground,
                int start = 0, int end = INT_MAX)
{
  if (end > int(values.size()))
    end = values.size();

  cairo_set_source_rgba(cairo, color_rgb(background), 1.0);
  cairo_rectangle(cairo, y0, double(x_pos[start]) * scale_x + offset_x,
                  y1 - y0, double(x_pos[end - 1] - x_pos[start]) * scale_x);
  cairo_fill(cairo);

  double base_y = (top_value > bottom_value) ? y1 : y0;
  cairo_move_to(cairo, base_y, double(x_pos[start]) * scale_x + offset_x);
  for (int i = start; i < end; ++i)
  {
    double v = 0.0;
    v = (double(values[i]) - double(top_value)) / (double(bottom_value) - double(top_value));
    if (v < 0.0)
      v = 0.0;
    if (v > 1.0)
      v = 1.0;
    cairo_line_to(cairo, y0 + v*(y1 - y0), double(x_pos[i]) * scale_x + offset_x);
  }
  cairo_line_to(cairo, base_y, double(x_pos[end - 1]) * scale_x + offset_x);
    
  cairo_set_source_rgba(cairo, color_rgb(foreground), 1.0);
  cairo_fill(cairo);

  cairo_set_source_rgba(cairo, color_rgb(0x000000), 1.0);
  cairo_set_line_width(cairo, 1.0 * line_width);
  cairo_move_to(cairo, y0, double(x_pos[start]) * scale_x + offset_x);
  cairo_rel_line_to(cairo, 0.0, double(x_pos[end - 1] - x_pos[start]) * scale_x);
  cairo_move_to(cairo, y1, double(x_pos[start]) * scale_x + offset_x);
  cairo_rel_line_to(cairo, 0.0, double(x_pos[end - 1] - x_pos[start]) * scale_x);
  cairo_stroke(cairo);
}


void GraphHaploModelGraphs(HaploModelGraphs &graphs)
{
  if (!select_graphs.empty())
  {
    vector<bool> selected(graphs.graphs.size(), false);
    for (int i = 0; i < select_graphs.size(); ++i)
    {
      Assert((select_graphs[i] >= 0) && (select_graphs[i] < selected.size()));
      selected[select_graphs[i]] = true;
    }

    vector< vector<HaploModelGraph *> > g;
    g.reserve(select_graphs.size());
    for (int i = 0; i < selected.size(); ++i)
      if (selected[i])
        g.push_back(graphs.graphs[i]);

    graphs.graphs.swap(g);
  }
    
  cerr << "* Generating graph...";

  int border_x = int(double(image_size_x) / 100.0 + 0.5);
  int border_y = border_x;

  int graph_image_width = image_size_x;
  int graph_image_height = image_size_y;

  Cairo graph_cairo(graph_image_height, graph_image_width,
                    CairoOutputSurface(graph_image_height, graph_image_width, false));

  vector<uint32> pop_color(4, 0x000000);
  pop_color[0] = 0xc04040;
  pop_color[1] = 0x4040c0;
  pop_color[2] = 0xc0c040;
  pop_color[3] = 0x40c040;

  vector<uint32> float_color(2, 0x000000);
  float_color[0] = 0x404040;
  float_color[1] = 0xc0c0c0;

  double total_height = 0.0;
  for (int i = 0; i < graphs.size(); ++i)
  {
    if (i > 0)
      total_height += 6.0;
    for (int j = 0; j < graphs[i].size(); ++j)
    {
      if (j > 0)
        total_height += 1.0;
      total_height += graphs[i][j]->height;
    }
  }
  Assert(total_height > 0.0);

  double graph_scale_x = double(graph_image_width - 2*border_x) / double(graphs.pos.back() - graphs.pos.front());
  double graph_offset_x = -double(graphs.pos.front()) * graph_scale_x + border_x;
  double graph_scale_y = (graph_image_height - 2*border_y) / total_height;
  double graph_offset_y = border_y;

  if (graphs.chrom_ranges.size() > 1)
  {
    for (int chrom = 0; chrom < graphs.chrom_ranges.size(); ++chrom)
    {
      double x_left = graphs.pos[graphs.chrom_ranges[chrom].first] * graph_scale_x + graph_offset_x;
      double x_right = graphs.pos[graphs.chrom_ranges[chrom].second - 1] * graph_scale_x + graph_offset_x;
      
      if ((chrom & 1) == 0)
        cairo_set_source_rgba(graph_cairo, color_rgb(0xb0ffd0), 1.0);
      else
        cairo_set_source_rgba(graph_cairo, color_rgb(0xb0d0ff), 1.0);
      cairo_rectangle(graph_cairo, border_y, x_left, graph_image_height - 2*border_y, x_right - x_left);
      cairo_fill(graph_cairo);
    }
  }

  if (!no_grid)
  {
    cairo_set_source_rgba(graph_cairo, color_rgb(0x808080), 1.0);
    cairo_set_line_width(graph_cairo, 1.0 * line_width);
    vector<double> dashes(2);
    dashes[0] = 1.0;
    dashes[1] = 1.0;
    cairo_set_dash(graph_cairo, (const double *)&dashes[0], dashes.size(), 0.0);
    double step = 0.01;
    if ((step * graph_scale_x) < 5)
      step = 0.1;
    for (int chrom = 0; chrom < graphs.chrom_ranges.size(); ++chrom)
    {
      for (double pos = graphs.pos[graphs.chrom_ranges[chrom].first]; pos <= graphs.pos[graphs.chrom_ranges[chrom].second - 1]; pos += step)
      {
        cairo_move_to(graph_cairo, border_y, pos*graph_scale_x + graph_offset_x);
        cairo_rel_line_to(graph_cairo, graph_image_height - 2*border_y, 0.0);
        cairo_stroke(graph_cairo);
      }
    }

    cairo_set_dash(graph_cairo, (const double *)NULL, 0, 0.0);
  }

  if (!no_grid)
  {
    cairo_set_source_rgba(graph_cairo, color_rgb(0x000000), 1.0);
    cairo_set_line_width(graph_cairo, 2.0 * line_width);
    vector<double> dashes(2);
    dashes[0] = 4.0;
    dashes[1] = 4.0;
    cairo_set_dash(graph_cairo, (const double *)&dashes[0], dashes.size(), 0.0);

    for (int chrom = 0; chrom < graphs.chrom_ranges.size(); ++chrom)
    {
      for (double pos = graphs.pos[graphs.chrom_ranges[chrom].first]; pos <= graphs.pos[graphs.chrom_ranges[chrom].second - 1]; pos += 1.0)
      {
        cairo_move_to(graph_cairo, border_y, pos*graph_scale_x + graph_offset_x);
        cairo_rel_line_to(graph_cairo, graph_image_height - 2*border_y, 0.0);
        cairo_stroke(graph_cairo);
      }
    }

    cairo_set_dash(graph_cairo, (const double *)NULL, 0, 0.0);
  }

  double graph_pos = 0.0;
  for (int i = 0; i < graphs.size(); ++i)
  {
    if (i > 0)
    {
      graph_pos += 3.0;
      
      cairo_set_source_rgba(graph_cairo, color_rgb(0x000000), 1.0);
      cairo_set_line_width(graph_cairo, 3.0 * line_width);
      cairo_move_to(graph_cairo, graph_pos * graph_scale_y + graph_offset_y, 0.5 * border_x);
      cairo_rel_line_to(graph_cairo, 0.0, graph_image_width - border_x);
      cairo_stroke(graph_cairo);

      graph_pos += 3.0;
    }

    double graph_pos_top = graph_pos;

    for (int j = 0; j < graphs[i].size(); ++j)
    {
      if (j > 0)
        graph_pos += 1.0;
      
      switch (graphs[i][j]->type)
      {
      case 1:
      {
        HaploModelGraphCode &graph = *((HaploModelGraphCode *)graphs[i][j]);
        const vector<uint32> &colors = (graph.palette == 0) ? graph.colors : pop_color;
        GraphCode(graph_cairo, graph.v, graphs.pos, 
                  graph_scale_x, graph_offset_x,
                  graph_pos * graph_scale_y + graph_offset_y, 
                  (graph_pos + graphs[i][j]->height) * graph_scale_y + graph_offset_y, 
                  colors);
        break;
      }
      
      case 2:
      {
        HaploModelGraphCodeProb &graph = *((HaploModelGraphCodeProb *)graphs[i][j]);
        const vector<uint32> &colors = (graph.palette == 0) ? graph.colors : pop_color;
        GraphProportion(graph_cairo, graph.p, float(1.0), float(0.0), 
                        graphs.pos, graph_scale_x, graph_offset_x, 
                        graph_pos * graph_scale_y + graph_offset_y, 
                        (graph_pos + graphs[i][j]->height) * graph_scale_y + graph_offset_y, 
                        colors);
        break;
      }

      case 3:
      {
        HaploModelGraphSpace &graph = *((HaploModelGraphSpace *)graphs[i][j]);
        break;
      }
      
      case 4:
      {
        HaploModelGraphFloat &graph = *((HaploModelGraphFloat *)graphs[i][j]);
        const vector<uint32> &colors = (graph.palette == 0) ? graph.colors : float_color;
        GraphGraph(graph_cairo, graph.v, 
                   float(graph.max_value), float(graph.min_value),
                   graphs.pos, graph_scale_x, graph_offset_x, 
                   graph_pos * graph_scale_y + graph_offset_y, 
                   (graph_pos + graphs[i][j]->height) * graph_scale_y + graph_offset_y, 
                   colors[0], colors[1]);
        break;
      }

      default:
        AssertMsg(false, ToStr(graphs[i][j]->type));
      }
      
      graph_pos += graphs[i][j]->height;
    }

    cairo_set_source_rgba(graph_cairo, color_rgb(0x000000), 1.0);
    cairo_set_line_width(graph_cairo, 2.0 * line_width);
    cairo_move_to(graph_cairo, graph_pos_top * graph_scale_y + graph_offset_y, graphs.pos.front() * graph_scale_x + graph_offset_x);
    cairo_rel_line_to(graph_cairo, (graph_pos - graph_pos_top) * graph_scale_y, 0.0);
    cairo_stroke(graph_cairo);

    cairo_move_to(graph_cairo, graph_pos_top * graph_scale_y + graph_offset_y, graphs.pos.back() * graph_scale_x + graph_offset_x);
    cairo_rel_line_to(graph_cairo, (graph_pos - graph_pos_top) * graph_scale_y, 0.0);
    cairo_stroke(graph_cairo);
  }

  cerr << " done." << endl;

  cerr << "* Writing image...";
  Cairo cairo(image_size_x, image_size_y,
              CairoOutputPNG(RelPath(output_dir, output_file), 
                             image_size_x, image_size_y, false));
  cairo_set_source_rgba(cairo, color_rgb(0xffffff), 1.0);
  cairo_rectangle(cairo, 0, 0, image_size_x, image_size_y);
  cairo_fill(cairo);

  cairo_pattern_t *pattern = cairo_pattern_create_for_surface(graph_cairo);
  cairo_pattern_set_filter(pattern, CAIRO_FILTER_NEAREST);
  cairo_pattern_set_extend(pattern, CAIRO_EXTEND_REPEAT);
  cairo_matrix_t transform;
  transform.xx = transform.yy = 0.0;
  transform.xy = transform.yx = 1.0;
  transform.y0 = 0.0;//-border_x;
  transform.x0 = 0.0;//-border_y;
  cairo_pattern_set_matrix(pattern, &transform);
  //cairo_rectangle(cairo, border_x, border_y, image_size_x - 2*border_x, image_size_y - 2*border_y);
  cairo_rectangle(cairo, 0, 0, image_size_x, image_size_y);
  cairo_set_source(cairo, pattern);
  cairo_fill(cairo);

  cairo_pattern_destroy(pattern);
  cairo.Done();
  cerr << " done." << endl;
}


void AddSpace(HaploModelGraphs &graphs, double height = 1.0)
{
  for (int i = 0; i < graphs.graphs.size(); ++i)
    graphs.graphs[i].push_back(new HaploModelGraphSpace(height));
}

void Graph()
{
  string snpprofile_file;
  for (int i = 0; i < input_files.size(); ++i)
    if ((GetExtension(RemoveCompressedExtension(input_files[i])) == "bin") &&
        ((GetExtension(RemoveExtension(RemoveCompressedExtension(input_files[i]))) == "snpprofile") ||
         (GetExtension(RemoveExtension(RemoveCompressedExtension(input_files[i]))) == "snpprofiledata")))
      snpprofile_file = input_files[i];

  Assert(!snpprofile_file.empty());
  SNPProfile profile;
  {
    istream *in = InFileStream(RelPath(input_dir, snpprofile_file));

    if (!in)
    {
      cerr << "Could not open file \"" << snpprofile_file << "\"";
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"";
      cerr << endl;
      Exit(1);
    }

    cerr << "* Reading SNPProfile...";
    profile.ReadBin(*in);
    cerr << " done." << endl;
    delete in;

    profile.Validate();
  }

  cerr << "* Reading graphs...";
  HaploModelGraphs graphs(profile, true);
  
  for (int i = 0; i < input_files.size(); ++i)
  {
    if (GetExtension(RemoveCompressedExtension(input_files[i])) == "bin")
    {
      if ((GetExtension(RemoveExtension(RemoveCompressedExtension(input_files[i]))) == "snpprofile") || (GetExtension(RemoveExtension(RemoveCompressedExtension(input_files[i]))) == "snpprofiledata"))
        continue;

      if (GetExtension(RemoveExtension(RemoveCompressedExtension(input_files[i]))) == "haplopopdata")
      {
        vector< vector<int8> > pop_data;
        istream *in = InFileStream(RelPath(input_dir, input_files[i]));
        AssertMsg(in, RelPath(input_dir, input_files[i]));
        ReadBin(*in, pop_data);
        delete in;

        if (graphs.graphs.empty())
          graphs.graphs.resize(pop_data.size());
        else
        {
          if (!ignore_missing)
            Assert(graphs.graphs.size() == pop_data.size());
          else
            graphs.graphs.resize(min(graphs.graphs.size(), pop_data.size()));
          AddSpace(graphs);
        }
        for (int i = 0; i < graphs.graphs.size(); ++i)
        {
          graphs.graphs[i].push_back(new HaploModelGraphCode(pop_data[i]));
          cerr << ".";
        }
        cerr << "/";
      }
      else if (GetExtension(RemoveExtension(RemoveCompressedExtension(input_files[i]))) == "diplopopdata")
      {
        vector< vector<int8> > pop_data;
        istream *in = InFileStream(RelPath(input_dir, input_files[i]));
        AssertMsg(in, RelPath(input_dir, input_files[i]));
        ReadBin(*in, pop_data);
        delete in;

        Assert((pop_data.size() % 2) == 0);
        if (graphs.graphs.empty())
          graphs.graphs.resize(pop_data.size() / 2);
        else
        {
          if (!ignore_missing)
            Assert(graphs.graphs.size() == (pop_data.size() / 2));
          else
            graphs.graphs.resize(min(graphs.graphs.size(), pop_data.size() / 2));
          AddSpace(graphs);
        }
        for (int i = 0; i < graphs.graphs.size(); ++i)
        {
          graphs.graphs[i].push_back(new HaploModelGraphCode(pop_data[i*2 + 0]));
          graphs.graphs[i].push_back(new HaploModelGraphCode(pop_data[i*2 + 1]));
          cerr << ".";
        }
        cerr << "/";
      }
      else if (GetExtension(RemoveExtension(RemoveCompressedExtension(input_files[i]))) == "haploposterior")
      {
        vector< vector< vector<float> > > pop_ps;
        istream *in = InFileStream(RelPath(input_dir, input_files[i]));
        AssertMsg(in, RelPath(input_dir, input_files[i]));
        ReadBin(*in, pop_ps);
        delete in;

        if (graphs.graphs.empty())
          graphs.graphs.resize(pop_ps.size());
        else
        {
          if (!ignore_missing)
            Assert(graphs.graphs.size() == pop_ps.size());
          else
            graphs.graphs.resize(min(graphs.graphs.size(), pop_ps.size()));
          AddSpace(graphs);
        }
        for (int i = 0; i < graphs.graphs.size(); ++i)
        {
          graphs.graphs[i].push_back(new HaploModelGraphCodeProb(pop_ps[i]));
          cerr << ".";
        }
        cerr << "/";
      }
      else if (GetExtension(RemoveExtension(RemoveCompressedExtension(input_files[i]))) == "diploposterior")
      {
        vector< vector< vector<float> > > pop_ps;
        istream *in = InFileStream(RelPath(input_dir, input_files[i]));
        AssertMsg(in, RelPath(input_dir, input_files[i]));
        ReadBin(*in, pop_ps);
        delete in;

        Assert((pop_ps.size() % 2) == 0);
        if (graphs.graphs.empty())
          graphs.graphs.resize(pop_ps.size() / 2);
        else
        {
          if (!ignore_missing)
            Assert(graphs.graphs.size() == (pop_ps.size() / 2));
          else
            graphs.graphs.resize(min(graphs.graphs.size(), pop_ps.size() / 2));
          AddSpace(graphs);
        }
        for (int i = 0; i < graphs.graphs.size(); ++i)
        {
          cerr << ".";
          graphs.graphs[i].push_back(new HaploModelGraphCodeProb(pop_ps[i*2 + 0]));
          graphs.graphs[i].push_back(new HaploModelGraphCodeProb(pop_ps[i*2 + 1]));
        }
        cerr << "/";
      }
      else if (GetExtension(RemoveExtension(RemoveCompressedExtension(input_files[i]))) == "phasingposterior")
      {
        vector< vector<float> > phasing_errors;
        istream *in = InFileStream(RelPath(input_dir, input_files[i]));
        AssertMsg(in, RelPath(input_dir, input_files[i]));
        ReadBin(*in, phasing_errors);
        delete in;

        if (graphs.graphs.empty())
          graphs.graphs.resize(phasing_errors.size());
        else
        {
          if (!ignore_missing)
            Assert(graphs.graphs.size() == phasing_errors.size());
          else
            graphs.graphs.resize(min(graphs.graphs.size(), phasing_errors.size()));
          AddSpace(graphs);
        }
        for (int i = 0; i < graphs.graphs.size(); ++i)
        {
          graphs.graphs[i].push_back(new HaploModelGraphFloat(phasing_errors[i]));
          cerr << ".";
        }
        cerr << "/";
      }
      else
        Assert(false);
    }
    else if (GetExtension(RemoveCompressedExtension(input_files[i])) == "lst")
    {
      if (GetExtension(RemoveExtension(RemoveCompressedExtension(input_files[i]))) == "saberviterbi")
      {
        vector< vector<int8> > pop_data;

        string path, dummy;
        SplitPath(RelPath(input_dir, input_files[i]), path, dummy);
        vector<string> files;
        {
          istream *in = InFileStream(RelPath(input_dir, input_files[i]));
          AssertMsg(in, RelPath(input_dir, input_files[i]));
          string line;
          while (getline(*in, line))
            files.push_back(line);
          delete in;
        }
        pop_data.resize(files.size() * 2);
        for (int file = 0; file < files.size(); ++file)
        {
          if (files[file].empty())
          {
            pop_data[file*2 + 0].resize(profile.seq_len, 0);
            pop_data[file*2 + 1].resize(profile.seq_len, 0);
            continue;
          }
          else
          {
            pop_data[file*2 + 0].resize(profile.seq_len, -1);
            pop_data[file*2 + 1].resize(profile.seq_len, -1);
          }
          
          istream *in = InFileStream(RelPath(path, files[file]));
          AssertMsg(in, files[file]);
          string line;
          vector<string> fields;
          int i = 0;
          while (getline(*in, line))
          {
            SplitFields(line, fields, " \t,;");
            if (fields.empty())
              continue;
            Assert(i < profile.seq_len);
            Assert(fields.size() == 2);
            int pop0, pop1;
            Assert(StringToInt(fields[0], pop0));
            Assert(StringToInt(fields[1], pop1));
            pop_data[file*2 + 0][i] = pop0 - 1;
            pop_data[file*2 + 1][i] = pop1 - 1;
            ++i;
          }
          delete in;
          Assert(i == profile.seq_len);
          cerr << ".";
        }

        Assert((pop_data.size() % 2) == 0);
        if (graphs.graphs.empty())
          graphs.graphs.resize(pop_data.size() / 2);
        else
        {
          if (!ignore_missing)
            Assert(graphs.graphs.size() == (pop_data.size() / 2));
          else
            graphs.graphs.resize(min(graphs.graphs.size(), pop_data.size() / 2));
          AddSpace(graphs);
        }
        for (int i = 0; i < graphs.graphs.size(); ++i)
        {
          graphs.graphs[i].push_back(new HaploModelGraphCode(pop_data[i*2 + 0]));
          graphs.graphs[i].push_back(new HaploModelGraphCode(pop_data[i*2 + 1]));
        }
        cerr << "/";
      }
      else if (GetExtension(RemoveExtension(RemoveCompressedExtension(input_files[i]))) == "saberposterior")
      {
        vector< vector< vector<float> > > pop_ps;

        string path, dummy;
        SplitPath(RelPath(input_dir, input_files[i]), path, dummy);
        vector<string> files;
        {
          istream *in = InFileStream(RelPath(input_dir, input_files[i]));
          AssertMsg(in, RelPath(input_dir, input_files[i]));
          string line;
          while (getline(*in, line))
            files.push_back(line);
          delete in;
        }
        pop_ps.resize(files.size());

        for (int file = 0; file < files.size(); ++file)
        {
          if (ignore_missing && files[file].empty())
          {
            pop_ps[file].resize(3, vector<float>(profile.seq_len, 0.0));
            continue;
          }

          istream *in = InFileStream(RelPath(path, files[file]));
          if (ignore_missing && !in)
          {
            pop_ps[file].resize(3, vector<float>(profile.seq_len, 0.0));
            continue;
          }
          AssertMsg(in, RelPath(path, files[file]));

          pop_ps[file].reserve(52);
          
          string line;
          vector<string> fields;
          vector<double> pop_p;
          int i = 0;
          while (getline(*in, line))
          {
            SplitFields(line, fields, " \t,;");
            if (fields.empty())
              continue;
            Assert(i < profile.seq_len);
            if (i == 0)
            {
              Assert(pop_ps[file].empty());
              pop_ps[file].resize(fields.size(), vector<float>(profile.seq_len, 0.0));
            }
            for (int pop = 0; pop < fields.size(); ++pop)
            {
              StrTo(fields[pop], pop_ps[file][pop][i]);
              Assert((pop_ps[file][pop][i] >= 0.0) && (pop_ps[file][pop][i] <= 1.0));
            }
            ++i;
          }
          delete in;
          Assert(i == profile.seq_len);
          cerr << ".";
        }

        if (graphs.graphs.empty())
          graphs.graphs.resize(pop_ps.size());
        else
        {
          if (!ignore_missing)
            Assert(graphs.graphs.size() == pop_ps.size());
          else
            graphs.graphs.resize(min(graphs.graphs.size(), pop_ps.size()));
          AddSpace(graphs);
        }
        for (int i = 0; i < graphs.graphs.size(); ++i)
        {
          graphs.graphs[i].push_back(new HaploModelGraphCodeProb(pop_ps[i]));
          graphs.graphs[i].back()->height = 8.0;
        }
        cerr << "/";
      }
      else
        Assert(false);
    }
    else
      Assert(false);
  }

  cerr << " done." << endl;

  GraphHaploModelGraphs(graphs);
}


int Main(vector<string> args)
{
  cerr << "GraphHaploModel" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;
  
  if ((input_files.size() == 1) && 
      (GetExtension(RemoveCompressedExtension(input_files[0])) == "bin") &&
      (GetExtension(RemoveExtension(RemoveCompressedExtension(input_files[0]))) == "haplomodelgraphs"))
  {
    HaploModelGraphs graphs;
    {
      cerr << "* Reading haplo model graphs...";
      istream *in = InFileStream(RelPath(input_dir, input_files[0]));
      AssertMsg(in, RelPath(input_dir, input_files[0]));
      graphs.ReadBin(*in);
      delete in;
      cerr << " done." << endl;
    }

    GraphHaploModelGraphs(graphs);
  }
  else
    Graph();

  cerr << "Done!" << endl;
  return 0;
}
