
#include "System.h"
#include "Utility.h"
#include "HaploModel.h"


namespace HaploModel
{

  template <typename T>
  void StateProfile<T>::Normalize(double &scale)
  {
    scale = 0.0;
    int i = 0;
    while (i < s.size())
    {
      for (int k = 0; k < pops_offset; ++k)
        scale += s[i + k];
      i += count_size;
    }

    if (scale == 0.0)
      scale = 1.0;
    //
    //scale = 1.0;
    //
    double rescale = 1.0 / scale;
    i = 0;
    while (i < s.size())
    {
      for (int k = 0; k < pops_offset; ++k)
        s[i + k] = T(s[i + k] * rescale);
      i += count_size;
    }
  }


  void Model::Init(int num_pops_)
  {
    num_pops = num_pops_;
    
    emission.clear();
    emission.resize(num_pops, vector< vector<double> > (5, vector<double>(5, 0.0)));
    pop_init.clear();
    pop_init.resize(num_pops, 0.0);
    pop_recomb.clear();
    pop_recomb.resize(num_pops, 0.0);
    pop_admix.clear();
    pop_admix.resize(num_pops, vector<double>(num_pops, 0.0));
  }

  void Model::DisplayEmission(ostream &out) const
  {
    for (int pop = 0; pop < num_pops; ++pop)
    {
      if (pop == 0)
        out << StrFitRight("emission", 10, ' ');
      else
        out << StrFitRight("", 10, ' ');
      out << StrFitRight(pop, 8, ' ') << StrFitRight("y\\x", 8, '.');
      for (int x = 0; x < 5; ++x)
        out << StrFitRight((x < 4) ? base2char[x] : '?', 8, '-');
      out << "\n";
      
      for (int y = 0; y < 5; ++y)
      {
        out << StrFitRight("", 18, ' ') << StrFitRight((y < 4) ? base2char[y] : '?', 8, ' ');
        for (int x = 0; x < 5; ++x)
          out << ToStr(emission[pop][y][x], 8, 5);
        out << "\n";
      }
    }
  }

  void Model::DisplayPopInit(ostream &out) const
  {
    for (int pop = 0; pop < num_pops; ++pop)
      out << StrFitRight((pop == 0) ? "pop_init" : "", 10, ' ') << StrFitRight(pop, 8, (pop == 0) ? '.' : ' ') << ToStr(pop_init[pop], 16, 5) << "\n";
  }

  void Model::DisplayPopRecomb(ostream &out) const
  {
    for (int pop = 0; pop < num_pops; ++pop)
      out << StrFitRight((pop == 0) ? "pop_recomb" : "", 10, ' ') << StrFitRight(pop, 8, (pop == 0) ? '.' : ' ') << ToStr(pop_recomb[pop], 16, 5) << "\n";
  }

  void Model::DisplayPopAdmix(ostream &out) const
  {
    out << StrFitRight("pop_admix", 10, ' ') << StrFitRight("fr\\to", 8, '.');
    for (int q = 0; q < num_pops; ++q)
      out << StrFitRight(q, 8, '-');
    out << "\n";
  
    for (int p = 0; p < num_pops; ++p)
    {
      out << StrFitRight("", 10, ' ') << StrFitRight(p, 8, ' ');
      for (int q = 0; q < num_pops; ++q)
        out << ToStr(pop_admix[p][q], 8, 5);
      out << "\n";
    }
  }

  void Model::SetDefaultEmission(double error_rate, double missing_rate)
  {
    for (int pop = 0; pop < num_pops; ++pop)
    {
      for (int y = 0; y < 5; ++y)
        for (int x = 0; x < 5; ++x)
        {
          if (y == 4)
            emission[pop][y][x] = 0.2;
          else if (x == 4)
            emission[pop][y][x] = missing_rate;
          else if (x == y)
            emission[pop][y][x] = (1.0 - missing_rate) * (1.0 - error_rate);
          else
            emission[pop][y][x] = (1.0 - missing_rate) * error_rate / 3.0;
        }
    }
  }
  
  void Model::SetDefaultPopInit()
  {
    for (int pop = 0; pop < num_pops; ++pop)
      pop_init[pop] = 1.0 / double(num_pops);
  }

  void Model::SetDefaultPopRecomb(double rate)
  {
    for (int pop = 0; pop < num_pops; ++pop)
      pop_recomb[pop] = rate;
  }

  void Model::SetDefaultPopAdmix(double self_p)
  {
    if (num_pops == 1)
      pop_admix[0][0] = 1.0;
    else
    {
      for (int p = 0; p < num_pops; ++p)
        for (int q = 0; q < num_pops; ++q)
        {
          if (p == q)
            pop_admix[p][q] = self_p;
          else
            pop_admix[p][q] = (1.0 - self_p) / double(num_pops - 1);
        }
    }
  }

  void Model::SetDefault()
  {
    SetDefaultEmission();
    SetDefaultPopInit();
    SetDefaultPopRecomb();
    SetDefaultPopAdmix();
  }

  void Model::Generate(const SNPData &data,
                       vector<int8> &output, vector<int> &haplos)
  {
    haplos.resize(data.seq_len);
    output.resize(data.seq_len);
  
    double prob = drand();
    int pop = 0;
    while (prob >= pop_init[pop])
      prob -= pop_init[pop++];
    Assert(pop < num_pops);
  
    prob = drand() * double(data.pops[pop].haplos.size());
    int index = int(floor(prob));
    int k = data.pops[pop].haplos[index];
  
    for (int i = 0; i < data.seq_len; ++i)
    {
      haplos[i] = k;
      int pop = data.haplos[k].pop;
    
      int y = data.haplos[k][i];
      prob = drand();
      int x = 0;
      while (prob >= emission[pop][y][x])
        prob -= emission[pop][y][x++];
      output[i] = int8(x);
    
      double recomb = exp(-pop_recomb[pop] * (data.locus_genetic_pos[i + 1] - data.locus_genetic_pos[i]));
      prob = drand();
      if (prob < recomb)
      {
        if (data.haplos[k].diplo >= 0)
        {
          if (drand() < SNPData::switch_decode(data.diplos[data.haplos[k].diplo].switches[i]))
            k = data.haplos[k].diplo_pair;
        }
      }
      else
      {
        prob = drand();
        int new_pop = 0;
        while (prob >= pop_admix[pop][new_pop])
          prob -= pop_admix[pop][new_pop++];
        pop = new_pop;
      
        prob = drand() * double(data.pops[pop].haplos.size());
        index = int(floor(prob));
        k = data.pops[pop].haplos[index];
      }
    }
  }


  void Model::Write(ostream &o) const
  {
    int32 w_num_pops = num_pops;
    ::Write(o, w_num_pops);

    for (int y = 0; y < 5; ++y)
      for (int x = 0; x < 5; ++x)
        ::Write(o, emission[0][y][x]);

    for (int pop = 0; pop < num_pops; ++pop)
      ::Write(o, pop_init[pop]);

    for (int pop = 0; pop < num_pops; ++pop)
      ::Write(o, pop_recomb[pop]);

    for (int p = 0; p < num_pops; ++p)
      for (int q = 0; q < num_pops; ++q)
        ::Write(o, pop_admix[p][q]);
  }

  void Model::Read(istream &i)
  {
    int32 r_num_pops;
    ::Read(i, r_num_pops);

    Init(r_num_pops);
    
    emission.resize(1);
    for (int y = 0; y < 5; ++y)
      for (int x = 0; x < 5; ++x)
        ::Read(i, emission[0][y][x]);
    emission.resize(num_pops, emission[0]);

    for (int pop = 0; pop < num_pops; ++pop)
      ::Read(i, pop_init[pop]);

    for (int pop = 0; pop < num_pops; ++pop)
      ::Read(i, pop_recomb[pop]);

    for (int p = 0; p < num_pops; ++p)
      for (int q = 0; q < num_pops; ++q)
        ::Read(i, pop_admix[p][q]);
  }


  void Model::Write2(ostream &o) const
  {
    int32 w_num_pops = num_pops;
    ::Write(o, w_num_pops);

    for (int pop = 0; pop < num_pops; ++pop)
      for (int y = 0; y < 5; ++y)
        for (int x = 0; x < 5; ++x)
          ::Write(o, emission[pop][y][x]);

    for (int pop = 0; pop < num_pops; ++pop)
      ::Write(o, pop_init[pop]);

    for (int pop = 0; pop < num_pops; ++pop)
      ::Write(o, pop_recomb[pop]);

    for (int p = 0; p < num_pops; ++p)
      for (int q = 0; q < num_pops; ++q)
        ::Write(o, pop_admix[p][q]);
  }

  void Model::Read2(istream &i)
  {
    int32 r_num_pops;
    ::Read(i, r_num_pops);

    Init(r_num_pops);
    
    for (int pop = 0; pop < num_pops; ++pop)
      for (int y = 0; y < 5; ++y)
        for (int x = 0; x < 5; ++x)
          ::Read(i, emission[pop][y][x]);

    for (int pop = 0; pop < num_pops; ++pop)
      ::Read(i, pop_init[pop]);

    for (int pop = 0; pop < num_pops; ++pop)
      ::Read(i, pop_recomb[pop]);

    for (int p = 0; p < num_pops; ++p)
      for (int q = 0; q < num_pops; ++q)
        ::Read(i, pop_admix[p][q]);
  }


  void GetHaploPops(const SNPData &data, 
                    const vector<int> &haplos, vector<int8> &pops)
  {
    pops.resize(haplos.size());
    for (int i = 0; i < haplos.size(); ++i)
      pops[i] = data.haplos[haplos[i]].pop;
  }

  void GetPopMask(const SNPData &data, const vector<int8> &pops, 
                  vector<uint64> &pop_mask)
  {
    pop_mask.resize(pops.size());
    for (int i = 0; i < pops.size(); ++i)
      pop_mask[i] = uint64(1) << pops[i];
  }


  void ComputeInitEdges(const SNPData &data, const Model &model,
                        vector<double> &init_edges, uint64 pop_mask,
                        const vector<int> &haplo_mask,
                        const vector<int> &haplo_exclude)
  {
    vector<int> pop_haplo_count(data.num_pops, 0);
    for (int k = 0; k < data.num_haplos; ++k)
      if ((haplo_mask.empty() || haplo_mask[k]) &&
          (haplo_exclude.empty() || (haplo_exclude[0] != k)))
        ++pop_haplo_count[data.haplos[k].pop];

    vector<double> pop_haplo_scale(data.num_pops, 0.0);
    for (int pop = 0; pop < data.num_pops; ++pop)
    {
      if ((!pop_mask || (pop_mask & (uint64(1) << pop))) &&
          (pop_haplo_count[pop] > 0))
        pop_haplo_scale[pop] = 1.0 / double(pop_haplo_count[pop]);
      else
        pop_haplo_scale[pop] = 0.0;
    }

    double pop_total = 0.0;
    vector<double> pop_haplo_init = model.pop_init;
    for (int pop = 0; pop < data.num_pops; ++pop)
      if ((!pop_mask || (pop_mask & (uint64(1) << pop))) &&
          (pop_haplo_count[pop] > 0))
        pop_total += model.pop_init[pop];
    AssertMsg(pop_total > 0.0, ToStr(pop_total) + ", " + ToBinary(pop_mask));

    for (int pop = 0; pop < data.num_pops; ++pop)
    {
      if ((!pop_mask || (pop_mask & (uint64(1) << pop))) &&
          (pop_haplo_count[pop] > 0))
        pop_haplo_init[pop] = (model.pop_init[pop] / pop_total) * pop_haplo_scale[pop];
      else
        pop_haplo_init[pop] = 0.0;
    }

    init_edges.resize(data.num_haplos, 0.0);
    for (int k = 0; k < data.num_haplos; ++k)
      if ((haplo_mask.empty() || haplo_mask[k]) &&
          (haplo_exclude.empty() || (haplo_exclude[0] != k)))
        init_edges[k] = pop_haplo_init[data.haplos[k].pop];
  }


  void Forward::Calculate(const SNPData &data, const Model &model,
                          const vector<int8> &x,
                          const vector<uint64> &pop_mask_from,
                          const vector<uint64> &pop_mask_to,
                          const vector<int> &haplo_mask,
                          const vector<int> &haplo_exclude,
                          int max_count_)
  {
    Assert(x.size() >= data.seq_len);
    Assert(pop_mask_from.empty() || (pop_mask_from.size() >= data.seq_len));
    Assert(pop_mask_to.empty() || (pop_mask_to.size() >= data.seq_len));
    Assert((pop_mask_from.empty() && pop_mask_to.empty()) ||
           (!pop_mask_from.empty() && !pop_mask_to.empty()));
    bool pop_mask = !pop_mask_from.empty();
    Assert(haplo_mask.empty() || (haplo_mask.size() == data.num_haplos));
    Assert(haplo_exclude.empty() || (haplo_exclude.size() >= data.seq_len));

    max_count = max_count_;
    int num_c = max(max_count, 1);
    s.clear();
    s.resize(data.seq_len, StateProfile<Prob>(data, num_c));
    scale.clear();
    scale.resize(data.seq_len, 1.0);
    log_scale_total.clear();
    log_scale_total.resize(data.seq_len);

    StateProfile<double> last_s(data, num_c);
    StateProfile<double> new_s(data, num_c);
  
    for (int c = 0; c < num_c; ++c)
      for (int p = 0; p < data.num_pops; ++p)
      {
        new_s.Z(0, p, c) = 0.0;
        new_s.Z(1, p, c) = 0.0;
      }

    vector<double> init_edges;
    ComputeInitEdges(data, model, init_edges,
                     !pop_mask ? uint64(0) : (pop_mask_from[0]|pop_mask_to[0]),
                     haplo_mask, haplo_exclude);
    init_edges.clear();
    init_edges.resize(data.num_haplos, 1.0);
    for (int k = 0; k < data.num_haplos; ++k)
    {
      new_s.O(k, 0) = init_edges[k] * model.emission[data.haplos[k].pop][data.haplos[k][0]][x[0]];
      for (int c = 1; c < num_c; ++c)
        new_s.O(k, c) = 0.0;
    }
  
    new_s.Normalize(scale[0]);
    s[0] = new_s;

    vector<double> no_recomb_prob(data.num_pops);
    vector<int> pop_haplo_count(data.num_pops, 0);
    vector<double> pop_haplo_scale(data.num_pops, 0.0);

    for (int i = 1; i < data.seq_len; ++i)
    {
      last_s.swap(new_s);
      new_s.clear();

      pop_haplo_count.clear();
      pop_haplo_count.resize(data.num_pops, 0);
      for (int k = 0; k < data.num_haplos; ++k)
        if ((haplo_mask.empty() || haplo_mask[k]) &&
            (haplo_exclude.empty() || (haplo_exclude[i] != k)))
          ++pop_haplo_count[data.haplos[k].pop];
      
      for (int pop = 0; pop < data.num_pops; ++pop)
      {
        if ((!pop_mask || 
             ((pop_mask_from[i] | pop_mask_to[i]) & (uint64(1) << pop))) &&
            (pop_haplo_count[pop] > 0))
          pop_haplo_scale[pop] = 1.0 / double(pop_haplo_count[pop]);
        else
          pop_haplo_scale[pop] = 0.0;
      }

      for (int pop = 0; pop < data.num_pops; ++pop)
        no_recomb_prob[pop] = exp(-model.pop_recomb[pop] * (data.locus_genetic_pos[i] - data.locus_genetic_pos[i - 1]));
      
      for (int k = 0; k < data.num_haplos; ++k)
        if ((haplo_mask.empty() || (haplo_mask[k])) &&
            (haplo_exclude.empty() || (haplo_exclude[i - 1] != k)))
        {
          int pop = data.haplos[k].pop;
          for (int c = 0; c < num_c; ++c)
            new_s.Z(0, pop, c) += last_s.O(k, c) * (1.0 - no_recomb_prob[pop]);
        }

      for (int c = 0; c < num_c; ++c)
      {
        for (int p = 0; p < data.num_pops; ++p)
        {
          bool from = !pop_mask || (pop_mask_from[i - 1] & (uint64(1) << p));
          double total = 0.0;
          for (int q = 0; q < data.num_pops; ++q)
            if ((p == q) || 
                (((max_count == 0) || (c < (num_c - 1))) &&
                 (!pop_mask || (from && (pop_mask_to[i] & (uint64(1) << q))))))
              total += model.pop_admix[p][q];

          if (total > 0.0)
          {
            double scale = 1.0 / total;
            for (int q = 0; q < data.num_pops; ++q)
              if ((p == q) || 
                  (((max_count == 0) || (c < (num_c - 1))) &&
                   (!pop_mask || (from && (pop_mask_to[i] & (uint64(1) << q))))))
                new_s.Z(1, q, c) += new_s.Z(0, p, (max_count == 0) ? 0 : ((p == q) ? c : (c - 1))) * model.pop_admix[p][q] * scale;
          }
        }
      }

      for (int k = 0; k < data.num_haplos; ++k)
      {
        int pop = data.haplos[k].pop;
        if ((!pop_mask ||
             ((pop_mask_from[i] | pop_mask_to[i]) & (uint64(1) << pop))) &&
            (haplo_mask.empty() || haplo_mask[k]) &&
            (haplo_exclude.empty() || (haplo_exclude[i] != k)))
        {
          int pair_k = data.haplos[k].diplo_pair;
          if ((pair_k >= 0) && (haplo_mask.empty() || haplo_mask[pair_k]) &&
              (haplo_exclude.empty() || (haplo_exclude[i - 1] != pair_k)))
          {
            double switch_prob = SNPData::switch_decode(data.diplos[data.haplos[k].diplo].switches[i - 1]);
            for (int c = 0; c < num_c; ++c)
              new_s.O(k, c) = (last_s.O(k, c) * (1.0 - switch_prob) + last_s.O(pair_k, c) * switch_prob) * no_recomb_prob[pop];
          }
          else
          {
            for (int c = 0; c < num_c; ++c)
              new_s.O(k, c) = last_s.O(k, c) * no_recomb_prob[pop];
          }

          for (int c = 0; c < num_c; ++c)
          {
            new_s.O(k, c) += new_s.Z(1, pop, c) * pop_haplo_scale[pop];
            new_s.O(k, c) *= model.emission[pop][data.haplos[k][i]][x[i]];
          }
        }
      }

      {
        total_s = 0.0;
        for (int c = 0; c < num_c; ++c)
          for (int k = 0; k < data.num_haplos; ++k)
            total_s += new_s.O(k, c);

        AssertMsg(total_s != 0.0, ToStr(i) + ", " + ToStr(int(x[i])) + (pop_mask ? (", " + ToStr(pop_mask_from[i]) + ", " + ToStr(pop_mask_to[i]) + ", " + ToStr(pop_mask_from[i - 1]) + ", " + ToStr(pop_mask_to[i - 1])) : ""));
      }

      new_s.Normalize(scale[i]);
      s[i] = new_s;
    }

    double log_total = 0.0;
    for (int i = 0; i < data.seq_len; ++i)
      log_scale_total[i] = log_total += log(scale[i]);

#if 0
    cerr << endl;
    for (int i = data.seq_len - 5; i <= data.seq_len; ++i)
      cerr << ToStr(scale[i], 20, 12);
    cerr << endl;
    for (int i = data.seq_len - 5; i <= data.seq_len; ++i)
      cerr << ToStr(log_scale_total[i], 20, 12);
    cerr << endl;
    for (int k = lower_limit; k <= upper_limit; ++k)
    {
      cerr << StrFitRight(k, 3, ' ') << ": ";
      for (int i = 0; i < data.seq_len; ++i)
        cerr << int(data.haplos[k][i]);
      cerr << endl;
    }
      cerr << StrFitRight("x", 3, ' ') << ": ";
      for (int i = 0; i < data.seq_len; ++i)
        cerr << int(x[i]);
      cerr << endl;
#endif

    total_s = 0.0;
    for (int c = 0; c < num_c; ++c)
      for (int k = 0; k < data.num_haplos; ++k)
        total_s += new_s.O(k, c);

    Assert(total_s != 0.0);
    log_partition = log(total_s) + log_scale_total[data.seq_len - 1];
  }


  void Backward::Calculate(const SNPData &data, const Model &model,
                           const vector<int8> &x,
                           const vector<uint64> &pop_mask_from,
                           const vector<uint64> &pop_mask_to,
                           const vector<int> &haplo_mask,
                           const vector<int> &haplo_exclude,
                           int max_count_)
  {
    Assert(x.size() >= data.seq_len);
    Assert(pop_mask_from.empty() || (pop_mask_from.size() >= data.seq_len));
    Assert(pop_mask_to.empty() || (pop_mask_to.size() >= data.seq_len));
    Assert((pop_mask_from.empty() && pop_mask_to.empty()) ||
           (!pop_mask_from.empty() && !pop_mask_to.empty()));
    bool pop_mask = !pop_mask_from.empty();
    Assert(haplo_mask.empty() || (haplo_mask.size() == data.num_haplos));
    Assert(haplo_exclude.empty() || (haplo_exclude.size() >= data.seq_len));

    max_count = max_count_;
    int num_c = max(max_count, 1);
    s.clear();
    s.resize(data.seq_len, StateProfile<Prob>(data, num_c));
    scale.clear();
    scale.resize(data.seq_len, 1.0);
    log_scale_total.clear();
    log_scale_total.resize(data.seq_len);
  
    StateProfile<double> last_s(data, num_c);
    StateProfile<double> new_s(data, num_c);
  
    for (int c = 0; c < num_c; ++c)
      for (int k = 0; k < data.num_haplos; ++k)
        new_s.O(k, c) = 1.0;
    new_s.Normalize(scale[data.seq_len - 1]);
    
    vector<double> init_edges;
    ComputeInitEdges(data, model, init_edges,
                     !pop_mask ? uint64(0) : (pop_mask_from[0]|pop_mask_to[0]),
                     haplo_mask, haplo_exclude);
    //init_edges.clear();
    //init_edges.resize(data.num_haplos, 1.0);
    vector<double> no_recomb_prob(data.num_pops);
    vector<int> pop_haplo_count(data.num_pops, 0);
    vector<double> pop_haplo_scale(data.num_pops, 0.0);

    for (int i = data.seq_len - 2; i >= 0; --i)
    {
      pop_haplo_count.clear();
      pop_haplo_count.resize(data.num_pops, 0);
      for (int k = 0; k < data.num_haplos; ++k)
        if ((haplo_mask.empty() || haplo_mask[k]) &&
            (haplo_exclude.empty() || (haplo_exclude[i + 1] != k)))
          ++pop_haplo_count[data.haplos[k].pop];
      
      for (int pop = 0; pop < data.num_pops; ++pop)
      {
        if ((!pop_mask || 
             ((pop_mask_from[i+1] | pop_mask_to[i+1]) & (uint64(1) << pop))) &&
            (pop_haplo_count[pop] > 0))
          pop_haplo_scale[pop] = 1.0 / double(pop_haplo_count[pop]);
        else
          pop_haplo_scale[pop] = 0.0;
      }

      for (int pop = 0; pop < data.num_pops; ++pop)
        no_recomb_prob[pop] = exp(-model.pop_recomb[pop] * (data.locus_genetic_pos[i + 1] - data.locus_genetic_pos[i]));
      
      for (int k = 0; k < data.num_haplos; ++k)
      {
        int pop = data.haplos[k].pop;
        if ((!pop_mask || 
             ((pop_mask_from[i+1] | pop_mask_to[i+1]) & (uint64(1) << pop))) &&
            (haplo_mask.empty() || haplo_mask[k]) &&
            (haplo_exclude.empty() || (haplo_exclude[i + 1] != k)))
        {
          for (int c = 0; c < num_c; ++c)
            new_s.Z(1, pop, c) += pop_haplo_scale[pop] * model.emission[pop][data.haplos[k][i + 1]][x[i + 1]] * new_s.O(k, c);
        }
      }

      for (int c = 0; c < num_c; ++c)
      {
        for (int p = 0; p < data.num_pops; ++p)
        {
          bool from = !pop_mask || (pop_mask_from[i] & (uint64(1) << p));
          double total = 0.0;
          for (int q = 0; q < data.num_pops; ++q)
            if ((p == q) || 
                (((max_count == 0) || (c < (num_c - 1))) &&
                 (!pop_mask || (from && (pop_mask_to[i + 1] & (uint64(1) << q))))))
            total += model.pop_admix[p][q];

          if (total > 0.0)
          {
            double scale = 1.0 / total;
            for (int q = 0; q < data.num_pops; ++q)
              if ((p == q) || 
                  (((max_count == 0) || (c < (num_c - 1))) &&
                   (!pop_mask || (from && (pop_mask_to[i + 1] & (uint64(1) << q))))))
                new_s.Z(0, p, (max_count == 0) ? 0 : ((p == q) ? c : (c - 1))) += model.pop_admix[p][q] * scale * new_s.Z(1, q, c);
          }
        }
      }

      s[i + 1] = new_s;
      last_s.swap(new_s);
      new_s.clear();

      for (int k = 0; k < data.num_haplos; ++k)
      {
        int pop = data.haplos[k].pop;
        if ((!pop_mask || 
             ((pop_mask_from[i] | pop_mask_to[i]) & (uint64(1) << pop))) &&
            (haplo_mask.empty() || haplo_mask[k]) &&
            (haplo_exclude.empty() || (haplo_exclude[i] != k)))
        {
          int pair_k = data.haplos[k].diplo_pair;
          if ((pair_k >= 0) && (haplo_mask.empty() || haplo_mask[pair_k]) &&
              (haplo_exclude.empty() || (haplo_exclude[i + 1] != pair_k)))
          {
            double switch_prob = SNPData::switch_decode(data.diplos[data.haplos[k].diplo].switches[i]);
            for (int c = 0; c < num_c; ++c)
              new_s.O(k, c) = no_recomb_prob[pop] * ((1.0 - switch_prob) * model.emission[pop][data.haplos[k][i + 1]][x[i + 1]] * last_s.O(k, c) + switch_prob * model.emission[pop][data.haplos[pair_k][i + 1]][x[i + 1]] * last_s.O(pair_k, c));
          }
          else
          {
            for (int c = 0; c < num_c; ++c)
              new_s.O(k, c) = no_recomb_prob[pop] * model.emission[pop][data.haplos[k][i + 1]][x[i + 1]] * last_s.O(k, c);
          }
          
          for (int c = 0; c < num_c; ++c)
            new_s.O(k, c) += (1.0 - no_recomb_prob[pop]) * last_s.Z(0, pop, c);
        }
      
      }

      {
        total_s = 0.0;
        for (int k = 0; k < data.num_haplos; ++k)
          total_s += new_s.O(k, 0);
        AssertMsg(total_s != 0.0, ToStr(i) + ", " + ToStr(int(x[i])) + (pop_mask ? (", " + ToStr(pop_mask_from[i]) + ", " + ToStr(pop_mask_to[i]) + ", " + ToStr(pop_mask_from[i + 1]) + ", " + ToStr(pop_mask_to[i + 1])) : ""));
      }

      new_s.Normalize(scale[i]);
    }
    s[0] = new_s;
    
    double log_total = 0.0;
    for (int i = data.seq_len - 1; i >= 0; --i)
      log_scale_total[i] = log_total += log(scale[i]);

#if 0
    cerr << endl;
    for (int i = data.seq_len - 50; i <= data.seq_len; ++i)
      cerr << ToStr(scale[i], 20, 12);
    cerr << endl;
    for (int i = data.seq_len - 50; i <= data.seq_len; ++i)
      cerr << ToStr(log_scale_total[i], 20, 12);
    cerr << endl;
    for (int i = data.seq_len - 50; i < data.seq_len; ++i)
      cerr << ToStr(s[i].O(0), 20, 12);
    cerr << endl;
    for (int i = data.seq_len - 50; i < data.seq_len; ++i)
      cerr << ToStr(s[i].O(1), 20, 12);
    cerr << endl;
    for (int i = 0; i < 50; ++i)
      cerr << ToStr(s[i].O(0), 20, 12);
    cerr << endl;
    for (int i = 0; i < 50; ++i)
      cerr << ToStr(s[i].O(1), 20, 12);
    cerr << endl;
    for (int k = lower_limit; k <= upper_limit; ++k)
    {
      cerr << StrFitRight(k, 3, ' ') << ": ";
      for (int i = data.seq_len - 50; i < data.seq_len; ++i)
        cerr << int(data.haplos[k][i]);
      cerr << endl;
    }
      cerr << StrFitRight("x", 3, ' ') << ": ";
      for (int i = data.seq_len - 50; i < data.seq_len; ++i)
        cerr << int(x[i]);
      cerr << endl;
#endif

    total_s = 0.0;
    for (int k = 0; k < data.num_haplos; ++k)
      total_s += new_s.O(k, 0) * init_edges[k] * model.emission[data.haplos[k].pop][data.haplos[k][0]][x[0]];

    if (total_s == 0)
    {
      for (int k = 0; k < data.num_haplos; ++k)
        cerr << k << ": " << new_s.O(k, 0) << ", " << init_edges[k] << endl;
      if (pop_mask)
      {
        cerr << "pop_mask = " << pop_mask_from[0] << ", " << pop_mask_to[0] << ", " << pop_mask_from[1] << ", " << pop_mask_to[1] << endl;
      }
    }
  
    Assert(total_s != 0.0);
    log_partition = log(total_s) + log_scale_total[0];
  }


  void Posterior::Calculate(const SNPData &data,
                            const Forward &forward, 
                            const Backward &backward)
  {
    Assert(forward.s.size() >= data.seq_len);
    Assert(backward.s.size() >= data.seq_len);
  
    int max_count = forward.max_count;
    int num_c = max(max_count, 1);
    s.clear();
    s.resize(data.seq_len, StateProfile<Prob>(data, num_c));
    pop_p.clear();
    pop_p.resize(data.num_pops, vector<Prob>(data.seq_len, 0.0));
  
    log_partition = forward.log_partition;
  
    for (int i = 0; i < data.seq_len; ++i)
    {
      double Oscale = exp(forward.log_scale_total[i] + backward.log_scale_total[i] - log_partition);
      double Zscale = exp(forward.log_scale_total[i - 1] + backward.log_scale_total[i] - log_partition);

      for (int c = 0; c < num_c; ++c)
      {
        for (int k = 0; k < data.num_haplos; ++k)
          s[i].O(k, c) = forward[i].O(k, c) * backward[i].O(k, c) * Oscale;
    
        for (int p = 0; p < data.num_pops; ++p)
          for (int w = 0; w < 2; ++w)
            s[i].Z(w, p) = forward[i].Z(w, p) * backward[i].Z(w,p) * Zscale;
          
        for (int k = 0; k < data.num_haplos; ++k)
          pop_p[data.haplos[k].pop][i] += s[i].O(k, c);
      }
    }
  }

  void Posterior::DecodeHaplos(const SNPData &data,
                               vector<int> &haplos) const
  {
    haplos.resize(data.seq_len);
    for (int i = 0; i < data.seq_len; ++i)
    {
      int max_k = -1;
      Prob max_p = 0.0;
      for (int k = 0; k < data.num_haplos; ++k)
        if (s[i].O(k) > max_p)
          max_k = k, max_p = s[i].O(k);

      haplos[i] = max_k;
    }
  }

  void Posterior::DecodePops(const SNPData &data,
                             vector<int8> &pops) const
  {
    pops.resize(data.seq_len);
    for (int i = 0; i < data.seq_len; ++i)
    {
      int max_pop = -1;
      Prob max_p = 0.0;
      for (int pop = 0; pop < data.num_pops; ++pop)
        if (pop_p[pop][i] > max_p)
          max_pop = pop, max_p = pop_p[pop][i];
    
      pops[i] = max_pop;
    }
  }


  void Posterior::DecodeHaploPops(const SNPData &data,
                                  vector<int8> &pops) const
  {
    pops.resize(data.seq_len);
    for (int i = 0; i < data.seq_len; ++i)
    {
      int max_k = -1;
      Prob max_p = 0.0;
      for (int k = 0; k < data.num_haplos; ++k)
        if (s[i].O(k) > max_p)
          max_k = k, max_p = s[i].O(k);

      pops[i] = data.haplos[max_k].pop;
    }
  }

  void BaumWelch::ClearCounts()
  {
    total_partition = 0.0;

    if (train & TRAIN_EMISSION)
    {
      emission_counts.clear();
      emission_counts.resize(data->num_pops, vector< vector<double> >(5, vector<double>(5, 1.0)));
    }
    if (train & TRAIN_RECOMB)
    {
      pop_recomb_counts.clear();
      pop_recomb_counts.resize(data->num_pops, vector<double>(data->seq_len - 1, 0.0));
      for (int pop = 0; pop < data->num_pops; ++pop)
        for (int i = 0; i < (data->seq_len - 1); ++i)
          pop_recomb_counts[pop][i] = 0.01 * max(0.0, 1.0 - (data->locus_genetic_pos[i + 1] - data->locus_genetic_pos[i]) * model.pop_recomb[pop]);

      pop_recomb_counts_recomb.clear();
      pop_recomb_counts_recomb.resize(data->num_pops, vector<double>(data->seq_len - 1, 0.0));
      for (int pop = 0; pop < data->num_pops; ++pop)
        for (int i = 0; i < (data->seq_len - 1); ++i)
          pop_recomb_counts_recomb[pop][i] = 0.01 * min(1.0, (data->locus_genetic_pos[i + 1] - data->locus_genetic_pos[i]) * model.pop_recomb[pop]);
    }
    if (train & TRAIN_ADMIX)
    {
      pop_admix_counts.clear();
      pop_admix_counts.resize(data->num_pops, vector<double>(data->num_pops, 1.0));
    }
  }

  void BaumWelch::AccumulateTraining(const vector<int8> &x,
                                     const vector<uint64> &pop_mask_from,
                                     const vector<uint64> &pop_mask_to,
                                     const vector<int> &haplo_mask,
                                     const vector<int> &haplo_exclude)
  {
    Assert(x.size() >= data->seq_len);
    Assert(pop_mask_from.empty() || (pop_mask_from.size() >= data->seq_len));
    Assert(pop_mask_to.empty() || (pop_mask_to.size() >= data->seq_len));
    Assert((pop_mask_from.empty() && pop_mask_to.empty()) ||
           (!pop_mask_from.empty() && !pop_mask_to.empty()));
    bool pop_mask = !pop_mask_from.empty();
    Assert(haplo_mask.empty() || (haplo_mask.size() == data->num_haplos));
    Assert(haplo_exclude.empty() || (haplo_exclude.size() >= data->seq_len));

    //cerr << " F:" << TimerSeconds();
    Forward forward(*data, model, x, pop_mask_from, pop_mask_to, 
                    haplo_mask, haplo_exclude);
    //cerr << " B:" << TimerSeconds();
    Backward backward(*data, model, x, pop_mask_from, pop_mask_to, 
                      haplo_mask, haplo_exclude);
    //cerr << " E:" << TimerSeconds();
    //cerr << "forward log(partition) = " << forward.log_partition << endl;
    //cerr << "backward log(partition) = " << backward.log_partition << endl;
    //cerr << "relative difference = " << fabs((forward.log_partition - backward.log_partition) / (0.5 * (forward.log_partition + backward.log_partition))) << endl;

    total_partition += forward.log_partition;

    vector<double> no_recomb_prob(data->num_pops);
  
    for (int i = 0; i < data->seq_len; ++i)
    {
      if (train & TRAIN_EMISSION)
      {
        double scale = exp(forward.log_scale_total[i] + backward.log_scale_total[i] - forward.log_partition);

        for (int k = 0; k < data->num_haplos; ++k)
        {
          int pop = data->haplos[k].pop;
          if ((!pop_mask || 
               ((pop_mask_from[i] | pop_mask_to[i]) & (uint64(1) << pop))) &&
              (haplo_mask.empty() || haplo_mask[k]) &&
              (haplo_exclude.empty() || (haplo_exclude[i] != k)))
            emission_counts[pop][data->haplos[k][i]][x[i]] += forward[i].O(k) * backward[i].O(k) * scale;
        }
      }
    
      if (i > 0)
      {
        double scale = exp(forward.log_scale_total[i - 1] + backward.log_scale_total[i] - forward.log_partition);

        if (train & TRAIN_ADMIX)
        {
          for (int p = 0; p < data->num_pops; ++p)
          {
            bool from = !pop_mask || (pop_mask_from[i - 1] & (uint64(1) << p));
            double total = 0.0;
            for (int q = 0; q < data->num_pops; ++q)
              if (!pop_mask || (p == q) || (from && (pop_mask_to[i] & (uint64(1) << q))))
                total += model.pop_admix[p][q];
            
            if (total > 0.0)
            {
              double scale = 1.0 / total;
              for (int q = 0; q < data->num_pops; ++q)
                if (!pop_mask || (p == q) || (from && (pop_mask_to[i] & (uint64(1) << q))))
                  pop_admix_counts[p][q] += forward[i].Z(0, p) * model.pop_admix[p][q] * scale * backward[i].Z(1, q) * scale;
            }
          }
        }
        
        if (train & TRAIN_RECOMB)
        {
          for (int pop = 0; pop < data->num_pops; ++pop)
            no_recomb_prob[pop] = exp(-model.pop_recomb[pop] * (data->locus_genetic_pos[i] - data->locus_genetic_pos[i - 1]));

          for (int k = 0; k < data->num_haplos; ++k)
          {
            int pop = data->haplos[k].pop;
            if ((!pop_mask || 
                 ((pop_mask_from[i - 1] | pop_mask_to[i - 1]) & (uint64(1) << pop))) &&
                (haplo_mask.empty() || haplo_mask[k]) &&
                (haplo_exclude.empty() || (haplo_exclude[i - 1] != k)))
            {
              if (!pop_mask || 
                  ((pop_mask_from[i] | pop_mask_to[i]) & (uint64(1) << pop)))
              {
                double expect;
                int pair_k = data->haplos[k].diplo_pair;
                if ((pair_k >= 0) && 
                    (haplo_mask.empty() || haplo_mask[pair_k]) &&
                    (haplo_exclude.empty() || (haplo_exclude[i] != pair_k)))
                {
                  double switch_prob = SNPData::switch_decode(data->diplos[data->haplos[k].diplo].switches[i - 1]);
                  expect = (1.0 - switch_prob) * model.emission[pop][data->haplos[k][i]][x[i]] * backward[i].O(k) + switch_prob * model.emission[pop][data->haplos[pair_k][i]][x[i]] * backward[i].O(pair_k);
                  expect = forward[i - 1].O(k) * no_recomb_prob[pop] * expect * exp(forward.log_scale_total[i - 1] + backward.log_scale_total[i] - forward.log_partition);
                }
                else
                  expect = forward[i - 1].O(k) * no_recomb_prob[pop] * model.emission[pop][data->haplos[k][i]][x[i]] * backward[i].O(k) * scale;
                pop_recomb_counts[pop][i - 1] += expect;
              }

              double expect = forward[i - 1].O(k) * (1.0 - no_recomb_prob[pop]) * backward[i].Z(0, pop) * scale;
              pop_recomb_counts_recomb[pop][i - 1] += expect;
            }
          }
        }
      }
    }

    //cerr << " D:" << TimerSeconds();
  }

  bool BaumWelch::UpdateModel(double &partition_rel_diff)
  {
    partition_rel_diff = fabs((total_partition - last_total_partition) / (0.5 * (total_partition + last_total_partition)));
    bool changed = partition_rel_diff >= 1e-8;
    last_total_partition = total_partition;
    last_model = model;
  
    if (train & TRAIN_EMISSION)
    {
      for (int pop = 0; pop < data->num_pops; ++pop)
        for (int y = 0; y < 5; ++y)
        {
          double total = 0.0;
          for (int x = 0; x < 5; ++x)
            total += emission_counts[pop][y][x];
          Assert(total != 0.0);
          
          for (int x = 0; x < 5; ++x)
            model.emission[pop][y][x] = emission_counts[pop][y][x] / total;
        }
    }
  
    if (train & TRAIN_RECOMB)
    {
      for (int pop = 0; pop < data->num_pops; ++pop)
      {
        int iterations = 0;
        const int max_iterations = 10000;
        while (iterations++ < max_iterations)
        {
          double f = 0.0, df = 0.0;
          for (int i = 0; i < (data->seq_len - 1); ++i)
          {
            double recombs = data->locus_genetic_pos[i + 1] - data->locus_genetic_pos[i];
            if (recombs != 0.0)
            {
              double e = exp(-model.pop_recomb[pop] * recombs);
              f += recombs * (e / (1.0 - e) * pop_recomb_counts_recomb[pop][i] - pop_recomb_counts[pop][i]);
            
              double e1s = (1.0 - e) * (1.0 - e);
              df -= 2.0 * recombs * recombs * e / e1s * pop_recomb_counts_recomb[pop][i];
            }
          }
          AssertMsg(df != 0.0, ToStr(f) + ", " + ToStr(df) + ", " + ToStr(model.pop_recomb[pop]));
        
          double delta = -f / df;
          if (fabs(delta) > (0.5 * model.pop_recomb[pop]))
            delta = ((delta < 0) ? -1.0 : 1.0) * 0.5 * model.pop_recomb[pop];

          double last_recomb = model.pop_recomb[pop];
          model.pop_recomb[pop] += delta;
          if (model.pop_recomb[pop] == last_recomb)
            break;

          if (model.pop_recomb[pop] < 1.0e-6)
          {
            model.pop_recomb[pop] = 1.0e-6;
            break;
          }
          if (model.pop_recomb[pop] > 1.0e6)
          {
            model.pop_recomb[pop] = 1.0e6;
            break;
          }
          if (fabs(delta / model.pop_recomb[pop]) < 1e-9)
            break;
        }
      }
    }
  
    if (train & TRAIN_ADMIX)
    {
      for (int p = 0; p < data->num_pops; ++p)
      {
        double total = 0.0;
        for (int q = 0; q < data->num_pops; ++q)
          total += pop_admix_counts[p][q];
        Assert(total != 0.0);
      
        for (int q = 0; q < data->num_pops; ++q)
          model.pop_admix[p][q] = pop_admix_counts[p][q] / total;
      }
#if 0
      cerr << "pop_admix Counts:" << endl;
      cerr << ToStr("from\\to", 12);
      for (int q = 0; q < data->num_pops; ++q)
        cerr << ToStr(q, 12);
      cerr << "\n";
    
      for (int p = 0; p < data->num_pops; ++p)
      {
        cerr << ToStr(p, 12);
        for (int q = 0; q < data->num_pops; ++q)
          cerr << ToStr(pop_admix_counts[p][q], 12, 5);
        cerr << "\n";
      }
#endif

    }
  
    ClearCounts();
    return changed;
  }

  bool BaumWelch::UpdateModel()
  {
    double partition_rel_diff;
    return UpdateModel(partition_rel_diff);
  }



  void Viterbi::Calculate(const SNPData &data, const Model &model,
                          const vector<int8> &x,
                          const vector<uint64> &pop_mask_from,
                          const vector<uint64> &pop_mask_to,
                          const vector<int> &haplo_mask,
                          const vector<int> &haplo_exclude,
                          int max_count_)
  {
    Assert(x.size() >= data.seq_len);
    Assert(pop_mask_from.empty() || (pop_mask_from.size() >= data.seq_len));
    Assert(pop_mask_to.empty() || (pop_mask_to.size() >= data.seq_len));
    Assert((pop_mask_from.empty() && pop_mask_to.empty()) ||
           (!pop_mask_from.empty() && !pop_mask_to.empty()));
    bool pop_mask = !pop_mask_from.empty();
    Assert(haplo_mask.empty() || (haplo_mask.size() == data.num_haplos));
    Assert(haplo_exclude.empty() || (haplo_exclude.size() >= data.seq_len));

    max_count = max_count_;
    int num_c = max(max_count, 1);
    ptr.clear();
    ptr.resize(data.seq_len, StateProfile<int>(data, num_c));
    scale.clear();
    scale.resize(data.seq_len, 1.0);
    log_scale_total.clear();
    log_scale_total.resize(data.seq_len);

    StateProfile<double> last_s(data, num_c);
    StateProfile<double> new_s(data, num_c);
    
    for (int c = 0; c < num_c; ++c)
      for (int p = 0; p < data.num_pops; ++p)
      {
        new_s.Z(0, p, c) = 0.0;
        new_s.Z(1, p, c) = 0.0;
      }

    vector<double> init_edges;
    ComputeInitEdges(data, model, init_edges,
                     !pop_mask ? uint64(0) : (pop_mask_from[0]|pop_mask_to[0]),
                     haplo_mask, haplo_exclude);
    init_edges.clear();
    init_edges.resize(data.num_haplos, 1.0);
    for (int k = 0; k < data.num_haplos; ++k)
    {
      new_s.O(k, 0) = init_edges[k] * model.emission[data.haplos[k].pop][data.haplos[k][0]][x[0]];
      for (int c = 1; c < num_c; ++c)
        new_s.O(k, c) = 0.0;
    }
  
    new_s.Normalize(scale[0]);

    vector<double> no_recomb_prob(data.num_pops);
    vector<int> pop_haplo_count(data.num_pops, 0);
    vector<double> pop_haplo_scale(data.num_pops, 0.0);

    for (int i = 1; i < data.seq_len; ++i)
    {
      last_s.swap(new_s);
      new_s.clear();

      pop_haplo_count.clear();
      pop_haplo_count.resize(data.num_pops, 0);
      for (int k = 0; k < data.num_haplos; ++k)
        if ((haplo_mask.empty() || haplo_mask[k]) &&
            (haplo_exclude.empty() || (haplo_exclude[i] != k)))
          ++pop_haplo_count[data.haplos[k].pop];
      
      for (int pop = 0; pop < data.num_pops; ++pop)
      {
        if ((!pop_mask || 
             ((pop_mask_from[i] | pop_mask_to[i]) & (uint64(1) << pop))) &&
            (pop_haplo_count[pop] > 0))
          pop_haplo_scale[pop] = 1.0 / double(pop_haplo_count[pop]);
        else
          pop_haplo_scale[pop] = 0.0;
      }

      for (int pop = 0; pop < data.num_pops; ++pop)
        no_recomb_prob[pop] = exp(-model.pop_recomb[pop] * (data.locus_genetic_pos[i] - data.locus_genetic_pos[i - 1]));
      
      for (int k = 0; k < data.num_haplos; ++k)
        if ((haplo_mask.empty() || (haplo_mask[k])) &&
            (haplo_exclude.empty() || (haplo_exclude[i - 1] != k)))
        {
          int pop = data.haplos[k].pop;
          for (int c = 0; c < num_c; ++c)
          {
            double prob = last_s.O(k, c) * (1.0 - no_recomb_prob[pop]);
            if (prob > new_s.Z(0, pop, c))
              new_s.Z(0, pop, c) = prob, ptr[i].Z(0, pop, c) = k;
          }
        }
    
      for (int p = 0; p < data.num_pops; ++p)
      {
        bool from = !pop_mask || (pop_mask_from[i - 1] & (uint64(1) << p));
        double total = 0.0;
        for (int q = 0; q < data.num_pops; ++q)
          if (!pop_mask || (p == q) || (from && (pop_mask_to[i] & (uint64(1) << q))))
            total += model.pop_admix[p][q];

        if (total > 0.0)
        {
          double scale = 1.0 / total;
          for (int q = 0; q < data.num_pops; ++q)
            if (!pop_mask || (p == q) || (from && (pop_mask_to[i] & (uint64(1) << q))))
            {
              if (max_count == 0)
              {
                double prob = new_s.Z(0, p) * model.pop_admix[p][q] * scale;
                if (prob > new_s.Z(1, q))
                  new_s.Z(1, q) = prob, ptr[i].Z(1, q) = p;
              }
              else
              {
                for (int c = 1; c < num_c; ++c)
                {
                  double prob = new_s.Z(0, p, c - 1) * model.pop_admix[p][q] * scale;
                  if (prob > new_s.Z(1, q, c))
                    new_s.Z(1, q, c) = prob, ptr[i].Z(1, q, c) = p;
                }
              }
            }
        }
      }

      for (int k = 0; k < data.num_haplos; ++k)
      {
        int pop = data.haplos[k].pop;
        if ((!pop_mask || 
             ((pop_mask_from[i] | pop_mask_to[i]) & (uint64(1) << pop))) &&
            (haplo_mask.empty() || haplo_mask[k]) &&
            (haplo_exclude.empty() || (haplo_exclude[i] != k)))
        {
          int pair_k = data.haplos[k].diplo_pair;
          if ((pair_k >= 0) && (haplo_mask.empty() || haplo_mask[pair_k]) &&
              (haplo_exclude.empty() || (haplo_exclude[i - 1] != pair_k)))
          {
            double switch_prob = SNPData::switch_decode(data.diplos[data.haplos[k].diplo].switches[i - 1]);
            for (int c = 0; c < num_c; ++c)
            {
              new_s.O(k, c) = last_s.O(k, c) * (1.0 - switch_prob) * no_recomb_prob[pop];
              ptr[i].O(k, c) = k;

              double prob = last_s.O(pair_k, c) * switch_prob * no_recomb_prob[pop];
              if (prob > new_s.O(k, c))
                new_s.O(k, c) = prob, ptr[i].O(k, c) = pair_k;
            }
          }
          else
          {
            for (int c = 0; c < num_c; ++c)
            {
              new_s.O(k, c) = last_s.O(k, c) * no_recomb_prob[pop];
              ptr[i].O(k, c) = k;
            }
          }

          for (int c = 0; c < num_c; ++c)
          {
            double prob = new_s.Z(1, pop, c) * pop_haplo_scale[pop];
            if (prob > new_s.O(k, c))
              new_s.O(k, c) = prob, ptr[i].O(k, c) = -1;
            
            new_s.O(k, c) *= model.emission[pop][data.haplos[k][i]][x[i]];
          }
        }
      }

      {
        int non_zero_count = 0;
        for (int c = 0; c < num_c; ++c)
          for (int k = 0; k < data.num_haplos; ++k)
            if (new_s.O(k, c) != 0.0)
              ++non_zero_count;
        Assert(non_zero_count > 0);
      }

      new_s.Normalize(scale[i]);
    }

    double log_total = 0.0;
    for (int i = 0; i < data.seq_len; ++i)
      log_scale_total[i] = log_total += log(scale[i]);

    parse.resize(data.seq_len, -1);
    recomb.resize(data.seq_len - 1, -1);

    int max_c = 0;
    int max_k = 0;
    for (int c = 0; c < num_c; ++c)
      for (int k = 0; k < data.num_haplos; ++k)
        if (new_s.O(k, c) > new_s.O(max_k, max_c))
          max_k = k, max_c = c;

    Assert(new_s.O(max_k, max_c) != 0.0);
    log_likelihood = log(new_s.O(max_k, max_c)) + log_scale_total[data.seq_len - 1];

    for (int i = data.seq_len - 1; i >= 0; --i)
    {
      parse[i] = max_k;
      if (i == 0)
        break;

      if (ptr[i].O(max_k, max_c) >= 0)
      {
        max_k = ptr[i].O(max_k);
        recomb[i - 1] = 0;
      }
      else
      {
        int pop = data.haplos[max_k].pop;
        pop = ptr[i].Z(1, pop, max_c);
        if (max_count > 0)
          --max_c;
        max_k = ptr[i].Z(0, pop, max_c);
        recomb[i - 1] = 1;
      }
    }
  }


  void Viterbi::DecodeHaplos(const SNPData &data,
                             vector<int> &haplos) const
  {
    haplos.resize(data.seq_len);
    for (int i = 0; i < data.seq_len; ++i)
      haplos[i] = parse[i];
  }

  void Viterbi::DecodePops(const SNPData &data,
                           vector<int8> &pops) const
  {
    pops.resize(data.seq_len);
    for (int i = 0; i < data.seq_len; ++i)
      pops[i] = data.haplos[parse[i]].pop;
  }



  void ViterbiForward::Calculate(const SNPData &data, const Model &model,
                                 const vector<int8> &x,
                                 const vector<uint64> &pop_mask_from,
                                 const vector<uint64> &pop_mask_to,
                                 const vector<int> &haplo_mask,
                                 const vector<int> &haplo_exclude)
  {
    Assert(x.size() >= data.seq_len);
    Assert(pop_mask_from.empty() || (pop_mask_from.size() >= data.seq_len));
    Assert(pop_mask_to.empty() || (pop_mask_to.size() >= data.seq_len));
    Assert((pop_mask_from.empty() && pop_mask_to.empty()) ||
           (!pop_mask_from.empty() && !pop_mask_to.empty()));
    bool pop_mask = !pop_mask_from.empty();
    Assert(haplo_mask.empty() || (haplo_mask.size() == data.num_haplos));
    Assert(haplo_exclude.empty() || (haplo_exclude.size() >= data.seq_len));

    s.clear();
    s.resize(data.seq_len, StateProfile<Prob>(data));
    scale.clear();
    scale.resize(data.seq_len, 1.0);
    log_scale_total.clear();
    log_scale_total.resize(data.seq_len);

    StateProfile<double> last_s(data);
    StateProfile<double> new_s(data);
  
    for (int p = 0; p < data.num_pops; ++p)
    {
      new_s.Z(0, p) = 0.0;
      new_s.Z(1, p) = 0.0;
    }

    vector<double> init_edges;
    ComputeInitEdges(data, model, init_edges,
                     !pop_mask ? uint64(0) : (pop_mask_from[0]|pop_mask_to[0]),
                     haplo_mask, haplo_exclude);
    init_edges.clear();
    init_edges.resize(data.num_haplos, 1.0);
    for (int k = 0; k < data.num_haplos; ++k)
      new_s.O(k) = init_edges[k] * model.emission[data.haplos[k].pop][data.haplos[k][0]][x[0]];
  
    new_s.Normalize(scale[0]);
    s[0] = new_s;

    vector<double> no_recomb_prob(data.num_pops);
    vector<int> pop_haplo_count(data.num_pops, 0);
    vector<double> pop_haplo_scale(data.num_pops, 0.0);

    for (int i = 1; i < data.seq_len; ++i)
    {
      last_s.swap(new_s);
      new_s.clear();

      pop_haplo_count.clear();
      pop_haplo_count.resize(data.num_pops, 0);
      for (int k = 0; k < data.num_haplos; ++k)
        if ((haplo_mask.empty() || haplo_mask[k]) &&
            (haplo_exclude.empty() || (haplo_exclude[i] != k)))
          ++pop_haplo_count[data.haplos[k].pop];
      
      for (int pop = 0; pop < data.num_pops; ++pop)
      {
        if ((!pop_mask || 
             ((pop_mask_from[i] | pop_mask_to[i]) & (uint64(1) << pop))) &&
            (pop_haplo_count[pop] > 0))
          pop_haplo_scale[pop] = 1.0 / double(pop_haplo_count[pop]);
        else
          pop_haplo_scale[pop] = 0.0;
      }

      for (int pop = 0; pop < data.num_pops; ++pop)
        no_recomb_prob[pop] = exp(-model.pop_recomb[pop] * (data.locus_genetic_pos[i] - data.locus_genetic_pos[i - 1]));
      
      for (int k = 0; k < data.num_haplos; ++k)
        if ((haplo_mask.empty() || (haplo_mask[k])) &&
            (haplo_exclude.empty() || (haplo_exclude[i - 1] != k)))
        {
          int pop = data.haplos[k].pop;
          double prob = last_s.O(k) * (1.0 - no_recomb_prob[pop]);
          if (prob > new_s.Z(0, pop))
            new_s.Z(0, pop) = prob;
        }
    
      for (int p = 0; p < data.num_pops; ++p)
      {
        bool from = !pop_mask || (pop_mask_from[i - 1] & (uint64(1) << p));
        double total = 0.0;
        for (int q = 0; q < data.num_pops; ++q)
          if (!pop_mask || (p == q) || (from && (pop_mask_to[i] & (uint64(1) << q))))
            total += model.pop_admix[p][q];

        if (total > 0.0)
        {
          double scale = 1.0 / total;
          for (int q = 0; q < data.num_pops; ++q)
            if (!pop_mask || (p == q) || (from && (pop_mask_to[i] & (uint64(1) << q))))
            {
              double prob = new_s.Z(0, p) * model.pop_admix[p][q] * scale;
              if (prob > new_s.Z(1, q))
                new_s.Z(1, q) = prob;
            }
        }
      }

      for (int k = 0; k < data.num_haplos; ++k)
      {
        int pop = data.haplos[k].pop;
        if ((!pop_mask || 
             ((pop_mask_from[i] | pop_mask_to[i]) & (uint64(1) << pop))) &&
            (haplo_mask.empty() || haplo_mask[k]) &&
            (haplo_exclude.empty() || (haplo_exclude[i] != k)))
        {
          int pair_k = data.haplos[k].diplo_pair;
          if ((pair_k >= 0) && (haplo_mask.empty() || haplo_mask[pair_k]) &&
              (haplo_exclude.empty() || (haplo_exclude[i - 1] != pair_k)))
          {
            double switch_prob = SNPData::switch_decode(data.diplos[data.haplos[k].diplo].switches[i - 1]);
            new_s.O(k) = last_s.O(k) * (1.0 - switch_prob) * no_recomb_prob[pop];
            double prob = last_s.O(pair_k) * switch_prob * no_recomb_prob[pop];
            if (prob > new_s.O(k))
              new_s.O(k) = prob;
          }
          else
            new_s.O(k) = last_s.O(k) * no_recomb_prob[pop];
          
          double prob = new_s.Z(1, pop) * pop_haplo_scale[pop];
          if (prob > new_s.O(k))
            new_s.O(k) = prob;

          new_s.O(k) *= model.emission[pop][data.haplos[k][i]][x[i]];
        }
      }

      {
        int non_zero_count = 0;
        for (int k = 0; k < data.num_haplos; ++k)
          if (new_s.O(k) != 0.0)
            ++non_zero_count;
        Assert(non_zero_count > 0);
      }

      new_s.Normalize(scale[i]);
      s[i] = new_s;
    }

    double log_total = 0.0;
    for (int i = 0; i < data.seq_len; ++i)
      log_scale_total[i] = log_total += log(scale[i]);

    int max_k = 0;
    for (int k = 1; k < data.num_haplos; ++k)
      if (new_s.O(k) > new_s.O(max_k))
        max_k = k;

    Assert(new_s.O(max_k) != 0.0);
    log_likelihood = log(new_s.O(max_k)) + log_scale_total[data.seq_len - 1];
  }


  void ViterbiBackward::Calculate(const SNPData &data, const Model &model,
                                  const vector<int8> &x,
                                  const vector<uint64> &pop_mask_from,
                                  const vector<uint64> &pop_mask_to,
                                  const vector<int> &haplo_mask,
                                  const vector<int> &haplo_exclude)
  {
    Assert(x.size() >= data.seq_len);
    Assert(pop_mask_from.empty() || (pop_mask_from.size() >= data.seq_len));
    Assert(pop_mask_to.empty() || (pop_mask_to.size() >= data.seq_len));
    Assert((pop_mask_from.empty() && pop_mask_to.empty()) ||
           (!pop_mask_from.empty() && !pop_mask_to.empty()));
    bool pop_mask = !pop_mask_from.empty();
    Assert(haplo_mask.empty() || (haplo_mask.size() == data.num_haplos));
    Assert(haplo_exclude.empty() || (haplo_exclude.size() >= data.seq_len));
  
    s.clear();
    s.resize(data.seq_len, StateProfile<Prob>(data));
    scale.clear();
    scale.resize(data.seq_len, 1.0);
    log_scale_total.clear();
    log_scale_total.resize(data.seq_len);
  
    StateProfile<double> last_s(data);
    StateProfile<double> new_s(data);
  
    for (int k = 0; k < data.num_haplos; ++k)
      new_s.O(k) = 1.0;
    new_s.Normalize(scale[data.seq_len - 1]);
  
    vector<double> init_edges;
    ComputeInitEdges(data, model, init_edges,
                     !pop_mask ? uint64(0) : (pop_mask_from[0]|pop_mask_to[0]),
                     haplo_mask, haplo_exclude);
    init_edges.clear();
    init_edges.resize(data.num_haplos, 1.0);
    vector<double> no_recomb_prob(data.num_pops);
    vector<int> pop_haplo_count(data.num_pops, 0);
    vector<double> pop_haplo_scale(data.num_pops, 0.0);

    for (int i = data.seq_len - 2; i >= 0; --i)
    {
      pop_haplo_count.clear();
      pop_haplo_count.resize(data.num_pops, 0);
      for (int k = 0; k < data.num_haplos; ++k)
        if ((haplo_mask.empty() || haplo_mask[k]) &&
            (haplo_exclude.empty() || (haplo_exclude[i + 1] != k)))
          ++pop_haplo_count[data.haplos[k].pop];
      
      for (int pop = 0; pop < data.num_pops; ++pop)
      {
        if ((!pop_mask || 
             ((pop_mask_from[i+1] | pop_mask_to[i+1]) & (uint64(1) << pop))) &&
            (pop_haplo_count[pop] > 0))
          pop_haplo_scale[pop] = 1.0 / double(pop_haplo_count[pop]);
        else
          pop_haplo_scale[pop] = 0.0;
      }

      for (int pop = 0; pop < data.num_pops; ++pop)
        no_recomb_prob[pop] = exp(-model.pop_recomb[pop] * (data.locus_genetic_pos[i + 1] - data.locus_genetic_pos[i]));
      
      for (int k = 0; k < data.num_haplos; ++k)
      {
        int pop = data.haplos[k].pop;
        if ((!pop_mask || 
             ((pop_mask_from[i+1] | pop_mask_to[i+1]) & (uint64(1) << pop))) &&
            (haplo_mask.empty() || haplo_mask[k]) &&
            (haplo_exclude.empty() || (haplo_exclude[i + 1] != k)))
        {
          double prob = pop_haplo_scale[pop] * model.emission[pop][data.haplos[k][i + 1]][x[i + 1]] * new_s.O(k);
          if (prob > new_s.Z(1, pop))
            new_s.Z(1, pop) = prob;
        }
      }
    
      for (int p = 0; p < data.num_pops; ++p)
      {
        bool from = !pop_mask || (pop_mask_from[i] & (uint64(1) << p));
        double total = 0.0;
        for (int q = 0; q < data.num_pops; ++q)
          if (!pop_mask || (p == q) || (from && (pop_mask_to[i + 1] & (uint64(1) << q))))
            total += model.pop_admix[p][q];

        if (total > 0.0)
        {
          double scale = 1.0 / total;
          for (int q = 0; q < data.num_pops; ++q)
            if (!pop_mask || (p == q) || (from && (pop_mask_to[i + 1] & (uint64(1) << q))))
            {
              double prob = model.pop_admix[p][q] * scale * new_s.Z(1, q);
              if (prob > new_s.Z(0, p))
                new_s.Z(0, p) = prob;
            }
        }
      }

      s[i + 1] = new_s;
      last_s.swap(new_s);
      new_s.clear();

      for (int k = 0; k < data.num_haplos; ++k)
      {
        int pop = data.haplos[k].pop;
        if ((!pop_mask || 
             ((pop_mask_from[i] | pop_mask_to[i]) & (uint64(1) << pop))) &&
            (haplo_mask.empty() || haplo_mask[k]) &&
            (haplo_exclude.empty() || (haplo_exclude[i] != k)))
        {
          int pair_k = data.haplos[k].diplo_pair;
          if ((pair_k >= 0) && (haplo_mask.empty() || haplo_mask[pair_k]) &&
              (haplo_exclude.empty() || (haplo_exclude[i + 1] != pair_k)))
          {
            double switch_prob = SNPData::switch_decode(data.diplos[data.haplos[k].diplo].switches[i]);
            new_s.O(k) = no_recomb_prob[pop] * (1.0 - switch_prob) * model.emission[pop][data.haplos[k][i + 1]][x[i + 1]] * last_s.O(k);
            double prob = no_recomb_prob[pop] * switch_prob * model.emission[pop][data.haplos[pair_k][i + 1]][x[i + 1]] * last_s.O(pair_k);
            if (prob > new_s.O(k))
              new_s.O(k) = prob;
          }
          else
            new_s.O(k) = no_recomb_prob[pop] * model.emission[pop][data.haplos[k][i + 1]][x[i + 1]] * last_s.O(k);

          double prob = (1.0 - no_recomb_prob[pop]) * last_s.Z(0, pop);
          if (prob > new_s.O(k))
            new_s.O(k) = prob;
        }
      }

      new_s.Normalize(scale[i]);
    }
    s[0] = new_s;
    
    double log_total = 0.0;
    for (int i = data.seq_len - 1; i >= 0; --i)
      log_scale_total[i] = log_total += log(scale[i]);

    int max_k = -1;
    double max_prob = 0.0;
    for (int k = 0; k < data.num_haplos; ++k)
    {
      double prob = new_s.O(k) * init_edges[k] * model.emission[data.haplos[k].pop][data.haplos[k][0]][x[0]];
      if ((max_k < 0) || (prob > max_prob))
        max_k = k, max_prob = prob;
    }

    Assert(max_prob != 0.0);
    log_likelihood = log(max_prob) + log_scale_total[0];
  }


}
