
#ifndef HAPLOMODEL_H
#define HAPLOMODEL_H

#include "System.h"
#include "SNPData.h"


namespace HaploModel
{
  
  typedef float Prob;
  
  template <typename T = Prob>
  class StateProfile
  {
    public:
    StateProfile(const SNPData &data, int num_count = 1)
      : pops_offset(data.num_haplos), 
        count_size(data.num_haplos + 2*data.num_pops),
        s(num_count * count_size, 0)
    {
    }

    StateProfile(const StateProfile &init)
      : pops_offset(init.pops_offset), count_size(init.count_size), s(init.s)
    {
    }

    T &O(int k, int c = 0) { return s[count_size*c + k]; }
    T O(int k, int c = 0) const { return s[count_size*c + k]; }

    T &Z(int w, int p, int c = 0) { return s[count_size*c + pops_offset + 2*p + w]; }
    T Z(int w, int p, int c = 0) const { return s[count_size*c + pops_offset + 2*p + w]; }

    void Normalize(double &scale);

    int num_haplos() const { return pops_offset; }
    int num_pops() const { return (count_size - pops_offset)/2; }
    int num_count() const { return (int(s.size()) / count_size - 1); }

    template <typename T2>
    const StateProfile<T> &operator=(const StateProfile<T2> &init)
    {
      pops_offset = init.pops_offset;
      count_size = init.count_size;
      s.resize(init.s.size());
      for (int i = 0; i < s.size(); ++i)
        s[i] = T(init.s[i]);
    }
  
    void swap(StateProfile<T> &other)
    {
      ::swap(pops_offset, other.pops_offset);
      ::swap(count_size, other.count_size);
      s.swap(other.s);
    }

    void clear()
    {
      int size = s.size();
      s.clear();
      s.resize(size, 0);
    }

    int pops_offset;
    int count_size;
    vector<T> s;
  };


  class Model
  {
  public:
    int num_pops;

    vector< vector< vector<double> > > emission;
    vector<double> pop_init;
    vector<double> pop_recomb;
    vector< vector<double> > pop_admix;

    Model(const SNPData &data)
    {
      Init(data.num_pops);
    }

    Model(int num_pops)
    {
      Init(num_pops);
    }

    void Init(int num_pops_);

    void DisplayEmission(ostream &out) const;
    void DisplayPopInit(ostream &out) const;
    void DisplayPopRecomb(ostream &out) const;
    void DisplayPopAdmix(ostream &out) const;

    void SetDefault();
    void SetDefaultEmission(double error_rate = 0.01, double missing_rate = 0.05);
    void SetDefaultPopInit();
    void SetDefaultPopRecomb(double rate = 1000.0);
    void SetDefaultPopAdmix(double self_p = 0.99);

    void Generate(const SNPData &data, 
                  vector<int8> &output, vector<int> &haplos);

    void Write(ostream &o) const;
    void Read(istream &i);

    void Write2(ostream &o) const;
    void Read2(istream &i);
  };


  void GetHaploPops(const SNPData &data, 
                    const vector<int> &haplos, vector<int8> &pops);

  void GetPopMask(const SNPData &data, const vector<int8> &pops, 
                  vector<uint64> &pop_mask);

  
  class Forward
  {
  public:
    vector< StateProfile<Prob> > s;
    vector<double> scale;
    vector<double> log_scale_total;
    double total_s, log_partition;
    int max_count;
    
    Forward() { }

    Forward(const SNPData &data, const Model &model,
            const vector<int8> &x, 
            const vector<uint64> &pop_mask = vector<uint64>(),
            const vector<int> &haplo_mask = vector<int>(),
            const vector<int> &haplo_exclude = vector<int>(), 
            int max_count = 0)
    {
      Calculate(data, model, x, pop_mask, haplo_mask, haplo_exclude, 
                max_count);
    }
    
    Forward(const SNPData &data, const Model &model,
            const vector<int8> &x,
            const vector<uint64> &pop_mask_from,
            const vector<uint64> &pop_mask_to,
            const vector<int> &haplo_mask = vector<int>(),
            const vector<int> &haplo_exclude = vector<int>(),
            int max_count = 0)
    {
      Calculate(data, model, x, pop_mask_from, pop_mask_to,
                haplo_mask, haplo_exclude, max_count);
    }

    void Calculate(const SNPData &data, const Model &model,
                   const vector<int8> &x,
                   const vector<uint64> &pop_mask = vector<uint64>(),
                   const vector<int> &haplo_mask = vector<int>(),
                   const vector<int> &haplo_exclude = vector<int>(),
                   int max_count = 0)
    {
      Calculate(data, model, x, pop_mask, pop_mask, haplo_mask, haplo_exclude, 
                max_count);
    }

    void Calculate(const SNPData &data, const Model &model,
                   const vector<int8> &x,
                   const vector<uint64> &pop_mask_from,
                   const vector<uint64> &pop_mask_to,
                   const vector<int> &haplo_mask = vector<int>(),
                   const vector<int> &haplo_exclude = vector<int>(),
                   int max_count_ = 0);

    const StateProfile<Prob> &operator[](int index) const { return s[index]; }

    int size() const { return int(s.size()); }
  };


  class Backward
  {
  public:
    vector< StateProfile<Prob> > s;
    vector<double> scale;
    vector<double> log_scale_total;
    double total_s, log_partition;
    int max_count;
    
    Backward() { }
    
    Backward(const SNPData &data, const Model &model,
             const vector<int8> &x,
             const vector<uint64> &pop_mask = vector<uint64>(),
             const vector<int> &haplo_mask = vector<int>(),
             const vector<int> &haplo_exclude = vector<int>(),
             int max_count = 0)
    {
      Calculate(data, model, x, pop_mask, haplo_mask, haplo_exclude,
                max_count);
    }
    
    Backward(const SNPData &data, const Model &model,
             const vector<int8> &x,
             const vector<uint64> &pop_mask_from,
             const vector<uint64> &pop_mask_to,
             const vector<int> &haplo_mask = vector<int>(),
             const vector<int> &haplo_exclude = vector<int>(),
             int max_count = 0)
    {
      Calculate(data, model, x, pop_mask_from, pop_mask_to,
                haplo_mask, haplo_exclude, max_count);
    }

    void Calculate(const SNPData &data, const Model &model,
                   const vector<int8> &x,
                   const vector<uint64> &pop_mask = vector<uint64>(),
                   const vector<int> &haplo_mask = vector<int>(),
                   const vector<int> &haplo_exclude = vector<int>(),
                   int max_count = 0)
    {
      Calculate(data, model, x, pop_mask, pop_mask, haplo_mask, haplo_exclude,
                max_count);
    }

    void Calculate(const SNPData &data, const Model &model,
                   const vector<int8> &x, 
                   const vector<uint64> &pop_mask_from,
                   const vector<uint64> &pop_mask_to,
                   const vector<int> &haplo_mask = vector<int>(),
                   const vector<int> &haplo_exclude = vector<int>(),
                   int max_count_ = 0);

    const StateProfile<Prob> &operator[](int index) const { return s[index]; }

    int size() const { return int(s.size()); }
  };


  class Posterior
  {
  public:
    vector< StateProfile<Prob> > s;
    vector< vector<Prob> > pop_p;
    double log_partition;
    
    Posterior() { }
  
    Posterior(const SNPData &data, 
              const Forward &forward, const Backward &backward)
    {
      Calculate(data, forward, backward);
    }
    
    void Calculate(const SNPData &data,
                   const Forward &forward, const Backward &backward);

    const StateProfile<Prob> &operator[](int index) const { return s[index]; }
    
    int size() const { return int(s.size()); }

    void DecodeHaplos(const SNPData &data, vector<int> &haplos) const;
    void DecodePops(const SNPData &data, vector<int8> &pops) const;
    void DecodeHaploPops(const SNPData &data, vector<int8> &pops) const;
  };



  class BaumWelch
  {
  public:
    vector< vector< vector<double> > > emission_counts;
    vector< vector<double> > pop_recomb_counts, pop_recomb_counts_recomb;
    vector< vector<double> > pop_admix_counts;
    
    static const uint32 TRAIN_EMISSION = 0x1;
    static const uint32 TRAIN_RECOMB = 0x2;
    static const uint32 TRAIN_ADMIX = 0x4;
    static const uint32 TRAIN_ALL = 0x7;
    
    uint32 train;
    
    Model model, last_model;
    double total_partition, last_total_partition;
    
    BaumWelch(const SNPData &data_, uint32 train_ = TRAIN_ALL) 
      : data(&data_), model(data_), last_model(data_), train(train_)
    {
      model.SetDefault();
      InitTraining();
    }
    
    BaumWelch(const SNPData &data_, const Model &init_model, 
              uint32 train_ = TRAIN_ALL) 
      : data(&data_), model(init_model), last_model(init_model), train(train_)
    {
      InitTraining();
    }

    void SetModel(const Model &init_model)
    {
      model = init_model;
      last_model = init_model;
    }
    
    void SetTrain(uint32 train_ = TRAIN_ALL)
    {
      train = train_;
    }
    
    void InitTraining()
    {
      last_model = model;
      last_total_partition = 0.0;
      ClearCounts();
    }
    
    void ClearCounts();

    void AccumulateTraining(const vector<int8> &x,
                            const vector<uint64> &pop_mask = vector<uint64>(),
                            const vector<int> &haplo_mask = vector<int>(),
                            const vector<int> &haplo_exclude = vector<int>())
    {
      AccumulateTraining(x, pop_mask, pop_mask, haplo_mask, haplo_exclude);
    }

    void AccumulateTraining(const vector<int8> &x,
                            const vector<uint64> &pop_mask_from,
                            const vector<uint64> &pop_mask_to,
                            const vector<int> &haplo_mask = vector<int>(),
                            const vector<int> &haplo_exclude = vector<int>());

    bool UpdateModel(double &partition_rel_diff);
    bool UpdateModel();

  private:
    const SNPData *data;
  };


  class Viterbi
  {
  public:
    vector< StateProfile<int> > ptr;
    vector<double> scale;
    vector<double> log_scale_total;
    double log_likelihood;
    int max_count;

    vector<int> parse;
    vector<int> recomb;

    Viterbi() { }

    Viterbi(const SNPData &data, const Model &model,
            const vector<int8> &x,
            const vector<uint64> &pop_mask = vector<uint64>(),
            const vector<int> &haplo_mask = vector<int>(),
            const vector<int> &haplo_exclude = vector<int>(),
            int max_count = 0)
    {
      Calculate(data, model, x, pop_mask, haplo_mask, haplo_exclude,
                max_count);
    }

    Viterbi(const SNPData &data, const Model &model,
            const vector<int8> &x,
            const vector<uint64> &pop_mask_from,
            const vector<uint64> &pop_mask_to,
            const vector<int> &haplo_mask = vector<int>(),
            const vector<int> &haplo_exclude = vector<int>(),
            int max_count = 0)
    {
      Calculate(data, model, x, pop_mask_from, pop_mask_to,
                haplo_mask, haplo_exclude, max_count);
    }

    void Calculate(const SNPData &data, const Model &model,
                   const vector<int8> &x,
                   const vector<uint64> &pop_mask = vector<uint64>(),
                   const vector<int> &haplo_mask = vector<int>(),
                   const vector<int> &haplo_exclude = vector<int>(),
                   int max_count = 0)
    {
      Calculate(data, model, x, pop_mask, pop_mask, haplo_mask, haplo_exclude,
                max_count);
    }

    void Calculate(const SNPData &data, const Model &model,
                   const vector<int8> &x,
                   const vector<uint64> &pop_mask_from,
                   const vector<uint64> &pop_mask_to,
                   const vector<int> &haplo_mask = vector<int>(),
                   const vector<int> &haplo_exclude = vector<int>(),
                   int max_count_ = 0);

    int operator[](int index) const { return parse[index]; }

    int size() const { return int(parse.size()); }

    void DecodeHaplos(const SNPData &data, vector<int> &haplos) const;
    void DecodePops(const SNPData &data, vector<int8> &pops) const;
  };


  class ViterbiForward
  {
  public:
    vector< StateProfile<Prob> > s;
    vector<double> scale;
    vector<double> log_scale_total;
    double log_likelihood;

    ViterbiForward() { }

    ViterbiForward(const SNPData &data, const Model &model,
                   const vector<int8> &x,
                   const vector<uint64> &pop_mask = vector<uint64>(),
                   const vector<int> &haplo_mask = vector<int>(),
                   const vector<int> &haplo_exclude = vector<int>())
    {
      Calculate(data, model, x, pop_mask, haplo_mask, haplo_exclude);
    }
    
    ViterbiForward(const SNPData &data, const Model &model,
                   const vector<int8> &x,
                   const vector<uint64> &pop_mask_from,
                   const vector<uint64> &pop_mask_to,
                   const vector<int> &haplo_mask = vector<int>(),
                   const vector<int> &haplo_exclude = vector<int>())
    {
      Calculate(data, model, x, pop_mask_from, pop_mask_to,
                haplo_mask, haplo_exclude);
    }

    void Calculate(const SNPData &data, const Model &model,
                   const vector<int8> &x,
                   const vector<uint64> &pop_mask = vector<uint64>(),
                   const vector<int> &haplo_mask = vector<int>(),
                   const vector<int> &haplo_exclude = vector<int>())
    {
      Calculate(data, model, x, pop_mask, pop_mask, haplo_mask, haplo_exclude);
    }

    void Calculate(const SNPData &data, const Model &model,
                   const vector<int8> &x,
                   const vector<uint64> &pop_mask_from,
                   const vector<uint64> &pop_mask_to,
                   const vector<int> &haplo_mask = vector<int>(),
                   const vector<int> &haplo_exclude = vector<int>());
    
    const StateProfile<Prob> &operator[](int index) const { return s[index]; }

    int size() const { return int(s.size()); }
  };


  class ViterbiBackward
  {
  public:
    vector< StateProfile<Prob> > s;
    vector<double> scale;
    vector<double> log_scale_total;
    double log_likelihood;

    ViterbiBackward() { }

    ViterbiBackward(const SNPData &data, const Model &model,
                    const vector<int8> &x,
                    const vector<uint64> &pop_mask = vector<uint64>(),
                    const vector<int> &haplo_mask = vector<int>(),
                    const vector<int> &haplo_exclude = vector<int>())
    {
      Calculate(data, model, x, pop_mask, haplo_mask, haplo_exclude);
    }
    
    ViterbiBackward(const SNPData &data, const Model &model,
                    const vector<int8> &x,
                    const vector<uint64> &pop_mask_from,
                    const vector<uint64> &pop_mask_to,
                    const vector<int> &haplo_mask = vector<int>(),
                    const vector<int> &haplo_exclude = vector<int>())
    {
      Calculate(data, model, x, pop_mask_from, pop_mask_to,
                haplo_mask, haplo_exclude);
    }

    void Calculate(const SNPData &data, const Model &model,
                   const vector<int8> &x,
                   const vector<uint64> &pop_mask = vector<uint64>(),
                   const vector<int> &haplo_mask = vector<int>(),
                   const vector<int> &haplo_exclude = vector<int>())
    {
      Calculate(data, model, x, pop_mask, pop_mask, haplo_mask, haplo_exclude);
    }

    void Calculate(const SNPData &data, const Model &model,
                   const vector<int8> &x,
                   const vector<uint64> &pop_mask_from,
                   const vector<uint64> &pop_mask_to,
                   const vector<int> &haplo_mask = vector<int>(),
                   const vector<int> &haplo_exclude = vector<int>());
    
    const StateProfile<Prob> &operator[](int index) const { return s[index]; }

    int size() const { return int(s.size()); }
  };

};

#endif
