
#ifndef HAPLOMODELGRAPH_H
#define HAPLOMODELGRAPH_H

#include "System.h"
#include "Utility.h"
#include "SNPData.h"
#include "HaploModel.h"

class HaploModelGraph
{
public:
  int type;
  double height;

  HaploModelGraph()
  {
  }

  HaploModelGraph(int type_ = 0, double height_ = 1.0) 
    : type(type_), height(height_)
  {
  }

  virtual void WriteBin(ostream &out) const
  {
    ::WriteBin(out, height);
  }

  virtual void ReadBin(istream &in)
  {
    ::ReadBin(in, height);
  }
};

class HaploModelGraphSpace : public HaploModelGraph
{
public:
  HaploModelGraphSpace(double height_ = 1.0) : HaploModelGraph(3, height_)
  {
  }

  virtual void WriteBin(ostream &out) const
  {
    HaploModelGraph::WriteBin(out);
  }

  virtual void ReadBin(istream &in)
  {
    HaploModelGraph::ReadBin(in);
  }
    
};

class HaploModelGraphLabel : public HaploModelGraph
{
public:
  string label;

  HaploModelGraphLabel(const string &label_ = "") : HaploModelGraph(5), label(label_)
  {
  }

  virtual void WriteBin(ostream &out) const
  {
    HaploModelGraph::WriteBin(out);
    Write(out, label);
  }

  virtual void ReadBin(istream &in)
  {
    HaploModelGraph::ReadBin(in);
    Read(in, label);
  }
    
};

class HaploModelGraphCode : public HaploModelGraph
{
public:
  int palette;
  vector<uint32> colors;

  vector<int> v;

  HaploModelGraphCode() : HaploModelGraph(1, 2.0), palette(1)
  {
  }

  HaploModelGraphCode(const vector<int8> &pops) 
    : HaploModelGraph(1, 2.0), palette(1)
  {
    v.resize(pops.size());
    for (int i = 0; i < pops.size(); ++i)
      v[i] = pops[i];
  }

  HaploModelGraphCode(const SNPData &data,
                      const HaploModel::Posterior &posterior) 
    : HaploModelGraph(1, 2.0), palette(1)
  {
    vector<int8> pops;
    posterior.DecodePops(data, pops);
    v.resize(pops.size());
    for (int i = 0; i < pops.size(); ++i)
      v[i] = pops[i];
  }

  virtual void WriteBin(ostream &out) const
  {
    HaploModelGraph::WriteBin(out);
    ::WriteBin(out, palette);
    ::WriteBin(out, colors);
    ::WriteBin(out, v);
  }

  virtual void ReadBin(istream &in)
  {
    HaploModelGraph::ReadBin(in);
    ::ReadBin(in, palette);
    ::ReadBin(in, colors);
    ::ReadBin(in, v);
  }
};


class HaploModelGraphCodeProb : public HaploModelGraph
{
public:
  int palette;
  vector<uint32> colors;

  vector< vector<float> > p;

  HaploModelGraphCodeProb() : HaploModelGraph(2)
  {
  }

  HaploModelGraphCodeProb(const HaploModel::Posterior &posterior) 
    : HaploModelGraph(2, 4.0), palette(1)
  {
    p.resize(posterior.size());
    for (int i = 0; i < posterior.pop_p.size(); ++i)
    {
      p[i].resize(posterior.pop_p[i].size());
      for (int j = 0; j < posterior.pop_p[i].size(); ++j)
        p[i][j] = float(posterior.pop_p[i][j]);
    }
  }

  template <typename T>
  HaploModelGraphCodeProb(const vector< vector<T> > &init)
    : HaploModelGraph(2, 4.0), palette(1)
  {
    p.resize(init.size());
    for (int i = 0; i < init.size(); ++i)
    {
      p[i].resize(init[i].size());
      for (int j = 0; j < init[i].size(); ++j)
        p[i][j] = float(init[i][j]);
    }
  }

  virtual void WriteBin(ostream &out) const
  {
    HaploModelGraph::WriteBin(out);
    ::WriteBin(out, palette);
    ::WriteBin(out, colors);
    ::Write(out, p);
  }

  virtual void ReadBin(istream &in)
  {
    HaploModelGraph::ReadBin(in);
    ::ReadBin(in, palette);
    ::ReadBin(in, colors);
    ::Read(in, p);
  }
};


class HaploModelGraphFloat : public HaploModelGraph
{
public:
  int palette;
  vector<uint32> colors;

  vector<float> v;
  double min_value, max_value;

  HaploModelGraphFloat() : HaploModelGraph(4)
  {
  }

  template <typename T>
  HaploModelGraphFloat(const vector<T> &init, T max_value_ = T(1), T min_value_ = T(0))
    : HaploModelGraph(4, 2.0), palette(1), max_value(max_value_), min_value(min_value_)
  {
    v.resize(init.size());
    for (int i = 0; i < init.size(); ++i)
      v[i] = float(init[i]);
  }

  virtual void WriteBin(ostream &out) const
  {
    HaploModelGraph::WriteBin(out);
    ::WriteBin(out, palette);
    ::WriteBin(out, colors);
    ::Write(out, v);
    ::WriteBin(out, min_value);
    ::WriteBin(out, max_value);
  }

  virtual void ReadBin(istream &in)
  {
    HaploModelGraph::ReadBin(in);
    ::ReadBin(in, palette);
    ::ReadBin(in, colors);
    ::Read(in, v);
    ::ReadBin(in, min_value);
    ::ReadBin(in, max_value);
  }
};


class HaploModelGraphs
{
public:
  vector<double> pos;
  vector< pair<int, int> > chrom_ranges;

  vector< vector<HaploModelGraph *> > graphs;

  HaploModelGraphs()
  {
  }

  HaploModelGraphs(const SNPProfile &profile, bool genetic = false, int gender = -1)
  {
    pos.resize(profile.seq_len);

    if (!genetic)
    {
      for (int i = 0; i < profile.seq_len; ++i)
        pos[i] = double(profile.loci[i].genomic_pos);
    }
    else
    {
      const vector<double> &profile_pos = (gender < 0) ? profile.locus_genetic_pos : profile.gender_locus_genetic_pos[gender];
      for (int i = 0; i < profile.seq_len; ++i)
        pos[i] = profile_pos[i];
    }

    for (int chrom = 0; chrom < profile.chroms.size(); ++chrom)
    {
      if (profile.chroms[chrom].range.len() > 0)
        chrom_ranges.push_back(pair<int, int>(profile.chroms[chrom].range.left(), profile.chroms[chrom].range.right()));
    }
  }

  void NewGraphs()
  {
    graphs.resize(graphs.size() + 1);
  }

  void AddGraph(HaploModelGraph *graph)
  {
    graphs.back().push_back(graph);
  }
  
  int size() const { return graphs.size(); }
  bool empty() const { return graphs.empty(); }
  
  vector<HaploModelGraph *> &operator[](int index) { return graphs[index]; }
    
  void WriteBin(ostream &out) const
  {
    ::WriteBin(out, pos);
    ::WriteBin(out, chrom_ranges);

    int32 size = graphs.size();
    ::WriteBin(out, size);
    for (int i = 0; i < size; ++i)
    {
      int32 size2 = graphs[i].size();
      ::WriteBin(out, size2);
      for (int j = 0; j < size2; ++j)
      {
        int type = graphs[i][j]->type;
        ::WriteBin(out, type);
        switch (graphs[i][j]->type)
        {
        case 1:
          ((HaploModelGraphCode *)graphs[i][j])->WriteBin(out);
          break;

        case 2:
          ((HaploModelGraphCodeProb *)graphs[i][j])->WriteBin(out);
          break;

        case 3:
          ((HaploModelGraphSpace *)graphs[i][j])->WriteBin(out);
          break;

        case 4:
          ((HaploModelGraphFloat *)graphs[i][j])->WriteBin(out);
          break;

        case 5:
          ((HaploModelGraphLabel *)graphs[i][j])->WriteBin(out);
          break;

        default:
          AssertMsg(false, ToStr(graphs[i][j]->type));
        }
      }
    }
  }

  void ReadBin(istream &in)
  {
    ::ReadBin(in, pos);
    ::ReadBin(in, chrom_ranges);

    int32 size;
    ::ReadBin(in, size);
    graphs.resize(size);
    for (int i = 0; i < size; ++i)
    {
      int32 size2;
      ::ReadBin(in, size2);
      graphs[i].resize(size2);
      for (int j = 0; j < size2; ++j)
      {
        int type;
        ::ReadBin(in, type);
        switch (type)
        {
        case 1:
          graphs[i][j] = new HaploModelGraphCode;
          ((HaploModelGraphCode *)graphs[i][j])->ReadBin(in);
          break;

        case 2:
          graphs[i][j] = new HaploModelGraphCodeProb;
          ((HaploModelGraphCodeProb *)graphs[i][j])->ReadBin(in);
          break;
 
        case 3:
          graphs[i][j] = new HaploModelGraphSpace;
          ((HaploModelGraphSpace *)graphs[i][j])->ReadBin(in);
          break;
 
        case 4:
          graphs[i][j] = new HaploModelGraphFloat;
          ((HaploModelGraphFloat *)graphs[i][j])->ReadBin(in);
          break;
 
        case 5:
          graphs[i][j] = new HaploModelGraphLabel;
          ((HaploModelGraphLabel *)graphs[i][j])->ReadBin(in);
          break;
 
        default:
          AssertMsg(false, ToStr(type));
        }
      }
    }
  }
        
};

#endif
