

#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"


Params params(false);

int debug = 0;
string input_dir, output_dir;
string input_file, snpprofile_file, output_file;
double switch_error = 0.5;


string required_param[] =
{
  "input_file",
  "snpprofile_file",
  "output_file"
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

void InitParams()
{
  params.ExpandImports();

  srand(TimerSeconds());
  if (params.Contains("rand_seed"))
    srand(params["rand_seed"].GetInt());

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("input_file"))
    input_file = params["input_file"].GetString();

  if (params.Contains("snpprofile_file"))
  {
    snpprofile_file = params["snpprofile_file"].GetString();
    string filetype = RemoveCompressedExtension(snpprofile_file);
    Assert(GetExtension(filetype) == "bin");
    filetype = GetExtension(RemoveExtension(filetype));
    Assert((filetype == "snpprofile") || (filetype == "snpprofiledata"));
  }

  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (params.Contains("output_file"))
  {
    output_file = params["output_file"].GetString();
    string filetype = RemoveCompressedExtension(output_file);
    Assert(GetExtension(filetype) == "bin");
    filetype = GetExtension(RemoveExtension(filetype));
    Assert((filetype == "snpdata") || (filetype == "snpprofiledata"));
  }

  if (params.Contains("switch_error"))
    switch_error = params["switch_error"].GetDouble();

  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;

  cerr << "- Input file = " << input_file << endl;
  cerr << "- SNPProfile file = " << snpprofile_file << endl;

  if (!output_dir.empty())
    cerr << "- Output directory = " << output_dir << endl;

  cerr << "- Output file = " << output_file << endl;

  cerr << "- Switch error rate = " << switch_error << endl;
}


int Main(vector<string> args)
{
  cerr << "ImportSNPData" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;

  SNPProfile profile;

  string pathname;
  RelPath(input_dir, snpprofile_file, pathname);
  cerr << "* Reading SNPProfile file " << pathname << " ...";

  istream *in = InFileStream(pathname);
  AssertMsg(in, pathname);
  profile.ReadBin(*in);
  delete in;

  cerr << " done." << endl;
  profile.ComputeDerived();
  profile.Validate();

  map<string, int> coords;
  for (int i = 0; i < profile.seq_len; ++i)
    coords[profile.chroms[profile.loci[i].chrom].label + "." + ToStr(profile.loci[i].genomic_pos)] = i;
  for (int i = 0; i < profile.seq_len; ++i)
    coords[profile.loci[i].label] = i;

  RelPath(input_dir, input_file, pathname);
  cerr << "* Reading input file " << pathname << " ...";

  SNPData data(profile);
  map<string, int> pop_names;
  int num_data = -1;
  vector<int> pops;
  vector<string> labels;
  vector<int> diplo_base, haplo_base;

  in = InFileStream(pathname);
  AssertMsg(in, pathname);
  
  string line;
  vector<string> fields;
  bool first = true;
  while (getline(*in, line))
  {
    SplitFields(line, fields, " \t,");
    if (fields.empty() || (fields[0][0] == ';') || (fields[0][0] == '#'))
      continue;

    if (num_data < 0)
      num_data = fields.size() - 1;
    else
      AssertMsg(num_data == (fields.size() - 1), line);
    
    AssertMsg(fields.size() >= 1, line);
    if (fields[0] == "pops")
    {
      for (int i = 1; i < fields.size(); ++i)
      {
        map<string, int>::const_iterator find = pop_names.find(fields[i]);
        int pop_index;
        if (find == pop_names.end())
        {
          pop_index = pop_names[fields[i]] = data.pops.size();
          data.pops.push_back(SNPData::Pop(fields[i]));
        }
        else
          pop_index = find->second;
        pops.push_back(pop_index);
      }
    }
    else if (fields[0] == "labels")
      labels.insert(labels.end(), fields.begin() + 1, fields.end());
    else
    {
      map<string, int>::const_iterator find = coords.find(fields[0]);
      AssertMsg(find != coords.end(), fields[0]);
      int pos = find->second;

      for (int i = 1; i < fields.size(); ++i)
        AssertMsg((fields[i].length() == 1) || (fields[i].length() == 2), fields[i]);

      if (first)
      {
        first = false;

        diplo_base.resize(num_data, -1);
        haplo_base.resize(num_data, -1);

        for (int i = 1; i < fields.size(); ++i)
        {
          haplo_base[i - 1] = data.haplos.size();
          if (fields[i].length() == 1)
          {
            diplo_base[i - 1] = -1;
            data.haplos.push_back(SNPData::Haplo());
            data.haplos.back().data.resize(profile.seq_len);
          }
          else
          {
            diplo_base[i - 1] = data.diplos.size();
            data.diplos.push_back(SNPData::Diplo());
            data.diplos.back().haplos[0] = data.haplos.size();
            data.diplos.back().haplos[1] = data.haplos.size() + 1;
            data.haplos.push_back(SNPData::Haplo());
            data.haplos.back().data.resize(profile.seq_len);
            data.haplos.push_back(SNPData::Haplo());
            data.haplos.back().data.resize(profile.seq_len);
          }
        }
      }

      for (int i = 1; i < fields.size(); ++i)
      {
        int b = (fields[i][0] == '-') ? 4 : char2base[fields[i][0]];
        AssertMsg(b >= 0, fields[i]);
        
        if (diplo_base[i - 1] < 0)
        {
          AssertMsg(fields[i].length() == 1, fields[i]);
          data.haplos[haplo_base[i - 1]].data[pos] = b;
        }
        else
        {
          AssertMsg(fields[i].length() == 2, fields[i]);
          int b2 = (fields[i][1] == '-') ? 4 : char2base[fields[i][1]];
          AssertMsg(b2 >= 0, fields[i]);
          data.haplos[haplo_base[i - 1] + 0].data[pos] = b;
          data.haplos[haplo_base[i - 1] + 1].data[pos] = b2;
        }
      }
    }
  }

  delete in;
  cerr << " done." << endl;

  Assert(!pops.empty());
  if (labels.empty())
    labels.resize(num_data);

  for (int i = 0; i < num_data; ++i)
  {
    if (diplo_base[i] < 0)
    {
      data.haplos[haplo_base[i]].label = labels[i];
      data.haplos[haplo_base[i]].pop = pops[i];
      data.haplos[haplo_base[i]].sry = 0;
    }
    else
    {
      data.haplos[haplo_base[i] + 0].label = labels[i] + ".0";
      data.haplos[haplo_base[i] + 1].label = labels[i] + ".1";
      data.haplos[haplo_base[i] + 0].pop = pops[i];
      data.haplos[haplo_base[i] + 1].pop = pops[i];
      data.haplos[haplo_base[i] + 0].sry = 0;
      data.haplos[haplo_base[i] + 1].sry = 0;
      data.diplos[diplo_base[i]].label = labels[i];
    }
  }

  data.ComputeDerived();
  data.SetAllDiploSwitch(switch_error);
  data.ComputeDerived();
  data.Validate();

  RelPath(output_dir, output_file, pathname);
  cerr << "* Writing file " << pathname << " ...";
  ostream *out = OutFileStream(pathname);
  AssertMsg(out, pathname);

  string filetype = GetExtension(RemoveExtension(RemoveCompressedExtension(output_file)));
  if (filetype == "snpdata")
    data.WriteBin(*out);
  else
    WriteBinSNPProfileData(*out, data);

  delete out;
  cerr << " done." << endl;
  
  cerr << "Done!" << endl;
  return 0;
}
