
#ifndef INTERVALMAP_H
#define INTERVALMAP_H

#include "System.h"
#include "Utility.h"
#include "Segment.h"

template <typename T, typename ST = int>
class IntervalMap;

template <typename T, typename ST = int>
class IntervalMapRange;

template <typename T, typename ST>
class IntervalMap
{
private:
  struct SegmentTouchLess
  {
    bool operator()(const TSegment<ST> &a, const TSegment<ST> &b) const
    {
      return a.right() <= b.left();
    }
  };
  
public:
  typedef map< TSegment<ST>, T, SegmentTouchLess > MapType;

  typedef typename MapType::iterator iterator;
  typedef typename MapType::const_iterator const_iterator;
    
  IntervalMap() { }

  int size() const { return int(m.size()); }

  iterator begin() { return m.begin(); }
  const_iterator begin() const { return m.begin(); }

  iterator end() { return m.end(); }
  const_iterator end() const { return m.end(); }

  iterator lower_bound(const TSegment<ST> &range) { return m.lower_bound(range); }
  const_iterator lower_bound(const TSegment<ST> &range) const { return m.lower_bound(range); }

  iterator lower_bound(ST x) { return lower_bound(TSegment<ST>(x, x + 1, true)); }
  const_iterator lower_bound(ST x) const { return lower_bound(TSegment<ST>(x, x + 1, true)); }

  iterator upper_bound(const TSegment<ST> &range) { return m.upper_bound(range); }
  const_iterator upper_bound(const TSegment<ST> &range) const { return m.upper_bound(range); }

  iterator upper_bound(ST x) { return upper_bound(TSegment<ST>(x, x + 1, true)); }
  const_iterator upper_bound(ST x) const { return upper_bound(TSegment<ST>(x, x + 1, true)); }


  iterator find(ST x)
  {
    TSegment<ST> dummy(x, x + 1, true);
    iterator iter = m.lower_bound(x);
    if ((iter != m.end()) && (iter->first.contains(x)))
      return iter;
    else
      return m.end();
  }

  const_iterator find(ST x) const
  {
    TSegment<ST> dummy(x, x + 1, true);
    const_iterator iter = m.lower_bound(x);
    if ((iter != m.end()) && (iter->first.contains(x)))
      return iter;
    else
      return m.end();
  }

  IntervalMapRange<T, ST> operator[](const TSegment<ST> &range)
  {
    //cerr << "[" << range.left() << ", " << range.right() << ")" << endl;
    return IntervalMapRange<T, ST>(*this, range);
  }

  void Write(ostream &o) const
  {
    ::Write(o, m);
  }

  void Read(istream &i)
  {
    ::Read(i, m);
  }

  void swap(IntervalMap<T> &other)
  {
    m.swap(other.m);
  }

  void clear()
  {
    m.clear();
  }

private:
  friend class IntervalMapRange<T, ST>;

  MapType m;

  typename MapType::iterator Split(typename MapType::iterator iter, ST x)
  {
    TSegment<ST> range = iter->first;
    T value = iter->second;
    m.erase(iter);
    m.insert(pair< TSegment<ST>, T >(TSegment<ST>(range.left(), x, true), value));
    return m.insert(pair< TSegment<ST>, T >(TSegment<ST>(x, range.right(), true), value)).first;
  }

  typename MapType::iterator Split(ST x)
  {
    TSegment<ST> dummy(x, x + 1, true);
    typename MapType::iterator iter = m.lower_bound(dummy);
    if ((iter != m.end()) && (x > iter->first.left()) && (x < iter->first.right()))
      return Split(iter, x);
    else
      return iter;
  }
};

template <typename T, typename ST>
class IntervalMapRange
{
public:
  const IntervalMapRange<T, ST> &operator=(const T &value)
  {
    typename IntervalMap<T, ST>::MapType::iterator left = im.m.lower_bound(range);
    typename IntervalMap<T, ST>::MapType::iterator right = im.m.upper_bound(range);

    ST left_overhang = range.left(), right_overhang = range.right();
    T left_overhang_value, right_overhang_value;
    ST seg_left = range.left(), seg_right = range.right();
    
    if (left != right)
    {
      if (left->first.left() < range.left())
      {
        if (left->second == value)
          seg_left = left->first.left();
        else
        {
          left_overhang = left->first.left();
          left_overhang_value = left->second;
        }
      }
      typename IntervalMap<T, ST>::MapType::iterator rright = right;
      --rright;
      if (rright->first.right() > range.right())
      {
        if (rright->second == value)
          seg_right = rright->first.right();
        else
        {
          right_overhang = rright->first.right();
          right_overhang_value = rright->second;
        }
      }
    }

    im.m.erase(left, right);
    if (left_overhang < seg_left)    
      im.m.insert(right, pair< TSegment<ST>, T >(TSegment<ST>(left_overhang, seg_left, true), left_overhang_value));
    im.m.insert(right, pair< TSegment<ST>, T >(TSegment<ST>(seg_left, seg_right, true), value));
    if (right_overhang > seg_right)
      im.m.insert(right, pair< TSegment<ST>, T >(TSegment<ST>(seg_right, right_overhang, true), right_overhang_value));

    return *this;
  }

  const IntervalMapRange<T, ST> &operator%=(const T &value)
  {
    typename IntervalMap<T, ST>::MapType::iterator right = im.Split(range.right());
    typename IntervalMap<T, ST>::MapType::iterator left = im.Split(range.left());
    for ( ; left != right; ++left)
      left->second = value;
    return *this;
  }

  const IntervalMapRange<T, ST> &operator+=(const T &delta)
  {
    typename IntervalMap<T, ST>::MapType::iterator right = im.Split(range.right());
    typename IntervalMap<T, ST>::MapType::iterator left = im.Split(range.left());

    int x;
    for (x = range.left(); left != right; x = left->first.right(), ++left)
    {
      if (left->first.left() > x)
        im.m.insert(pair< TSegment<ST>, T >(TSegment<ST>(x, left->first.left(), true), delta));
      left->second += delta;
    }
    if (x < range.right())
      im.m.insert(pair< TSegment<ST>, T >(TSegment<ST>(x, range.right(), true), delta));
    return *this;
  }

  const IntervalMapRange<T, ST> &operator-=(const T &delta)
  {
    return operator+=(-delta);
  }

  const IntervalMapRange<T, ST> &operator++()
  {
    return operator+=(T(1));
  }

  const IntervalMapRange<T, ST> &operator--()
  {
    return operator-=(T(1));
  }

private:
  friend class IntervalMap<T, ST>;

  IntervalMapRange(IntervalMap<T, ST> &im_, const TSegment<ST> &range_)
    : im(im_), range(range_)
  { 
  }

  IntervalMap<T, ST> &im;
  TSegment<ST> range;
};

typedef IntervalMap<int> IntIntervalMap;
typedef IntervalMap<bool> BoolIntervalMap;


template <typename T, typename ST>
void Write(ostream &o, const IntervalMap<T, ST> &x)
{
  x.Write(o);
}

template <typename T, typename ST>
void Read(istream &i, IntervalMap<T, ST> &x)
{
  x.Read(i);
}

#endif
