
#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"


Params params(false);

int debug = 0;

string input_dir;
string snpprofile_file;
vector<string> input_files;
vector<bool> input_profile;
string output_dir;
string output_file;
bool output_profile;
int merge_mode = -1;
vector<string> pop_labels;

string required_param[] =
{
  "input_files",
  "output_file"
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

void InitParams()
{
  params.ExpandImports();

  uint32 seed = RandomSeed();
  if (params.Contains("rand_seed"))
    seed = params["rand_seed"].GetInt();
  srand(seed);
  cerr << "- Random seed = " << seed << endl;

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("snpprofile_file"))
  {
    snpprofile_file = params["snpprofile_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(snpprofile_file)) == "bin");
    string filetype = GetExtension(RemoveExtension(RemoveCompressedExtension(snpprofile_file)));
    Assert((filetype == "snpprofile") || (filetype == "snpprofiledata"));
  }

  if (params.Contains("input_files"))
  {
    vector<ParamValue> values = params["input_files"].GetList();
    for (int i = 0; i < values.size(); ++i)
    {
      string filename = values[i].GetString();
      Assert(GetExtension(RemoveCompressedExtension(filename)) == "bin");
      string filetype = GetExtension(RemoveExtension(RemoveCompressedExtension(filename)));
      Assert((filetype == "snpdata") || (filetype == "snpprofiledata"));
      input_files.push_back(filename);
      input_profile.push_back(filetype == "snpprofiledata");
    }
  }

  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (params.Contains("output_file"))
  {
    output_file = params["output_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(output_file)) == "bin");
    string file_type = GetExtension(RemoveExtension(RemoveCompressedExtension(output_file)));
    if (file_type == "snpdata")
      output_profile = false;
    else if (file_type == "snpprofiledata")
      output_profile = true;
    else
      AssertMsg(false, "Neither snpdata nor snpprofiledata!");
  }

  if (!snpprofile_file.empty())
  {
    bool found_one = false;
    for (int i = 0; i < input_profile.size(); ++i)
      if (input_profile[i])
        found_one = true;
    Assert(found_one);
  }
  
  if (params.Contains("concat"))
  {
    Assert(merge_mode < 0);
    merge_mode = 0;
  }
  if (params.Contains("merge_pop_index"))
  {
    Assert(merge_mode < 0);
    merge_mode = 1;
  }
  if (params.Contains("merge_pop_label"))
  {
    Assert(merge_mode < 0);
    merge_mode = 2;
  }
  if (merge_mode < 0)
  {
    cerr << "No merge mode specified! Options are: concat, merge_pop_index, merge_pop_label" << endl;
    Exit(1);
  }

  if (params.Contains("pop_labels"))
  {
    vector<ParamValue> values = params["pop_labels"].GetList();
    pop_labels.resize(values.size());
    for (int i = 0; i < values.size(); ++i)
      pop_labels[i] = values[i].GetString();
  }

  switch (merge_mode)
  {
  case 0:
    cerr << "- Concatenating populations" << endl;
    break;
  case 1:
    cerr << "- Merging populations on index" << endl;
    break;
  case 2:
    cerr << "- Merging populations on label" << endl;
    break;
  }
  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;
  if (!snpprofile_file.empty())
    cerr << "- SNP Profile file = " << snpprofile_file << endl;
  cerr << "- Input files = {";
  for (int i = 0; i < input_files.size(); ++i)
    cerr << " " << input_files[i];
  cerr << " }" << endl;
  if (!output_dir.empty())
    cerr << "- Output directory = " << output_dir << endl;
  cerr << "- Output file = " << output_file << endl;
  if (output_profile)
    cerr << "- Outputing profile & data" << endl;
  else
    cerr << "- Outputing data only" << endl; 

  if (!pop_labels.empty())
  {
    cerr << "- Relabel populations to {";
    for (int i = 0; i < pop_labels.size(); ++i)
      cerr << " " << pop_labels[i];
    cerr << " }" << endl;
  }
}

int Main(vector<string> args)
{
  cerr << "MergeSNPData" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;

  SNPProfile profile;
  SNPData data;
  if (!snpprofile_file.empty())
  {
    istream *in = InFileStream(RelPath(input_dir, snpprofile_file));
    if (!in)
    {
      cerr << "Could not open file \"" << snpprofile_file;
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"";
      cerr << endl;
      Exit(1);
    }
    
    cerr << "* Reading SNP profile...";
    profile.ReadBin(*in);
    cerr << " done." << endl;
    delete in;
    
    profile.Validate();
    data.SetProfile(profile);
  }
  else
  {
    for (int i = 0; i < input_profile.size(); ++i)
      if (input_profile[i])
      {
        istream *in = InFileStream(RelPath(input_dir, input_files[i]));
        if (!in)
        {
          cerr << "Could not open file \"" << input_files[i];
          if (!input_dir.empty())
            cerr << " in directory \"" << input_dir << "\"";
          cerr << endl;
          Exit(1);
        }
        
        cerr << "* Reading SNP profile from \"" << input_files[i] << "\"...";
        profile.ReadBin(*in);
        data.SetProfile(profile);
        delete in;
        cerr << " done." << endl;
        break;
      }
  }

  for (int i = 0; i < input_files.size(); ++i)
  {
    cerr << "* Reading \"" << input_files[i] << "\"...";
    SNPProfile temp_profile(profile);
    SNPData temp_data;
    
    istream *in = InFileStream(RelPath(input_dir, input_files[i]));
    if (!in)
    {
      cerr << "Could not open file \"" << input_files[i];
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"";
      cerr << endl;
      Exit(1);
    }
    
    if (input_profile[i])
    {
      ReadBinSNPProfileData(*in, temp_profile, temp_data);
      Assert(profile.Compatible(temp_profile));
    }
    else
      temp_data.ReadBin(*in);

    delete in;
    cerr << " done." << endl;

    temp_data.Validate();

    cerr << "* Merging...";
    switch (merge_mode)
    {
    case 0:
      data.Concat(temp_data);
      break;
    case 1:
      data.MergePopIndex(temp_data);
      break;
    case 2:
      data.MergePopLabel(temp_data);
      break;
    }
    cerr << " done." << endl;
  }

  data.ComputeDerived();
  data.Validate();

  if (!pop_labels.empty())
  {
    Assert(pop_labels.size() == data.num_pops);
    for (int pop = 0; pop < data.num_pops; ++pop)
      data.pops[pop].label = pop_labels[pop];
  }

  cerr << "* Writing output...";
  {
    ostream *out = OutFileStream(RelPath(output_dir, output_file));
    AssertMsg(out, output_file);

    if (output_profile)
      WriteBinSNPProfileData(*out, data);
    else
      data.WriteBin(*out);

    delete out;
  }
  cerr << " done." << endl;

  cerr << "Done!" << endl;
  return 0;
}
