
#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"


Params params(false);

int debug = 0;

string input_dir;
string sample_file;
string input_file;
string output_dir;
string output_file;
string phaseerrors_file;
int max_kmer_len = 15;
double switch_error_rate = 0.1;
bool output_phase = false;


string required_param[] =
{
  "sample_file",
  "input_file",
  "output_file",
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

void InitParams()
{
  params.ExpandImports();

  uint32 seed = RandomSeed();
  if (params.Contains("rand_seed"))
    seed = params["rand_seed"].GetInt();
  srand(seed);
  cerr << "- Random seed = " << seed << endl;

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (params.Contains("sample_file"))
  {
    sample_file = params["sample_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(sample_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(sample_file))) == "snpprofiledata");
  }

  if (params.Contains("input_file"))
  {
    input_file = params["input_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(input_file)) == "bin");
    Assert((GetExtension(RemoveExtension(RemoveCompressedExtension(input_file))) == "snpprofiledata") || (GetExtension(RemoveExtension(RemoveCompressedExtension(input_file))) == "snpdata"));
  }

  if (params.Contains("output_file"))
  {
    output_file = params["output_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(output_file)) == "bin");
    Assert((GetExtension(RemoveExtension(RemoveCompressedExtension(output_file))) == "snpprofiledata") || (GetExtension(RemoveExtension(RemoveCompressedExtension(output_file))) == "snpdata") || (GetExtension(RemoveExtension(RemoveCompressedExtension(output_file))) == "phases"));
    if (GetExtension(RemoveExtension(RemoveCompressedExtension(output_file))) == "phases")
      output_phase = true;
  }

  if (params.Contains("phaseerrors_file"))
  {
    phaseerrors_file = params["phaseerrors_file"].GetString();    
    Assert(GetExtension(RemoveCompressedExtension(phaseerrors_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(phaseerrors_file))) == "phaseerrors");
  }

  if (params.Contains("max_kmer_len"))
    max_kmer_len = params["max_kmer_len"].GetInt();

  if (params.Contains("switch_error_rate"))
    switch_error_rate = params["switch_error_rate"].GetDouble();

  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;

  cerr << "- Sample file = " << sample_file << endl;

  cerr << "- Input file = " << input_file << endl;

  cerr << "- Output file = " << output_file << endl;

  if (!phaseerrors_file.empty())
    cerr << "- Phase errors file = " << phaseerrors_file << endl;

  cerr << "- Maximum kmer length = " << max_kmer_len << endl;

  cerr << "- Output switch error rate = " << (100.0 * switch_error_rate) << "%" << endl;
}


void Phase(const SNPData &data, const vector< pair<int8, int8> > &genotypes, 
           vector< vector<int8> > &result, vector<uint8> &switches,
           int max_kmer_len)
{
  Assert(genotypes.size() == data.seq_len);
  Assert(result.size() == 2);
  result[0].resize(data.seq_len);
  result[1].resize(data.seq_len);
  switches.resize(data.seq_len - 1);

  result[0][0] = genotypes[0].first;
  result[1][0] = genotypes[0].second;

  int counts[2][2];
  vector<int> parity(max_kmer_len + 1, 0);
  vector<double> rel_diff(max_kmer_len + 1, 0.0);
  vector<double> prob(max_kmer_len + 1, 0.0);
  vector<bool> candidate(max_kmer_len + 1, false);
  vector<bool> match[2];

  for (int i = 1; i < data.seq_len; ++i)
  {
    if ((i % 10000) == 0)
      cerr << ".";

    if (genotypes[i].first == genotypes[i].second)
    {
      result[0][i] = result[1][i] = genotypes[i].first;
      switches[i - 1] = SNPData::switch_encode(0.0);
      continue;
    }
    
    for (int p = 0; p < 2; ++p)
    {
      match[p].clear();
      match[p].resize(data.num_haplos, true);
    }

    int cur_max_kmer_len = min(i, max_kmer_len);
    for (int len = 1; len <= cur_max_kmer_len; ++len)
    {
      counts[0][0] = counts[0][1] = counts[1][0] = counts[1][1] = 0;
      for (int p = 0; p < 2; ++p)
      {
        for (int k = 0; k < data.num_haplos; ++k)
        {
          if (!match[p][k])
            continue;
          if ((data.haplos[k][i - len] != result[p][i - len]) &&
              (data.haplos[k][i - len] != 4) && (result[p][i - len] != 4))
          {
            match[p][k] = false;
            continue;
          }
          
          if (data.haplos[k][i] == genotypes[i].first)
            ++counts[p][0];
          if (data.haplos[k][i] == genotypes[i].second)
            ++counts[p][1];
        }
      }

      int p0 = counts[0][0] * counts[1][1];
      int p1 = counts[0][1] * counts[1][0];
      if ((p0 == 0) && (p1 == 0))
        candidate[len] = false;
      else
      {
        candidate[len] = true;
        rel_diff[len] = double(abs(p0 - p1)) / double(p0 + p1);
        parity[len] = (p0 >= p1) ? 0 : 1;
        if (p0 >= p1)
          prob[len] = double(p0) / double(p0 + p1);
        else
          prob[len] = double(p1) / double(p0 + p1);
      }
    }
    
    int best_len = 0;
    for (int len = 1; len <= cur_max_kmer_len; ++len)
      if (candidate[len])
        if ((best_len == 0) || (rel_diff[len] > rel_diff[best_len]))
          best_len = len;

    if (best_len == 0)
    {
      result[0][i] = genotypes[i].first;
      result[1][i] = genotypes[i].second;
      switches[i - 1] = SNPData::switch_encode(0.5);
    }
    else
    {
      int d = parity[best_len];
      result[d][i] = genotypes[i].first;
      result[1 - d][i] = genotypes[i].second;
      switches[i - 1] = SNPData::switch_encode(max(prob[best_len], switch_error_rate));
    }
  }
}


void ComputePhase(const vector<int8> &true0, const vector<int8> &true1,
                  const vector<int8> &result0, const vector<int8> &result1,
                  vector<int8> &phase)
{
  int seq_len = true0.size();
  Assert(true0.size() == seq_len);
  Assert(true1.size() == seq_len);
  Assert(result0.size() == seq_len);
  Assert(result1.size() == seq_len);
  phase.clear();
  phase.resize(seq_len, -1);

  int i;
  for (i = 0; i < seq_len; ++i)
    if (true0[i] != true1[i])
      break;
  
  int p = (result0[i] == true0[i]) ? 0 : 1;
  for (int j = i; j >= 0; --j)
    phase[j] = p;

  for (++i; i < seq_len; ++i)
  {
    if ((p ? result1 : result0)[i] != true0[i])
      p = 1 - p;
    phase[i] = p;
  }
}
  


double Check(const vector<int8> &true0, const vector<int8> &true1,
             const vector<int8> &result0, const vector<int8> &result1)
{
  int seq_len = true0.size();
  Assert(true0.size() == seq_len);
  Assert(true1.size() == seq_len);
  Assert(result0.size() == seq_len);
  Assert(result1.size() == seq_len);

  int total_count = seq_len - 1;
  int hetero_count = 0;
  for (int i = 0; i < (seq_len - 1); ++i)
    if ((true0[i] != true1[i]) || (true0[i] == 4) || (true1[i] == 4))
      ++hetero_count;

  int switch_count = 0;
  {
    int i;
    for (i = 0; i < seq_len; ++i)
      if ((true0[i] != true1[i]) || (true0[i] == 4) || (true1[i] == 4))
        break;

    int p = (result0[i] == true0[i]) ? 0 : 1;

    for (++i; i < seq_len; ++i)
      if ((true0[i] != true1[i]) || (true0[i] == 4) || (true1[i] == 4))
        if ((p ? result1 : result0)[i] != true0[i])
        {
          p = 1 - p;
          ++switch_count;
        }
  }

  if (hetero_count == 0)
    return 0.0;
  else
    return double(switch_count) / double(hetero_count);
}


int Main(vector<string> args)
{
  cerr << "PhaseIncremental" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;

  SNPProfile profile;
  SNPData data;
  {
    string filename;
    istream *in;
    if (!input_dir.empty())
      in = InFileStream(RelPath(input_dir, sample_file));
    else
      in = InFileStream(sample_file);

    if (!in)
    {
      cerr << "Could not open file \"" << sample_file << "\"";
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"";
      cerr << endl;
      Exit(1);
    }

    cerr << "* Reading input data...";
    ReadBinSNPProfileData(*in, profile, data);
    cerr << " done." << endl;
    delete in;

    profile.Validate();
    data.Validate();
  }

  SNPData input;
  {
    string filename;
    istream *in = NULL;
    in = InFileStream(RelPath(input_dir, input_file));

    if (!in)
    {
      cerr << "Could not open file \"" << input_file << "\"";
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"";
      cerr << endl;
      Exit(1);
    }

    cerr << "* Reading input data...";
    if (GetExtension(RemoveExtension(RemoveCompressedExtension(input_file))) == "snpprofiledata")
    {
      SNPProfile temp_profile;
      ReadBinSNPProfileData(*in, temp_profile, input);
      Assert(profile.Compatible(profile));
      input.SetProfile(profile);
    }
    else
    {
      input.SetProfile(profile);
      input.ReadBin(*in);
    }
             
    delete in;
    cerr << " done." << endl;

    input.Validate();
  }

  SNPData output(profile);
  output.pops = input.pops;
  output.ReserveHaplos(input.num_haplos);
  output.ReserveDiplos(input.num_diplos);

  cerr << "* Adding unpaired haplotypes...";
  for (int k = 0; k < input.num_haplos; ++k)
    if (input.haplos[k].diplo < 0)
      output.AddHaplo(input, k);
  cerr << " done." << endl;

  vector< pair<int8, int8> > genotypes(data.seq_len);
  vector< vector<int8> > result(2, vector<int8>(data.seq_len));
  vector<int8> phase(data.seq_len);
  vector< vector<int8> > phases;
  vector< vector<uint8> > phaseerrors;
  vector<uint8> switches;
  phases.reserve(input.num_diplos);
  phaseerrors.reserve(input.num_diplos);

  cerr << "* Phasing " << input.num_diplos << " genotypes:" << endl;
  for (int d = 0; d < input.num_diplos; ++d)
  {
    cerr << "- " << d << ":";
    int h0 = input.diplos[d].haplos[0];
    int h1 = input.diplos[d].haplos[1];
    for (int i = 0; i < data.seq_len; ++i)
    {
      genotypes[i].first = min(input.haplos[h0][i], input.haplos[h1][i]);
      genotypes[i].second = max(input.haplos[h0][i], input.haplos[h1][i]);
    }
    
    cerr << " Phasing...";
    Phase(data, genotypes, result, switches, max_kmer_len);
    ComputePhase(input.haplos[h0].data, input.haplos[h1].data,
                 result[0], result[1], phase);

    cerr << " Checking... ";
    double true_switch_error_rate = 
      Check(input.haplos[h0].data, input.haplos[h1].data,
            result[0], result[1]);
    cerr << " switch error = " << (100.0 * true_switch_error_rate) << "%";

    if (!output_phase)
      output.AddDiplo(result[0], result[1], switch_error_rate, 
                      input.diplos[d].pop, input.diplos[d].label,
                      input.haplos[h0].sry, input.haplos[h1].sry);
    else
      phases.push_back(phase);

    if (!phaseerrors_file.empty())
      phaseerrors.push_back(switches);
                                
    cerr << endl;
  }

  if (!output_phase)
  {
    output.ComputeDerived();
    output.Validate();
  }

  {
    cerr << "* Writing output...";
    ostream *out = OutFileStream(RelPath(output_dir, output_file));
    AssertMsg(out, RelPath(output_dir, output_file));

    if (GetExtension(RemoveExtension(RemoveCompressedExtension(output_file))) == "snpprofiledata")
      WriteBinSNPProfileData(*out, output);
    else if (GetExtension(RemoveExtension(RemoveCompressedExtension(output_file))) == "snpdata")
      output.WriteBin(*out);
    else if (GetExtension(RemoveExtension(RemoveCompressedExtension(output_file))) == "phases")
      Write(*out, phases);

    delete out;
    cerr << " done." << endl;
  }

  if (!phaseerrors_file.empty())
  {
    cerr << "* Writing phase errors...";
    ostream *out = OutFileStream(RelPath(output_dir, phaseerrors_file));
    AssertMsg(out, RelPath(output_dir, phaseerrors_file));
    WriteBin(*out, phaseerrors);
    delete out;
    cerr << " done." << endl;
  }

  cerr << "Done!" << endl;
  return 0;
}
