
#include "System.h"
#include "PhasingError.h"


void PhasingError::Calculate(const SNPData &data, 
                             const vector<int8> &x0,
                             const HaploModel::Forward &forward0,
                             const HaploModel::Backward &backward0,
                             const vector<int8> &x1,
                             const HaploModel::Forward &forward1,
                             const HaploModel::Backward &backward1,
                             const vector<uint8> &x_switch_prior)
{
  Assert(x0.size() >= data.seq_len);
  Assert(x1.size() >= data.seq_len);
  Assert(forward0.size() >= data.seq_len);
  Assert(forward1.size() >= data.seq_len);
  Assert(backward0.size() >= data.seq_len);
  Assert(backward1.size() >= data.seq_len);
  Assert(x_switch_prior.size() >= (data.seq_len - 1));

  x_switch.clear();
  x_switch.resize(data.seq_len - 1, 0.0);
  double prob[2];
  for (int i = 0; i < (data.seq_len - 1); ++i)
  {
    prob[0] = CalculateSwitchSum(data, forward0[i], backward0[i]) * CalculateSwitchSum(data, forward1[i], backward1[i]);
    prob[1] = CalculateSwitchSum(data, forward0[i], backward1[i]) * CalculateSwitchSum(data, forward1[i], backward0[i]);

    double switch_prior = SNPData::switch_decode(x_switch_prior[i]);
    double p_switch = prob[1] * switch_prior;
    double n_switch = prob[0] * (1.0 - switch_prior);
    x_switch[i] = float(p_switch / (p_switch + n_switch));
  }
}


double PhasingError::CalculateSwitchSum(const SNPData &data,
                                        const HaploModel::StateProfile<HaploModel::Prob> &s0,
                                        const HaploModel::StateProfile<HaploModel::Prob> &s1)
{
  double sum = 0.0;
  for (int k = 0; k < data.num_haplos; ++k)
    sum += s0.O(k) * s1.O(k);
  return sum;
}


void PhasingError::Calculate(const SNPData &data, 
                             const vector<int8> &x0,
                             const HaploModel::ViterbiForward &forward0,
                             const HaploModel::ViterbiBackward &backward0,
                             const vector<int8> &x1,
                             const HaploModel::ViterbiForward &forward1,
                             const HaploModel::ViterbiBackward &backward1,
                             const vector<uint8> &x_switch_prior)
{
  Assert(x0.size() >= data.seq_len);
  Assert(x1.size() >= data.seq_len);
  Assert(forward0.size() >= data.seq_len);
  Assert(forward1.size() >= data.seq_len);
  Assert(backward0.size() >= data.seq_len);
  Assert(backward1.size() >= data.seq_len);
  Assert(x_switch_prior.size() >= (data.seq_len - 1));

  x_switch.clear();
  x_switch.resize(data.seq_len - 1, 0.0);
  double prob[2];
  for (int i = 0; i < (data.seq_len - 1); ++i)
  {
    prob[0] = CalculateSwitchMax(data, forward0[i], backward0[i]) * CalculateSwitchMax(data, forward1[i], backward1[i]);
    prob[1] = CalculateSwitchMax(data, forward0[i], backward1[i]) * CalculateSwitchMax(data, forward1[i], backward0[i]);

    double switch_prior = SNPData::switch_decode(x_switch_prior[i]);
    double p_switch = prob[1] * switch_prior;
    double n_switch = prob[0] * (1.0 - switch_prior);
    x_switch[i] = float(p_switch / (p_switch + n_switch));
  }
}

double PhasingError::CalculateSwitchMax(const SNPData &data,
                                        const HaploModel::StateProfile<HaploModel::Prob> &s0,
                                        const HaploModel::StateProfile<HaploModel::Prob> &s1)
{
  int max_k = -1;
  double max_prob = 0.0;
  for (int k = 0; k < data.num_haplos; ++k)
  {
    double prob = s0.O(k) * s1.O(k);
    if ((max_k < 0) || (prob > max_prob))
      max_k = k, max_prob = prob;
  }

  return max_prob;
}


