

#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"
#include "Segment.h"
#include "IntervalMap.h"


Params params(false);

int debug = 0;
string input_dir, output_dir;
string input_file, output_file;
string illumina_file = "Hap550v3_561466_SNPs";


string required_param[] =
{
  "input_file",
  "output_file"
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

void InitParams()
{
  params.ExpandImports();

  srand(TimerSeconds());
  if (params.Contains("rand_seed"))
    srand(params["rand_seed"].GetInt());

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("input_file"))
    input_file = params["input_file"].GetString();

  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (params.Contains("output_file"))
    output_file = params["output_file"].GetString();

  if (params.Contains("illumina_file"))
    illumina_file = params["illumina_file"].GetString();

  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;

  cerr << "- Input file = " << input_file << endl;

  if (!output_dir.empty())
    cerr << "- Output directory = " << output_dir << endl;

  cerr << "- Output file = " << output_file << endl;

  cerr << "- Illumina SNP file = " << illumina_file << endl;

}


void RestrictSites(const SNPProfile &profile, const SNPData &data,
                   string filename,
                   SNPProfile &new_profile, SNPData &new_data)
{
  string pathname;
  RelPath(input_dir, filename, pathname);
  cerr << "Reading restriction sites file " << pathname << " ...";
  istream *in = InFileStream(pathname);
  AssertMsg(in, pathname);
  
  string line;
  Assert(getline(*in, line));
  vector<string> fields;
  vector<string> sites;
  sites.reserve(1000000);
  while (getline(*in, line))
  {
    SplitFields(line, fields, " \t\n\r");
    if (fields.empty())
      continue;
    sites.push_back(fields[0]);
  }
  delete in;
  cerr << " done." << endl;

  cerr << "Restricting sites...";

  sort(sites.begin(), sites.end());

  vector< pair<string, int> > labels;
  labels.reserve(profile.seq_len);
  for (int i = 0; i < profile.seq_len; ++i)
    labels.push_back(pair<string, int>(profile.loci[i].label, i));
  sort(labels.begin(), labels.end(), FirstLess<string, int>());

  vector<int> select(profile.seq_len, 0);
  for (int i = 0, j = 0; (i < labels.size()) && (j < sites.size()); )
  {
    if (labels[i].first == sites[j])
    {
      select[labels[i++].second] = 1;
      ++j;
    }
    else if (labels[i].first < sites[j])
      ++i;
    else
      ++j;
  }

  new_profile.CopyRestrictLoci(profile, select);
  new_data.SetProfile(new_profile);
  new_data.CopyRestrictLoci(data, select);
  cerr << " done." << endl;
  new_profile.Validate();
  new_data.Validate();
}


int Main(vector<string> args)
{
  cerr << "RestrictIlluminaSites" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;

  SNPProfile profile;
  SNPData data;

  string pathname;
  RelPath(input_dir, input_file, pathname);
  cerr << "Reading " << pathname << " ...";
  ReadBinSNPProfileData(pathname, profile, data);
  cerr << " done." << endl;

  profile.Validate();
  data.Validate();

  SNPProfile rprofile;
  SNPData rdata;
  RestrictSites(profile, data, illumina_file, rprofile, rdata);

  RelPath(output_dir, output_file, pathname);
  cerr << "Writing " << pathname << " ...";
  WriteBinSNPProfileData(pathname, rdata);
  cerr << " done." << endl;
  
  cerr << "Done!" << endl;
  return 0;
}
