
#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"


Params params(false);

int debug = 0;

string input_dir;
string input_profile_file;
string input_file;
string output_dir;
string output_profile_file;
string output_file;

vector<int> loci;
vector<int> chroms;
vector<int> pops;
vector<string> pop_labels;
vector<int> diplos;
vector<int> haplos;
vector<int> pop_diplos;
vector<int> pop_haplos;
vector<int> gender;
vector<int> map_pops;

bool restrict_pop_missing = false;
bool restrict_monomorphic = false;

string required_param[] =
{
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

string BinFileType(string filename)
{
  Assert(GetExtension(RemoveCompressedExtension(filename)) == "bin");
  return GetExtension(RemoveExtension(RemoveCompressedExtension(filename)));
}

void GetIntegerSpec(string s, vector<int> &result);

void InitParams()
{
  params.ExpandImports();

  uint32 seed = RandomSeed();
  if (params.Contains("rand_seed"))
    seed = params["rand_seed"].GetInt();
  srand(seed);
  cerr << "- Random seed = " << seed << endl;

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("input_profile_file"))
  {
    input_profile_file = params["input_profile_file"].GetString();
    Assert((BinFileType(input_profile_file) == "snpprofile") || (BinFileType(input_profile_file) == "snpprofiledata"));
  }

  if (params.Contains("input_file"))
  {
    input_file = params["input_file"].GetString();
    if (BinFileType(input_file) == "snpdata")
      Assert(!input_profile_file.empty());
    else
    {
      Assert(input_profile_file.empty());
      Assert(BinFileType(input_file) == "snpprofiledata");
    }
  }
  
  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (params.Contains("output_file"))
  {
    output_file = params["output_file"].GetString();
    Assert((BinFileType(output_file) == "snpprofiledata") || (BinFileType(output_file) == "snpdata"));
  }
  
  if (params.Contains("output_profile_file"))
  {
    output_profile_file = params["output_profile_file"].GetString();
    Assert(BinFileType(output_profile_file) == "snpprofile");
  }

  Assert(!input_profile_file.empty() || !input_file.empty());
  Assert(!output_profile_file.empty() || !output_file.empty());
  if (!input_file.empty())
    Assert(!output_file.empty());

  if (params.Contains("loci"))
    GetIntegerSpec(params["loci"].GetString(), loci);

  if (params.Contains("chroms"))
  {
    vector<ParamValue> values = params["chroms"].GetList();
    chroms.resize(values.size());
    for (int i = 0; i < values.size(); ++i)
      chroms[i] = values[i].GetInt();
  }

  if (params.Contains("pops"))
  {
    vector<ParamValue> values = params["pops"].GetList();
    pops.resize(values.size());
    for (int i = 0; i < values.size(); ++i)
      pops[i] = values[i].GetInt();
  }

  if (params.Contains("pop_labels"))
  {
    vector<ParamValue> values = params["pop_labels"].GetList();
    pop_labels.resize(values.size());
    for (int i = 0; i < values.size(); ++i)
      pop_labels[i] = values[i].GetString();
  }

  if (params.Contains("diplos"))
    GetIntegerSpec(params["diplos"].GetString(), diplos);

  if (params.Contains("haplos"))
    GetIntegerSpec(params["haplos"].GetString(), haplos);

  if (params.Contains("pop_diplos"))
    GetIntegerSpec(params["pop_diplos"].GetString(), pop_diplos);

  if (params.Contains("pop_haplos"))
    GetIntegerSpec(params["pop_haplos"].GetString(), pop_haplos);

  if (params.Contains("gender"))
  {
    vector<ParamValue> values = params["gender"].GetList();
    gender.resize(values.size());
    for (int i = 0; i < values.size(); ++i)
      gender[i] = values[i].GetInt();
  }

  if (params.Contains("map_pops"))
  {
    vector<ParamValue> values = params["map_pops"].GetList();
    map_pops.resize(values.size());
    for (int i = 0; i < values.size(); ++i)
      map_pops[i] = values[i].GetInt();
  }

  if (params.Contains("restrict_pop_missing"))
    restrict_pop_missing = true;
  
  if (params.Contains("restrict_monomorphic"))
    restrict_monomorphic = true;

  Assert(!loci.empty() || !chroms.empty() || 
         !pops.empty() || !diplos.empty() || !haplos.empty() || 
         !pop_diplos.empty() || !pop_haplos.empty() || !gender.empty() ||
         !map_pops.empty() || restrict_pop_missing || restrict_monomorphic);

  if (!pops.empty() || !diplos.empty() || !haplos.empty() || 
      !pop_diplos.empty() || !pop_haplos.empty() || !gender.empty() ||
      !map_pops.empty() || restrict_pop_missing || restrict_monomorphic)
    Assert(!input_file.empty());

  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;
  if (!input_profile_file.empty())
    cerr << "- Input SNP Profile file = " << input_profile_file << endl;
  if (!input_file.empty())
    cerr << "- Input file = " << input_file << endl;
  if (!output_dir.empty())
    cerr << "- Output directory = " << output_dir << endl;
  if (!output_profile_file.empty())
    cerr << "- Output SNP Profile file = " << output_profile_file << endl;
  if (!output_file.empty())
    cerr << "- Output file = " << output_file << endl;

  if (!loci.empty())
    cerr << "- Restrict to loci { " << IntegerListString(loci.begin(), loci.end()) << " }" << endl;
  if (!chroms.empty())
    cerr << "- Restrict to chromosomes { " << IntegerListString(chroms.begin(), chroms.end()) << " }" << endl;
  if (!pops.empty())
    cerr << "- Restrict to populations { " << IntegerListString(pops.begin(), pops.end()) << " }" << endl;
  if (!diplos.empty())
    cerr << "- Restrict to diploid individuals { " << IntegerListString(diplos.begin(), diplos.end()) << " }" << endl;
  if (!haplos.empty())
    cerr << "- Restrict to haplotypes { " << IntegerListString(haplos.begin(), haplos.end()) << " }" << endl;
  if (!pop_diplos.empty())
    cerr << "- Restrict to population diploid individuals { " << IntegerListString(pop_diplos.begin(), pop_diplos.end()) << " }" << endl;
  if (!pop_haplos.empty())
    cerr << "- Restrict to population haplotypes { " << IntegerListString(pop_haplos.begin(), pop_haplos.end()) << " }" << endl;
  if (!gender.empty())
    cerr << "- Restrict to genders { " << IntegerListString(gender.begin(), gender.end()) << " }" << endl;
  if (!map_pops.empty())
    cerr << "- Map populations { " << IntegerListString(map_pops.begin(), map_pops.end()) << " }" << endl;
  if (restrict_pop_missing)
    cerr << "- Restrict sites with missing population data" << endl;
  if (restrict_monomorphic)
    cerr << "- Restrict monomorphic sites" << endl;

  if (!pop_labels.empty())
  {
    cerr << "- Relabel populations to {";
    for (int i = 0; i < pop_labels.size(); ++i)
      cerr << " " << pop_labels[i];
    cerr << " }" << endl;
  }
}


void GetIntegerSpec(string s, vector<int> &result)
{
  int p = 0;
  while (p < s.length())
  {
    if (isdigit(s[p]))
    {
      int i;
      Assert(StringToInt(s, i, p, p));
      if ((p == s.length()) || (s[p] == ','))
      {
        result.push_back(i);
        if (p < s.length())
          ++p;
        continue;
      }

      Assert(s[p++] == '-');
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p == s.length()) || (s[p] == ','));
      if (i <= j)
      {
        for (int k = i; k <= j; ++k)
          result.push_back(k);
      }
      else
      {
        for (int k = i; k >= j; --k)
          result.push_back(k);
      }
      if (p < s.length());
      ++p;
    }
    else if (s[p] == '(')
    {
      int i;
      Assert(StringToInt(s, i, p, ++p));
      Assert((p < s.length()) && (s[p] == ','));
      ++p;
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p < s.length()) && (s[p] == ')'));
      Assert(j >= 0);
      for (int k = 0; k < j; ++k)
        result.push_back(i + k);
      ++p;
      Assert((p == s.length()) || (s[p] == ','));
      if (p < s.length())
        ++p;
    }
    else
      Assert(false);
  }
}


int Main(vector<string> args)
{
  cerr << "RestrictSNPData" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;

  SNPProfile profile;
  SNPData data;
  if (!input_profile_file.empty())
  {
    istream *in = NULL;
    if (!input_dir.empty())
      in = InFileStream(RelPath(input_dir, input_profile_file));
    else
      in = InFileStream(input_profile_file);

    if (!in)
    {
      cerr << "Could not open file \"" << input_profile_file << "\" in directory";
      if (!input_dir.empty())
        cerr << " \"" << input_dir << "\""
;
      Exit(1);
    }

    cerr << "* Reading SNP profile...";
    profile.ReadBin(*in);
    cerr << " done." << endl;
    delete in;
    
    profile.Validate();
    data.SetProfile(profile);
  }

  {
    istream *in = NULL;
    if (!input_dir.empty())
      in = InFileStream(RelPath(input_dir, input_file));
    else
      in = InFileStream(input_file);

    if (!in)
    {
      cerr << "Could not open file \"" << input_file << "\" in directory";
      if (!input_dir.empty())
        cerr << " \"" << input_dir << "\"";
      Exit(1);
    }

    cerr << "* Reading SNP data...";
    if (BinFileType(input_file) == "snpprofiledata")
      ReadBinSNPProfileData(*in, profile, data);
    else
      data.ReadBin(*in);
    cerr << " done." << endl;
    delete in;

    if (BinFileType(input_file) == "snpprofiledata")
      profile.Validate();
    data.Validate();
  }

  {
    vector<int> select_loci(profile.loci.size(), 0);

    if (!loci.empty())
      for (int i = 0; i < loci.size(); ++i)
      {
        Assert((loci[i] >= 0) && (loci[i] < profile.loci.size()));
        select_loci[loci[i]] = 1;
      }

    if (!chroms.empty())
      for (int i = 0; i < chroms.size(); ++i)
      {
        Assert((chroms[i] >= 0) && (chroms[i] < profile.chroms.size()));
        for (int j = profile.chroms[chroms[i]].range.left();
             j < profile.chroms[chroms[i]].range.right(); ++j)
          select_loci[j] = 1;
      }

    if (!loci.empty() || !chroms.empty())
    {
      cerr << "* Restricting loci...";

      SNPProfile new_profile;
      SNPData new_data;
      new_profile.CopyRestrictLoci(profile, select_loci);
      new_data.SetProfile(new_profile);
      new_data.CopyRestrictLoci(data, select_loci);

      profile.Swap(new_profile);
      data.Swap(new_data);
      data.SetProfile(profile);
      cerr << " done." << endl;

      profile.Validate();
      data.Validate();
    }
  }

  {
    vector<int> select_haplos(data.num_haplos, 0);
    
    if (!pops.empty())
      for (int i = 0; i < pops.size(); ++i)
      {
        Assert((pops[i] >= 0) && (pops[i] < data.num_pops));
        for (int j = 0; j < data.pops[pops[i]].haplos.size(); ++j)
          select_haplos[data.pops[pops[i]].haplos[j]] = 1;
      }

    if (!diplos.empty())
      for (int i = 0; i < diplos.size(); ++i)
      {
        Assert((diplos[i] >= 0) && (diplos[i] < data.num_diplos));
        select_haplos[data.diplos[diplos[i]].haplos[0]] = 1;
        select_haplos[data.diplos[diplos[i]].haplos[1]] = 1;
      }

    if (!haplos.empty())
      for (int i = 0; i < haplos.size(); ++i)
      {
        Assert((haplos[i] >= 0) && (haplos[i] < data.num_haplos));
        select_haplos[haplos[i]] = 1;
      }

    if (!pop_diplos.empty())
      for (int i = 0; i < pop_diplos.size(); ++i)
        for (int pop = 0; pop < data.num_pops; ++pop)
        {
          Assert((pop_diplos[i] >= 0) && (pop_diplos[i] < data.pops[pop].diplos.size()));
          select_haplos[data.diplos[data.pops[pop].diplos[pop_diplos[i]]].haplos[0]] = 1;
          select_haplos[data.diplos[data.pops[pop].diplos[pop_diplos[i]]].haplos[1]] = 1;
        }

    if (!pop_haplos.empty())
      for (int i = 0; i < pop_haplos.size(); ++i)
        for (int pop = 0; pop < data.num_pops; ++pop)
        {
          Assert((pop_haplos[i] >= 0) && (pop_haplos[i] < data.pops[pop].haplos.size()));
          select_haplos[data.pops[pop].haplos[pop_haplos[i]]] = 1;
        }

    if (!gender.empty())
      for (int i = 0; i < gender.size(); ++i)
      {
        Assert((gender[i] == 0) || (gender[i] == 1));
        for (int d = 0; d < data.num_diplos; ++d)
          if (data.diplos[d].gender == gender[i])
          {
            select_haplos[data.diplos[d].haplos[0]] = 1;
            select_haplos[data.diplos[d].haplos[1]] = 1;
          }
      }

    if (!pops.empty() || !diplos.empty() || !haplos.empty() ||
        !pop_diplos.empty() || !pop_haplos.empty() || !gender.empty())
    {
      cerr << "* Restricting data...";

      SNPData new_data(profile);
      new_data.CopyRestrictHaplos(data, select_haplos);

      data.Swap(new_data);
      cerr << " done." << endl;

      data.Validate();
    }
  }

  if (!map_pops.empty())
  {
    cerr << "* Mapping populations...";
    data.MapPops(map_pops);
    cerr << " done." << endl;
    data.Validate();
  }

  if (restrict_pop_missing)
  {
    cerr << "* Restricting missing population data...";
    vector<int> select_loci(data.seq_len, 1);
    int missing_count = 0;

    vector<int> pop_counts;
    for (int i = 0; i < data.seq_len; ++i)
    {
      pop_counts.clear();
      pop_counts.resize(data.num_pops, 0);
      for (int k = 0; k < data.num_haplos; ++k)
      {
        Assert((data.haplos[k][i] >= 0) && (data.haplos[k][i] <= 4));
        if (data.haplos[k][i] < 4)
          ++pop_counts[data.haplos[k].pop];
      }

      bool missing_pop = false;
      for (int pop = 0; pop < data.num_pops; ++pop)
        if (pop_counts[pop] == 0)
          missing_pop = true;

      if (missing_pop)
      {
        select_loci[i] = 0;
        ++missing_count;
      }
    }

    if (missing_count == 0)
      cerr << " No missing data!" << endl;
    else
    {
      SNPProfile new_profile;
      SNPData new_data;
      new_profile.CopyRestrictLoci(profile, select_loci);
      new_data.SetProfile(new_profile);
      new_data.CopyRestrictLoci(data, select_loci);

      profile.Swap(new_profile);
      data.Swap(new_data);
      data.SetProfile(profile);
      cerr << " done." << endl;

      profile.Validate();
      data.Validate();
    }
  }

  if (restrict_monomorphic)
  {
    cerr << "* Restricting monomorphic sites...";
    vector<int> select_loci(data.seq_len, 1);
    int mono_count = 0;

    vector<int> counts;
    for (int i = 0; i < data.seq_len; ++i)
    {
      counts.clear();
      counts.resize(5, 0);
      for (int k = 0; k < data.num_haplos; ++k)
        ++counts[data.haplos[k][i]];

      int non_zero_counts = 0;
      for (int b = 0; b < 4; ++b)
        if (counts[b] > 0)
          ++non_zero_counts;
      
      if (non_zero_counts == 1)
      {
        select_loci[i] = 0;
        ++mono_count;
      }
    }

    if (mono_count == 0)
      cerr << " No monomorphic sites!" << endl;
    else
    {
      cerr << " Found " << mono_count << " monomorphic sites...";
      SNPProfile new_profile;
      SNPData new_data;
      new_profile.CopyRestrictLoci(profile, select_loci);
      new_data.SetProfile(new_profile);
      new_data.CopyRestrictLoci(data, select_loci);

      profile.Swap(new_profile);
      data.Swap(new_data);
      data.SetProfile(profile);
      cerr << " done." << endl;

      profile.Validate();
      data.Validate();
    }
  }

  if (!pop_labels.empty())
  {
    Assert(pop_labels.size() == data.num_pops);
    for (int pop = 0; pop < data.num_pops; ++pop)
      data.pops[pop].label = pop_labels[pop];
  }

  if (!output_profile_file.empty())
  {
    cerr << "* Writing profile output...";
    ostream *out = OutFileStream(RelPath(output_dir, output_profile_file));
    AssertMsg(out, RelPath(output_dir, output_profile_file));
    profile.WriteBin(*out);
    delete out;
    cerr << " done." << endl;
  }
  
  if (!output_file.empty())
  {
    if (data.num_diplos > 0)
    {
      vector<int> gender_count(2, 0);
      for (int d = 0; d < data.num_diplos; ++d)
      {
        Assert((data.diplos[d].gender == 0) || (data.diplos[d].gender == 1));
        ++gender_count[data.diplos[d].gender];
      }
      cerr << "- # females = " << gender_count[0] << ", # males = " << gender_count[1] << endl;
    }

    cerr << "* Writing SNP data output...";
    ostream *out = OutFileStream(RelPath(output_dir, output_file));
    AssertMsg(out, RelPath(output_dir, output_file));

    if (BinFileType(output_file) == "snpprofiledata")
      WriteBinSNPProfileData(*out, data);
    else
      data.WriteBin(*out);

    delete out;
    cerr << " done." << endl;
  }

  cerr << "Done!" << endl;
  return 0;
}
