
#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"
#include "Simulate.h"
#include "HaploModel.h"
#include "PhasingError.h"


Params params(false);

int debug = 0;

string input_dir;
string output_dir;
string model_file;
bool default_model = false;
string sample_file;
string testing_file;
string simulate_testing_file;
vector<string> composite_files;
bool do_haplos = false;
string phases_file;
bool empty_phases;
string phaseerrors_file;
vector<int> diplos;
bool smooth_posterior = false;
double smooth_stddev = 0.01;
double default_diplo_switch = 0.1;
double phase_posterior_threshold = 0.5;
bool haploposterior = false;
bool viterbi_phasing = false;
double min_block_len = 0.0;

bool set_default_pop_recomb = false;
double default_pop_recomb;
bool set_default_pop_admix = false;
double default_pop_admix;

string posterior_file;
string posterior_popdata_file;
string viterbi_popdata_file;
string phaseposterior_file;

int max_phasing_iterations = -1;

int job_num = -1;
int num_jobs = 10;

string required_param[] =
{
  "sample_file",
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

void GetIntegerSpec(const string &s, vector<int> &result);

void InitParams()
{
  params.ExpandImports();

  uint32 seed = RandomSeed();
  if (params.Contains("rand_seed"))
    seed = params["rand_seed"].GetInt();
  srand(seed);
  cerr << "- Random seed = " << seed << endl;

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (params.Contains("model_file"))
  {
    model_file = params["model_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(model_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(model_file))) == "haplomodel");
  }

  if (params.Contains("default_model"))
    default_model = true;

  if (params.Contains("sample_file"))
  {
    sample_file = params["sample_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(sample_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(sample_file))) == "snpprofiledata");
  }

  if (params.Contains("testing_file"))
  {
    testing_file = params["testing_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(testing_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(testing_file))) == "snpprofiledata");
  }

  if (params.Contains("simulate_testing_file"))
  {
    simulate_testing_file = params["simulate_testing_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(simulate_testing_file)) == "bin");
    Assert((GetExtension(RemoveExtension(RemoveCompressedExtension(simulate_testing_file))) == "diplopop") || (GetExtension(RemoveExtension(RemoveCompressedExtension(simulate_testing_file))) == "haplopop"));
  }

  Assert(!testing_file.empty() || !simulate_testing_file.empty());
  Assert(!model_file.empty() || default_model);

  if (params.Contains("do_haplos"))
    do_haplos = true;

  if (params.Contains("composite_files"))
  {
    Assert(!simulate_testing_file.empty());
    vector<ParamValue> values = params["composite_files"].GetList();
    composite_files.resize(values.size());
    bool has_haplo = false;
    bool has_diplo = false;
    for (int i = 0; i < values.size(); ++i)
    {
      string filename = values[i].GetString();
      Assert(GetExtension(RemoveCompressedExtension(filename)) == "bin");
      if (GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "diplopop")
        has_diplo = true;
      else if (GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "haplopop")
        has_haplo = true;
      else
        AssertMsg(false, filename);
      composite_files[i] = filename;
    }
    Assert(!has_haplo || !has_diplo);
  }

  if (params.Contains("phases_file"))
  {
    phases_file = params["phases_file"].GetString();    
    Assert(GetExtension(RemoveCompressedExtension(phases_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(phases_file))) == "phases");
  }

  if (params.Contains("empty_phases"))
    empty_phases = true;

  if (params.Contains("phaseerrors_file"))
  {
    phaseerrors_file = params["phaseerrors_file"].GetString();    
    Assert(GetExtension(RemoveCompressedExtension(phaseerrors_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(phaseerrors_file))) == "phaseerrors");
  }

  if (params.Contains("set_default_pop_recomb"))
  {
    set_default_pop_recomb = true;
    default_pop_recomb = params["set_default_pop_recomb"].GetDouble();
  }

  if (params.Contains("set_default_pop_admix"))
  {
    set_default_pop_admix = true;
    default_pop_admix = params["set_default_pop_admix"].GetDouble();
  }

  if (params.Contains("diplos"))
    GetIntegerSpec(params["diplos"].GetString(), diplos);
  
  if (params.Contains("posterior_file"))
  {
    posterior_file = params["posterior_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(posterior_file)) == "bin");
    Assert((GetExtension(RemoveExtension(RemoveCompressedExtension(posterior_file))) == "diploposterior") || (GetExtension(RemoveExtension(RemoveCompressedExtension(posterior_file))) == "haploposterior"));
  }

  if (params.Contains("posterior_popdata_file"))
  {
    posterior_popdata_file = params["posterior_popdata_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(posterior_popdata_file)) == "bin");
    Assert((GetExtension(RemoveExtension(RemoveCompressedExtension(posterior_popdata_file))) == "diplopopdata") || (GetExtension(RemoveExtension(RemoveCompressedExtension(posterior_popdata_file))) == "haplopopdata"));
  }

  if (params.Contains("viterbi_popdata_file"))
  {
    viterbi_popdata_file = params["viterbi_popdata_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(viterbi_popdata_file)) == "bin");
    Assert((GetExtension(RemoveExtension(RemoveCompressedExtension(viterbi_popdata_file))) == "diplopopdata") || (GetExtension(RemoveExtension(RemoveCompressedExtension(viterbi_popdata_file))) == "haplopopdata"));
  }

  if (params.Contains("smooth_posterior"))
    smooth_posterior = true;

  if (params.Contains("smooth_stddev"))
    smooth_stddev = params["smooth_stddev"].GetDouble();

  if (params.Contains("max_phasing_iterations"))
    max_phasing_iterations = params["max_phasing_iterations"].GetInt();

  if (params.Contains("phaseposterior_file"))
  {
    phaseposterior_file = params["phaseposterior_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(phaseposterior_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(phaseposterior_file))) == "phaseposterior");
  }

  if (params.Contains("phase_posterior_threshold"))
    phase_posterior_threshold = params["phase_posterior_threshold"].GetDouble();

  if (params.Contains("haploposterior"))
    haploposterior = true;

  if (params.Contains("viterbi_phasing"))
    viterbi_phasing = true;

  if (params.Contains("min_block_len"))
    min_block_len = params["min_block_len"].GetDouble();

  if (params.Contains("job_num"))
    job_num = params["job_num"].GetInt();

  if (params.Contains("num_jobs"))
    num_jobs = params["num_jobs"].GetInt();

  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;

  cerr << "- Sample file = " << sample_file << endl;

  cerr << "- Testing file = " << testing_file << endl;

  if (!simulate_testing_file.empty())
    cerr << "- Simulated testing file = " << simulate_testing_file << endl;

  if (!model_file.empty())
    cerr << "- Model file = " << model_file << endl;

  if (default_model)
    cerr << "- Default initial model" << endl;

  if (do_haplos)
    cerr << "- Doing haplotypes only" << endl;

  if (!composite_files.empty())
  {
    cerr << "- Composite files {";
    for (int i = 0; i < composite_files.size(); ++i)
      cerr << " " << composite_files[i];
    cerr << " }" << endl;
  }

  if (!phases_file.empty())
    cerr << "- Phases file = " << phases_file << endl;
  if (empty_phases)
    cerr << "- Empty phases" << endl;

  if (!phaseerrors_file.empty())
    cerr << "- Phase errors file = " << phaseerrors_file << endl;

  if (!output_dir.empty())
    cerr << "- Output directory = " << output_dir << endl;

  if (!posterior_file.empty())
    cerr << "- Posterior output file = " << posterior_file << endl;

  if (!posterior_popdata_file.empty())
    cerr << "- Posterior decoding output file = " << posterior_popdata_file << endl;
  
  if (haploposterior)
    cerr << "- Decoding haplotype posteriors" << endl;

  if (!viterbi_popdata_file.empty())
    cerr << "- Viterbi output file = " << viterbi_popdata_file << endl;

  if (!diplos.empty())
    cerr << "- Selected diploid individuals { " << IntegerListString(diplos.begin(), diplos.end()) << " }" << endl;

  if (set_default_pop_recomb)
    cerr << "- Setting default population recombination scale factor = " << default_pop_recomb << endl;

  if (set_default_pop_admix)
    cerr << "- Setting default population admixture rate = " << default_pop_admix << endl;

  if (smooth_posterior)
    cerr << "- Smooth posteriors with stddev = " << smooth_stddev  << endl;

  if (max_phasing_iterations >= 0)
    cerr << "- Max phasing iterations = " << max_phasing_iterations << endl;

  if (!phaseposterior_file.empty())
    cerr << "- Phasing posterior file = " << phaseposterior_file << endl;

  if (!phases_file.empty() || empty_phases)
    cerr << "- Phase posterior threshold = " << phase_posterior_threshold << endl;
  if (viterbi_phasing)
    cerr << "- Viterbi phasing" << endl;

  if (min_block_len > 0.0)
    cerr << "- Minimum block length = " << min_block_len << endl;
}


void GetIntegerSpec(const string &s, vector<int> &result)
{
  int p = 0;
  while (p < s.length())
  {
    if (isdigit(s[p]))
    {
      int i;
      Assert(StringToInt(s, i, p, p));
      if ((p == s.length()) || (s[p] == ','))
      {
        result.push_back(i);
        if (p < s.length())
          ++p;
        continue;
      }

      Assert(s[p++] == '-');
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p == s.length()) || (s[p] == ','));
      if (i <= j)
      {
        for (int k = i; k <= j; ++k)
          result.push_back(k);
      }
      else
      {
        for (int k = i; k >= j; --k)
          result.push_back(k);
      }
      if (p < s.length());
      ++p;
    }
    else if (s[p] == '(')
    {
      int i;
      Assert(StringToInt(s, i, p, ++p));
      Assert((p < s.length()) && (s[p] == ','));
      ++p;
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p < s.length()) && (s[p] == ')'));
      Assert(j >= 0);
      for (int k = 0; k < j; ++k)
        result.push_back(i + k);
      ++p;
      Assert((p == s.length()) || (s[p] == ','));
      if (p < s.length())
        ++p;
    }
    else
      Assert(false);
  }
}


bool GaussianInRange(double Z)
{
  return (Z >= -4.0) && (Z <= 4.0);
}

double Gaussian(double Z)
{
  return exp(-0.5*Z*Z);
}


void CreateErrors(vector<int8> haplo, 
                  double error_rate, double missing_rate)
{
  for (int i = 0; i < haplo.size(); ++i)
  {
    if (haplo[i] == 4)
      continue;
    if (drand() < missing_rate)
      haplo[i] = 4;
    else
    {
      if (drand() < error_rate)
        haplo[i] = (haplo[i] + 1 + (irand() % 3)) & 3;
    }
  }
}


vector< pair<int, int> > smooth_ranges;
vector<double> smooth_scales;


void InitSmooth(const SNPData &data, double stddev)
{
  smooth_ranges.resize(data.seq_len);
  smooth_scales.resize(data.seq_len);
  double inv_stddev = 1.0 / stddev;

  for (int i = 0; i < data.seq_len; ++i)
  {
    double genetic_pos = data.locus_genetic_pos[i];

    int start = i;
    while ((start >= 0) && GaussianInRange((data.locus_genetic_pos[start] - genetic_pos) * inv_stddev))
      --start;
    ++start;
    
    int end = i;
    while ((end < data.seq_len) && GaussianInRange((data.locus_genetic_pos[end] - genetic_pos) * inv_stddev))
      ++end;
    --end;

    double total = 0.0;

    for (int j = start; j <= end; ++j)
      total += Gaussian((data.locus_genetic_pos[j] - genetic_pos) * inv_stddev);

    smooth_ranges[i] = pair<int, int>(start, end);
    smooth_scales[i] = 1.0 / total;
  }  
}
  

template <typename TS, typename TD>
void Smooth(const SNPData &data, const vector< vector<TS> > &source, 
            vector< vector<TD> > &dest, double stddev)
{
  dest.resize(data.num_pops, vector<TD>(data.seq_len, 0));
  vector<double> pop_total;
  double inv_stddev = 1.0 / stddev;
  
  for (int i = 0; i < data.seq_len; ++i)
  {
    double genetic_pos = data.locus_genetic_pos[i];

    int start = smooth_ranges[i].first;
    int end = smooth_ranges[i].second;
    
    pop_total.clear();
    pop_total.resize(data.num_pops, 0.0);

    for (int j = start; j <= end; ++j)
    {
      double profile = Gaussian((data.locus_genetic_pos[j] - genetic_pos) * inv_stddev);
      for (int pop = 0; pop < data.num_pops; ++pop)
        pop_total[pop] += double(source[pop][j]) * profile;
    }

    double scale = smooth_scales[i];
    for (int pop = 0; pop < data.num_pops; ++pop)
      dest[pop][i] = TD(pop_total[pop] * scale);
  }  
}


template <typename TS, typename TD>
void SmoothNonZero(const SNPData &data, const vector<TS> &source, 
                   vector<TD> &dest, double stddev)
{
  dest.resize(data.seq_len, 0);
  
  for (int i = 0; i < data.seq_len; ++i)
  {
    int start = i;
    while ((start >= 0) && GaussianInRange((data.locus_genetic_pos[start] - data.locus_genetic_pos[i]) / stddev))
      --start;
    ++start;
    
    int end = i;
    while ((end < data.seq_len) && GaussianInRange((data.locus_genetic_pos[end] - data.locus_genetic_pos[i]) / stddev))
      ++end;
    --end;
    
    double total = 0.0;
    for (int j = start; j <= end; ++j)
      if (source[j] != TS(0))
        total += Gaussian((data.locus_genetic_pos[j] - data.locus_genetic_pos[i]) / stddev);
    double scale = 1.0 / total;
    
    total = 0.0;
    for (int j = start; j <= end; ++j)
    {
      double profile = Gaussian((data.locus_genetic_pos[j] - data.locus_genetic_pos[i]) / stddev);
      total += double(source[j]) * profile;
    }

    dest[i] = TD(total * scale * 10);
  }  
}


template <typename T>
void DecodePops(const vector< vector<T> > &pop_p, vector<int8> &pops)
{
  Assert(!pop_p.empty());
  Assert(!pop_p[0].empty());
  int num_pops = pop_p.size();
  int seq_len = pop_p[0].size();
  pops.resize(seq_len);
  for (int i = 0; i < seq_len; ++i)
  {
    int max_pop = 0;
    for (int pop = 1; pop < num_pops; ++pop)
      if (pop_p[pop][i] > pop_p[max_pop][i])
        max_pop = pop;
    pops[i] = max_pop;
  }
}


template <typename TS, typename TD>
void Set(const vector< vector<TS> > &source, vector< vector<TD> > &dest)
{
  Assert(!source.empty());
  dest.resize(source.size());
  for (int i = 0; i < source.size(); ++i)
  {
    dest[i].resize(source[i].size());
    for (int j = 0; j < source[i].size(); ++j)
      dest[i][j] = TD(source[i][j]);
  }
}


void DecodeHaploPops(const SNPData &data,
                     const HaploModel::Posterior &posterior,
                     vector<int8> &pops)
{
  if (!smooth_posterior)
    posterior.DecodeHaploPops(data, pops);
  else
  {
    pops.resize(data.seq_len, -1);
    vector<double> haplo_total;

    for (int i = 0; i < data.seq_len; ++i)
    {
      int start = i;
      while ((start >= 0) && GaussianInRange((data.locus_genetic_pos[start] - data.locus_genetic_pos[i]) / smooth_stddev))
        --start;
      ++start;
      
      int end = i;
      while ((end < data.seq_len) && GaussianInRange((data.locus_genetic_pos[end] - data.locus_genetic_pos[i]) / smooth_stddev))
        ++end;
      --end;
      
      double total = 0.0;
      for (int j = start; j <= end; ++j)
        total += Gaussian((data.locus_genetic_pos[j] - data.locus_genetic_pos[i]) / smooth_stddev);
      double scale = 1.0 / total;
      
      haplo_total.clear();
      haplo_total.resize(data.num_haplos, 0.0);
      for (int j = start; j <= end; ++j)
      {
        double profile = Gaussian((data.locus_genetic_pos[j] - data.locus_genetic_pos[i]) / smooth_stddev);
        for (int k = 0; k < data.num_haplos; ++k)
          haplo_total[k] += double(posterior[i].O(k)) * profile;
      }
    
      int max_k = 0;
      for (int k = 1; k < data.num_haplos; ++k)
        if (haplo_total[k] > haplo_total[max_k])
          max_k = k;
      
      pops[i] = data.haplos[max_k].pop;
    }  
  }
}


void TestHaploPop(const SNPData &data, const SNPData &testing_data,
                  const HaploModel::Model &model,
                  const Simulate::HaploPop &haplo_testing, 
                  const vector<Simulate::HaploPop> &composite,
                  vector< vector< vector<float> > > &pop_ps,
                  vector< vector<int8> > &posterior_pops,
                  vector< vector<int8> > &viterbi_pops)
{
  if (haplo_testing.haplos.empty())
    return;

  if (!posterior_popdata_file.empty())
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(posterior_popdata_file))) == "haplopopdata");
  if (!viterbi_popdata_file.empty())
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(viterbi_popdata_file))) == "haplopopdata");

  pop_ps.reserve(pop_ps.size() + haplo_testing.haplos.size());
  posterior_pops.reserve(posterior_pops.size() + haplo_testing.haplos.size());
  viterbi_pops.reserve(viterbi_pops.size() + haplo_testing.haplos.size());

  cerr << "* # haplotypes = " << haplo_testing.haplos.size() << endl;
  Simulate::Haplo temp_haplo;
  for (int k = 0; k < haplo_testing.haplos.size(); ++k)
  {
    cerr << "* Testing on haplotype " << k << "... ";
    vector<int8> haplo, pops;
    vector<int> haplo_exclude;

    if (composite.empty())
      temp_haplo = haplo_testing.haplos[k];
    else
      temp_haplo.Composite(composite, haplo_testing.haplos[k]);

    temp_haplo.Generate(testing_file.empty() ? data : testing_data, haplo);
    haplo_testing.haplos[k].LabelPop(pops);
    if (testing_file.empty())
      temp_haplo.Label(haplo_exclude);
    else
      haplo_exclude.clear();
    
    //CreateErrors(haplo, 0.01, 0.05);

    if (!posterior_file.empty() || !posterior_popdata_file.empty())
    {
      cerr << "F";
      HaploModel::Forward forward(data, model, haplo, vector<uint64>(),
                                  vector<int>(), haplo_exclude);
      cerr << "B";
      HaploModel::Backward backward(data, model, haplo, vector<uint64>(),
                                    vector<int>(), haplo_exclude);
      cerr << "P";
      HaploModel::Posterior posterior(data, forward, backward);

      vector< vector<float> > pop_p;
      if (smooth_posterior)
        Smooth(data, posterior.pop_p, pop_p, smooth_stddev);
      else
        Set(posterior.pop_p, pop_p);

      if (!posterior_file.empty())
        pop_ps.push_back(pop_p);

      if (!posterior_popdata_file.empty())
      {
        vector<int8> decode_pops;
        if (!haploposterior)
          DecodePops(pop_p, decode_pops);
        else
          DecodeHaploPops(data, posterior, decode_pops);

        posterior_pops.push_back(decode_pops);
      }
    }

    if (!viterbi_popdata_file.empty())
    {
      cerr << "V";
      HaploModel::Viterbi viterbi(data, model, haplo, vector<uint64>(), 
                                  vector<int>(), haplo_exclude);
      vector<int8> decode_pops;
      viterbi.DecodePops(data, decode_pops);
      viterbi_pops.push_back(decode_pops);
    }

    cerr << endl;
  }
}

void TestDiploPop(const SNPData &data, const SNPData &testing_data,
                  const HaploModel::Model &model,
                  Simulate::DiploPop &diplo_testing, 
                  const vector<Simulate::DiploPop> &composite,
                  const vector< vector<int8> > &phases,
                  const vector< vector<uint8> > &phaseerrors,
                  vector< vector< vector<float> > > &pop_ps,
                  vector< vector<int8> > &posterior_pops,
                  vector< vector<int8> > &viterbi_pops,
                  vector< vector<float> > &phasing_errors)
{
  if (diplo_testing.diplos.empty())
    return;

  if (!posterior_popdata_file.empty())
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(posterior_popdata_file))) == "diplopopdata");
  if (!viterbi_popdata_file.empty())
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(viterbi_popdata_file))) == "diplopopdata");

  cerr << "* # diploid samples = " << diplo_testing.diplos.size() << endl;
  Simulate::Diplo temp_diplo;
  vector<uint8> x_switch_prior;

  if (diplos.empty())
  {
    diplos.resize(diplo_testing.diplos.size());
    for (int d = 0; d < diplos.size(); ++d)
      diplos[d] = d;
  }
  else
  {
    vector<bool> included(diplo_testing.diplos.size(), false);
    for (int diplo_index = 0; diplo_index < diplos.size(); ++diplo_index)
    {
      Assert((diplos[diplo_index] >= 0) && (diplos[diplo_index] < included.size()));
      included[diplos[diplo_index]] = true;
    }

    for (int d = 0; d < diplo_testing.diplos.size(); ++d)
      if (!included[d])
        diplo_testing.diplos[d].Free();

    cerr << "* Testing on " << diplos.size() << " samples" << endl;
  }

  pop_ps.reserve(pop_ps.size() + 2*diplos.size());
  posterior_pops.reserve(posterior_pops.size() + 2*diplos.size());
  viterbi_pops.reserve(viterbi_pops.size() + 2*diplos.size());
  phasing_errors.reserve(phasing_errors.size() + diplos.size());

  for (int diplo_index = 0; diplo_index < diplos.size(); ++diplo_index)
  {
    int d = diplos[diplo_index];
    cerr << "* Testing on diploid sample " << d << ":";

    if (composite.empty())
      temp_diplo = diplo_testing.diplos[d];
    else
      temp_diplo.Composite(composite, diplo_testing.diplos[d]);

    vector< vector<int8> > pops(2), haplo(2);
    vector< vector<int> > haplo_exclude(2);
    temp_diplo.Generate(testing_file.empty() ? data : testing_data, haplo[0], haplo[1]);
    diplo_testing.diplos[d].LabelPop(testing_file.empty() ? data : testing_data, pops[0], pops[1]);
    if (testing_file.empty())
      temp_diplo.Label(haplo_exclude[0], haplo_exclude[1]);
    else
    {
      haplo_exclude[0].clear();
      haplo_exclude[1].clear();
    }

    //CreateErrors(haplo[0], 0.01, 0.05);
    //CreateErrors(haplo[1], 0.01, 0.05);

    if (!phases.empty())
    {
      Assert(d < phases.size());
      for (int i = 0; i < data.seq_len; ++i)
        if (phases[d][i])
        {
          swap(haplo[0][i], haplo[1][i]);
          if (!haplo_exclude[0].empty())
            swap(haplo_exclude[0][i], haplo_exclude[1][i]);
        }

      if (phaseerrors.empty())
        data.SetDiploSwitch(haplo[0], haplo[1], diplo_testing.diplos[d].gender,
                            x_switch_prior, default_diplo_switch);
      else
        x_switch_prior = phaseerrors[d];
    }
    else
      data.SetDiploSwitch(haplo[0], haplo[1], diplo_testing.diplos[d].gender,
                          x_switch_prior, 0.0);
    
    if (!phases.empty())
      cerr << " Phasing...";
    vector<int> phase_switched(data.seq_len - 1, 0);
    vector<HaploModel::Forward> forwards(2);
    vector<HaploModel::Backward> backwards(2);
    vector<HaploModel::ViterbiForward> viterbi_forwards(2);
    vector<HaploModel::ViterbiBackward> viterbi_backwards(2);
    PhasingError phasing;
    
    int phasing_iteration = 0;
    while (true)
    {
      cerr << " ";
      for (int a = 0; a < 2; ++a)
      {
        if (!viterbi_phasing)
        {
          cerr << "F";
          forwards[a].Calculate(data, model, haplo[a], vector<uint64>(),
                                vector<int>(), haplo_exclude[a]);
          cerr << "B";
          backwards[a].Calculate(data, model, haplo[a], vector<uint64>(),
                                 vector<int>(), haplo_exclude[a]);
        }
        else
        {
          cerr << "Vf";
          viterbi_forwards[a].Calculate(data, model, haplo[a], vector<uint64>(), vector<int>(), haplo_exclude[a]);
          cerr << "Vb";
          viterbi_backwards[a].Calculate(data, model, haplo[a], vector<uint64>(), vector<int>(), haplo_exclude[a]);
        }
      }

      if (phases.empty())
        break;

      cerr << "E";
      if (!viterbi_phasing)
        phasing.Calculate(data, haplo[0], forwards[0], backwards[0], haplo[1],
                          forwards[1], backwards[1], x_switch_prior);
      else
        phasing.Calculate(data, haplo[0], 
                          viterbi_forwards[0], viterbi_backwards[0],
                          haplo[1],
                          viterbi_forwards[1], viterbi_backwards[1],
                          x_switch_prior);

      vector<float> phasing_temp;
#if 0
      SmoothNonZero(data, phasing.x_switch, phasing_temp, 0.01);
      phasing.x_switch = phasing_temp;
#else
      phasing_temp = phasing.x_switch;
#endif

      if ((max_phasing_iterations >= 0) && (phasing_iteration >= max_phasing_iterations))
        break;

      vector<int> phase_switch(data.seq_len - 1, 0);
      int num_switch = 0;
      while (true)
      {
        int max_i = 0;
        for (int i = 1; i < (data.seq_len - 1); ++i)
          if (phasing_temp[i] > phasing_temp[max_i])
            max_i = i;
        
        if (phasing_temp[max_i] < phase_posterior_threshold)
          break;

        double phase_total = 0.0;
        for (int i = max_i - 1; i >= 0; --i)
        {
          phase_total += SNPData::switch_decode(x_switch_prior[i]);
          if (phase_total > 2.0)
            break;
          phasing_temp[i] = 0.0;
        }

        phase_total = 0.0;
        for (int i = max_i; i < (data.seq_len - 1); ++i)
        {
          phasing_temp[i] = 0.0;
          phase_total += SNPData::switch_decode(x_switch_prior[i]);
          if (phase_total > 2.0)
            break;
        }

        if (!phase_switched[max_i])
        {
          phase_switch[max_i] = 1;
          phase_switched[max_i] = 1;
          ++num_switch;
        }
      }

      cerr << "(" << num_switch << ")";
      if (num_switch == 0)
        break;

      int phase = 0;
      for (int i = 0; i < data.seq_len; ++i)
      {
        if (phase != 0)
        {
          swap(haplo[0][i], haplo[1][i]);
          if (!haplo_exclude[0].empty())
            swap(haplo_exclude[0][i], haplo_exclude[1][i]);
        }

        if ((i < (data.seq_len - 1)) && phase_switch[i])
          phase = 1 - phase;
      }

      ++phasing_iteration;
    }

    if (!phaseposterior_file.empty())
    {
      vector<float> v(data.seq_len, 0.0);
      for (int i = 1; i < data.seq_len; ++i)
        v[i] = phasing[i - 1];
      phasing_errors.push_back(v);
    }

    if (min_block_len > 0.0)
    {
      Assert(!viterbi_phasing);
      
      for (int a = 0; a < 2; ++a)
      {
        vector<uint64> pop_mask_from(data.seq_len), pop_mask_to(data.seq_len);
        uint64 all_pops_mask = (uint64(1) << data.num_pops) - 1;
        while (true)
        {
          cerr << "P";
          HaploModel::Posterior posterior(data, forwards[a], backwards[a]);
          vector<int8> pops;
          posterior.DecodePops(data, pops);
          
          vector< vector< pair<Segment, int> > > blocks(data.profile->chroms.size());
          for (int chrom = 0; chrom < data.profile->chroms.size(); ++chrom)
          {
            for (int i = data.profile->chroms[chrom].range.left(); 
                 i < data.profile->chroms[chrom].range.right(); )
            {
              int j = i + 1;
              while ((j < data.profile->chroms[chrom].range.right()) && 
                     (pops[j] == pops[i]))
                ++j;
              blocks[chrom].push_back(pair<Segment, int>(Segment(i, j, true), pops[i]));
              i = j;
            }
          }

          vector< vector<double> > block_len(data.profile->chroms.size());
          for (int chrom = 0; chrom < data.profile->chroms.size(); ++chrom)
          {
            block_len[chrom].resize(blocks[chrom].size());
            for (int i = 0; i < blocks[chrom].size(); ++i)
              block_len[chrom][i] = data.locus_genetic_pos[blocks[chrom][i].first.right() - 1] - data.locus_genetic_pos[blocks[chrom][i].first.left()];
          }

          bool small_block = false;
          for (int chrom = 0; chrom < data.profile->chroms.size(); ++chrom)
          {
            Assert(chrom < blocks.size());
            Assert(chrom < block_len.size());
            Assert(blocks[chrom].size() == block_len[chrom].size());
            bool all_small = true;
            for (int i = 0; i < blocks[chrom].size(); ++i)
              if (block_len[chrom][i] >= min_block_len)
              {
                all_small = false;
                break;
              }

            if (all_small)
            {
              for (int i = 0; i < blocks[chrom].size(); ++i)
              {
                int left = blocks[chrom][i].first.left();
                int right = blocks[chrom][i].first.right() - 1;
                for (int j = left; j <= right; ++j)
                  pop_mask_from[j] = pop_mask_to[j] = uint64(1) << blocks[chrom][i].second;
              }
            }
            else
            {
              for (int i = 0; i < blocks[chrom].size(); ++i)
              {
                int left = blocks[chrom][i].first.left();
                int right = blocks[chrom][i].first.right() - 1;
                if (block_len[chrom][i] < min_block_len)
                {
                  uint64 mask_from = 0, mask_to = 0;
                  int l;
                  for (l = i - 1; (l >= 0) && (block_len[chrom][l] < min_block_len); --l)
                    ;
                  if (l >= 0)
                    mask_from |= uint64(1) << blocks[chrom][l].second;
                  int r;
                  for (r = i + 1; (r < blocks[chrom].size()) && (block_len[chrom][r] < min_block_len); ++r)
                    ;
                  if (r < blocks[chrom].size())
                    mask_to |= uint64(1) << blocks[chrom][r].second;
                  AssertMsg((mask_to | mask_from) != 0, "<" + ToBinary(mask_to) + ", " + ToBinary(mask_from) + ">, " + ToStr(l) + ":" + ToStr(data.profile->chroms[chrom].range.left()) + ", " + ToStr(r) + ":" + ToStr(data.profile->chroms[chrom].range.right()));
                  for (int k = left; k <= right; ++k)
                    pop_mask_from[k] = mask_from, pop_mask_to[k] = mask_to;
                  small_block = true;
                }
                else
                {
                  for (int j = left; j <= right; ++j)
                    pop_mask_from[j] = pop_mask_to[j] = uint64(1) << blocks[chrom][i].second;
                }
              }
            }
          }
          
          cerr << "F";
          forwards[a].Calculate(data, model, haplo[a], 
                                pop_mask_from, pop_mask_to,
                                vector<int>(), haplo_exclude[a]);
          cerr << "B";
          backwards[a].Calculate(data, model, haplo[a], 
                                 pop_mask_from, pop_mask_to,
                                 vector<int>(), haplo_exclude[a]);
          
          //if (!small_block)
            break;
        }
      }
    }

    if (!posterior_file.empty() || !posterior_popdata_file.empty())
    {
      cerr << " ";
      if (viterbi_phasing)
      {
        for (int a = 0; a < 2; ++a)
        {
          cerr << "F";
          forwards[a].Calculate(data, model, haplo[a], vector<uint64>(),
                                vector<int>(), haplo_exclude[a]);
          cerr << "B";
          backwards[a].Calculate(data, model, haplo[a], vector<uint64>(),
                                 vector<int>(), haplo_exclude[a]);
        }
      }

      HaploModel::Posterior posterior;
      vector< vector<float> > pop_p;
      for (int a = 0; a < 2; ++a)
      {
        cerr << "P";
        posterior.Calculate(data, forwards[a], backwards[a]);

        if (smooth_posterior)
        {
          cerr << "S";
          Smooth(data, posterior.pop_p, pop_p, smooth_stddev);
        }
        else
          Set(posterior.pop_p, pop_p);
        
        if (!posterior_file.empty())
          pop_ps.push_back(pop_p);
        
        if (!posterior_popdata_file.empty())
        {
          vector<int8> decode_pops;
          DecodePops(pop_p, decode_pops);
          posterior_pops.push_back(decode_pops);
        }
      }
    }

    if (!viterbi_popdata_file.empty())
    {
      cerr << " ";
      for (int a = 0; a < 2; ++a)
      {
        cerr << "V";
        HaploModel::Viterbi viterbi(data, model, haplo[a], vector<uint64>(), 
                                    vector<int>(), haplo_exclude[a]);
        vector<int8> decode_pops;
        viterbi.DecodePops(data, decode_pops);
        viterbi_pops.push_back(decode_pops);
      }
    }

    cerr << endl;
  }
}


int Main(vector<string> args)
{
  cerr << "RunHaploModel" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;

  SNPProfile profile;
  SNPData data;
  {
    string filename;
    istream *in = InFileStream(RelPath(input_dir, sample_file));

    if (!in)
    {
      cerr << "Could not open file \"" << sample_file << "\"";
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"";
      cerr << endl;
      Exit(1);
    }

    cerr << "* Reading sample data...";
    ReadBinSNPProfileData(*in, profile, data);
    cerr << " done." << endl;
    delete in;

    profile.Validate();
    data.Validate();
  }

  SNPProfile testing_profile;
  SNPData testing_data;
  if (!testing_file.empty())
  {
    string filename;
    istream *in = InFileStream(RelPath(input_dir, testing_file));

    if (!in)
    {
      cerr << "Could not open file \"" << testing_file << "\"";
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"";
      cerr << endl;
      Exit(1);
    }

    cerr << "* Reading testing data...";
    ReadBinSNPProfileData(*in, testing_profile, testing_data);
    cerr << " done." << endl;
    delete in;

    testing_profile.Validate();
    testing_data.Validate();
  }

  Simulate::HaploPop haplo_testing(testing_file.empty() ? data : testing_data);
  Simulate::DiploPop diplo_testing(testing_file.empty() ? data : testing_data);
  if (!simulate_testing_file.empty())
  {
    istream *in = InFileStream(RelPath(input_dir, simulate_testing_file));
    if (!in)
    {
      cerr << "Could not open file \"" << simulate_testing_file << "\"";
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"";
      cerr << endl;
      Exit(1);
    }

    cerr << "* Reading simulated test set \"" << simulate_testing_file << "\"...";
    if (GetExtension(RemoveExtension(RemoveCompressedExtension(simulate_testing_file))) == "diplopop")
      diplo_testing.ReadBin(*in);
    else
      haplo_testing.ReadBin(*in);
      
    delete in;
    cerr << " done.";

    if (GetExtension(RemoveExtension(RemoveCompressedExtension(simulate_testing_file))) == "diplopop")
      cerr << " # diploid individuals = " << diplo_testing.diplos.size() << endl;
    else
      cerr << " # haplotypes = " << haplo_testing.haplos.size() << endl;
  }
  else
  {
    for (int d = 0; d < testing_data.num_diplos; ++d)
      diplo_testing.Add(Simulate::Diplo(testing_data, d));
    for (int k = 0; k < testing_data.num_haplos; ++k)
      if (testing_data.haplos[k].diplo < 0)
        haplo_testing.Add(Simulate::Haplo(testing_data, k));
  }    

  vector<Simulate::HaploPop> haplo_composite;
  vector<Simulate::DiploPop> diplo_composite;
  if (!composite_files.empty())
  {
    for (int i = 0; i < composite_files.size(); ++i)
    {
      cerr << "* Reading simulated population \"" << composite_files[i] << "\"...";
      if (GetExtension(RemoveExtension(RemoveCompressedExtension(composite_files[i]))) == "diplopop")
      {
        diplo_composite.push_back(Simulate::DiploPop(testing_file.empty() ? data : testing_data));
        diplo_composite.back().ReadBin(RelPath(input_dir, composite_files[i]));
      }
      else
      {
        haplo_composite.push_back(Simulate::HaploPop(testing_file.empty() ? data : testing_data));
        haplo_composite.back().ReadBin(RelPath(input_dir, composite_files[i]));
      }
      cerr << " done.";
      if (GetExtension(RemoveExtension(RemoveCompressedExtension(composite_files[i]))) == "diplopop")
        cerr << " # diploid individuals = " << diplo_composite.back().size() << endl;
      else
        cerr << " # haplotypes = " << haplo_composite.back().size() << endl;
    }
  }

  if (do_haplos)
  {
    for (int d = 0; d < diplo_testing.diplos.size(); ++d)
    {
      haplo_testing.haplos.push_back(Simulate::Haplo());
      haplo_testing.haplos.back().swap(diplo_testing.diplos[d].haplos[0]);
      haplo_testing.haplos.push_back(Simulate::Haplo());
      haplo_testing.haplos.back().swap(diplo_testing.diplos[d].haplos[1]);
    }

    diplo_testing.clear();
    testing_data.ClearDiploPairing();

    if (!diplo_composite.empty())
    {
      Assert(haplo_composite.empty());
      haplo_composite.resize(diplo_composite.size());
      for (int pop = 0; pop < diplo_composite.size(); ++pop)
      {
        for (int d = 0; d < diplo_composite[pop].size(); ++d)
        {
          haplo_composite[pop].Add(diplo_composite[pop][d].haplos[0]);
          haplo_composite[pop].Add(diplo_composite[pop][d].haplos[1]);
        }
      }
      diplo_composite.clear();
    }
  }

  vector< vector<int8> > phases;
  if (!phases_file.empty())
  {
    cerr << "* Reading phasing...";
    istream *in = InFileStream(RelPath(input_dir, phases_file));
    AssertMsg(in, RelPath(input_dir, phases_file));
    Read(*in, phases);
    delete in;
    cerr << " done." << endl;
    
    Assert(phases.size() == diplo_testing.diplos.size());
  }
  if (empty_phases)
  {
    phases.resize(diplo_testing.diplos.size(), vector<int8>(data.seq_len ,0));
  }

  vector< vector<uint8> > phaseerrors;
  if (!phaseerrors_file.empty())
  {
    cerr << "* Reading phase errors...";
    istream *in = InFileStream(RelPath(input_dir, phaseerrors_file));
    AssertMsg(in, RelPath(input_dir, phaseerrors_file));
    Read(*in, phaseerrors);
    delete in;
    cerr << " done." << endl;
    
    Assert(phaseerrors.size() == diplo_testing.diplos.size());
  }

  HaploModel::Model model(data);
  if (default_model)
    model.SetDefault();
  else
  {
    Assert(!model_file.empty());
    istream *in = InFileStream(RelPath(output_dir, model_file));
    AssertMsg(in, RelPath(output_dir, model_file));
    cerr << "* Reading model from " << RelPath(output_dir, model_file) << "...";
    model.Read(*in);
    delete in;
    cerr << " done." << endl;
  }

  if (set_default_pop_recomb)
    model.SetDefaultPopRecomb(default_pop_recomb);
  if (set_default_pop_admix)
    model.SetDefaultPopAdmix(default_pop_admix);
  
  cerr << "* Model:" << endl;
  model.DisplayEmission(cerr);
  model.DisplayPopInit(cerr);
  model.DisplayPopRecomb(cerr);
  model.DisplayPopAdmix(cerr);

  if (smooth_posterior)
    InitSmooth(data, smooth_stddev);

  vector< vector< vector<float> > > pop_ps;
  vector< vector<int8> > posterior_pops, viterbi_pops;
  vector< vector<float> > phasing_errors;

  if (job_num < 0)
  {
    TestHaploPop(data, testing_data, model, haplo_testing, haplo_composite,
                 pop_ps, posterior_pops, viterbi_pops);
    TestDiploPop(data, testing_data, model, diplo_testing, diplo_composite,
                 phases, phaseerrors, pop_ps, posterior_pops, viterbi_pops, 
                 phasing_errors);
  }
  else
  {
    Assert((job_num >= 0) && (job_num < num_jobs));
    Assert(haplo_testing.size() == 0);

    Assert(diplos.empty());
    int start = (diplo_testing.size() * job_num) / num_jobs;
    int end = (diplo_testing.size() * (job_num + 1)) / num_jobs;
    for (int d = start; d < end; ++d)
      diplos.push_back(d);
    TestDiploPop(data, testing_data, model, diplo_testing, diplo_composite,
                 phases, phaseerrors, pop_ps, posterior_pops, viterbi_pops, 
                 phasing_errors);
  }

  if (!posterior_file.empty())
  {
    ostream *out = OutFileStream(RelPath(output_dir, posterior_file));
    AssertMsg(out, RelPath(output_dir, posterior_file));
    Write(*out, pop_ps);
    delete out;
  }

  if (!posterior_popdata_file.empty())
  {
    ostream *out = OutFileStream(RelPath(output_dir, posterior_popdata_file));
    AssertMsg(out, RelPath(output_dir, posterior_popdata_file));
    WriteBin(*out, posterior_pops);
    delete out;
  }

  if (!viterbi_popdata_file.empty())
  {
    ostream *out = OutFileStream(RelPath(output_dir, viterbi_popdata_file));
    AssertMsg(out, RelPath(output_dir, viterbi_popdata_file));
    WriteBin(*out, viterbi_pops);
    delete out;
  }

  if (!phaseposterior_file.empty())
  {
    ostream *out = OutFileStream(RelPath(output_dir, phaseposterior_file));
    AssertMsg(out, RelPath(output_dir, phaseposterior_file));
    WriteBin(*out, phasing_errors);
    delete out;
  }

  cerr << "Done!" << endl;
  return 0;
}
