
#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"


Params params(false);

int debug = 0;

string input_dir;
string input_file;

string required_param[] =
{
  "input_file"
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

string BinFileType(string filename)
{
  Assert(GetExtension(RemoveCompressedExtension(filename)) == "bin");
  return GetExtension(RemoveExtension(RemoveCompressedExtension(filename)));
}

void InitParams()
{
  params.ExpandImports();

  uint32 seed = RandomSeed();
  if (params.Contains("rand_seed"))
    seed = params["rand_seed"].GetInt();
  srand(seed);
  cerr << "- Random seed = " << seed << endl;

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("input_file"))
  {
    input_file = params["input_file"].GetString();
    Assert((BinFileType(input_file) == "snpprofile") ||
           (BinFileType(input_file) == "snpprofiledata"));
  }
  
  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;

  if (!input_file.empty())
    cerr << "- Input file = " << input_file << endl;
}


int Main(vector<string> args)
{
  cerr << "SNPInfo" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;

  SNPProfile profile;
  SNPData data;

  cerr << "* Reading file...";
  if (BinFileType(input_file) == "snpprofile")
  {
    istream *in = InFileStream(RelPath(input_dir, input_file));
    AssertMsg(in, RelPath(input_dir, input_file));
    profile.ReadBin(*in);
    delete in;
  }
  else if (BinFileType(input_file) == "snpprofiledata")
  {
    istream *in = InFileStream(RelPath(input_dir, input_file));
    AssertMsg(in, RelPath(input_dir, input_file));
    ReadBinSNPProfileData(*in, profile, data);
    delete in;
  }
  cerr << " done." << endl;

  cout << "hash\t" << ToHex(profile.ComputeHash(), true) << endl;
  cout << "seq_len\t" << profile.loci.size() << endl;
  cout << "sry_genomic_pos\t" << profile.sry_genomic_pos << endl;
  cout << "x_y_par2_genomic_pos_delta\t" << profile.x_y_par2_genomic_pos_delta << endl;
  
  cout << "chrom_count\t" << profile.chroms.size() << endl;
  for (int chrom = 0; chrom < profile.chroms.size(); ++chrom)
    cout << "chrom\t" << chrom << "\t" << profile.chroms[chrom].label << "\t" << profile.chroms[chrom].range.left() << "\t" << profile.chroms[chrom].range.right() << "\t" << profile.chroms[chrom].range.len() << endl;

  cout << "par1\t" << profile.par1_range.left() << "\t" << profile.par1_range.right() << "\t" << profile.par1_range.len() << endl;
  cout << "par2\t" << profile.par2_range.left() << "\t" << profile.par2_range.right() << "\t" << profile.par2_range.len() << endl;
  cout << "x_nonpar\t" << profile.x_nonpar_range.left() << "\t" << profile.x_nonpar_range.right() << "\t" << profile.x_nonpar_range.len() << endl;
  cout << "y_nonpar\t" << profile.y_nonpar_range.left() << "\t" << profile.y_nonpar_range.right() << "\t" << profile.y_nonpar_range.len() << endl;
  cout << "mito\t" << profile.mito_range.left() << "\t" << profile.mito_range.right() << "\t" << profile.mito_range.len() << endl;

  if (params.Contains("show_loci"))
  {
    for (int i = 0; i < profile.loci.size(); ++i)
    {
      cout << "locus\t" << i << "\t" << profile.loci[i].label << "\t" << profile.loci[i].chrom << "\t" << (profile.loci[i].strand ? '-' : '+') << "\t" << profile.loci[i].genomic_pos << "\t" << ToStr(profile.loci[i].genetic_pos, 12, 8) << "\t" << ToStr(profile.loci[i].gender_genetic_pos[0], 12, 8) << "\t" << ToStr(profile.loci[i].gender_genetic_pos[1], 12, 8) << "\t" << ((i > 0) ? ToStr(profile.loci[i].genetic_pos - profile.loci[i - 1].genetic_pos, 12, 8) : "n/a") << endl;
    }
  }

  if (BinFileType(input_file) == "snpprofiledata")
  {
    cout << "pop_count\t" << data.pops.size() << endl;
    for (int pop = 0; pop < data.pops.size(); ++pop)
      cout << "pop\t" << pop << "\t" << data.pops[pop].label << "\t" << data.pops[pop].haplos.size() << "\t" << data.pops[pop].diplos.size() << "\t\"" << IntegerListString(data.pops[pop].haplos.begin(), data.pops[pop].haplos.end()) << "\"\t\"" << IntegerListString(data.pops[pop].diplos.begin(), data.pops[pop].diplos.end()) << "\"" << endl;

    cout << "haplo_count\t" << data.haplos.size() << endl;
    for (int k = 0; k < data.haplos.size(); ++k)
      cout << "haplo\t" << k << "\t" << data.haplos[k].label << "\t" << data.haplos[k].pop << "\t" << data.haplos[k].sry << endl;

    cout << "diplo_count\t" << data.diplos.size() << endl;
    for (int d = 0; d < data.diplos.size(); ++d)
      cout << "diplo\t" << d << "\t" << data.diplos[d].label << "\t" << data.diplos[d].haplos[0] << "\t" << data.diplos[d].haplos[1] << "\t" << (data.haplos[data.diplos[d].haplos[0]].sry + data.haplos[data.diplos[d].haplos[1]].sry) << endl;
  }

  cerr << "Done!" << endl;
  return 0;
}
