
#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"


Params params(false);

int debug = 0;

string input_dir;
string snpprofile_file;
string true_file, guess_file;
string output_dir;
bool diplo;
bool minor = false;
double match_stddev = 0.01;
bool do_haplos = false;
bool tabulate = false;
bool ignore_missing = false;
vector<int> haplos, diplos;
bool no_ends = false;
bool no_gaussian = false;
bool random_pause = false;

string required_param[] =
{
  "snpprofile_file",
  "true_file",
  "guess_file"
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

void GetIntegerSpec(string s, vector<int> &result);

void InitParams()
{
  params.ExpandImports();

  uint32 seed = RandomSeed();
  if (params.Contains("rand_seed"))
    seed = params["rand_seed"].GetInt();
  srand(seed);
  cerr << "- Random seed = " << seed << endl;

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (params.Contains("snpprofile_file"))
  {
    snpprofile_file = params["snpprofile_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(snpprofile_file)) == "bin");
    Assert((GetExtension(RemoveExtension(RemoveCompressedExtension(snpprofile_file))) == "snpprofile") || (GetExtension(RemoveExtension(RemoveCompressedExtension(snpprofile_file))) == "snpprofiledata"));
  }

  if (params.Contains("true_file"))
  {
    true_file = params["true_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(true_file)) == "bin");
    Assert((GetExtension(RemoveExtension(RemoveCompressedExtension(true_file))) == "haplopopdata") || (GetExtension(RemoveExtension(RemoveCompressedExtension(true_file))) == "diplopopdata"));
  }

  if (params.Contains("guess_file"))
  {
    guess_file = params["guess_file"].GetString();
    if (GetExtension(RemoveCompressedExtension(guess_file)) == "bin")
      Assert((GetExtension(RemoveExtension(RemoveCompressedExtension(guess_file))) == "haploposterior") || (GetExtension(RemoveExtension(RemoveCompressedExtension(guess_file))) == "diploposterior"));
    else if (GetExtension(RemoveCompressedExtension(guess_file)) == "lst")
      Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(guess_file))) == "saberposterior");
    else
      Assert(false);
  }

  if (GetExtension(RemoveExtension(RemoveCompressedExtension(true_file))) == "haplopopdata")
    diplo = false;
  else
    diplo = true;

  Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(true_file))) == GetExtension(RemoveExtension(RemoveCompressedExtension(true_file))));

  if (params.Contains("minor"))
    minor = true;

  if (params.Contains("match_stddev"))
    match_stddev = params["match_stddev"].GetDouble();

  if (params.Contains("do_haplos"))
    do_haplos = true;

  if (params.Contains("tabulate"))
    tabulate = true;

  if (params.Contains("ignore_missing"))
    ignore_missing = true;

  if (params.Contains("haplos"))
  {
    Assert(!diplo);
    GetIntegerSpec(params["haplos"].GetString(), haplos);
  }

  if (params.Contains("diplos"))
  {
    Assert(diplo);
    GetIntegerSpec(params["diplos"].GetString(), diplos);
  }

  if (params.Contains("no_ends"))
    no_ends = true;

  if (params.Contains("no_gaussian"))
    no_gaussian = true;

  if (params.Contains("random_pause"))
    random_pause = true;

  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;

  cerr << "- SNPProfile file = " << snpprofile_file << endl;

  cerr << "- True file = " << true_file << endl;

  cerr << "- Guess file = " << guess_file << endl;

  if (!output_dir.empty())
    cerr << "- Output directory = " << output_dir << endl;

  if (minor)
    cerr << "- Identify minor population" << endl;

  cerr << "- Match stddev = " << match_stddev << endl;
  
  if (do_haplos)
    cerr << "- Do Haplotypes" << endl;

  if (tabulate)
    cerr << "- Tabulate stdout" << endl;

  if (ignore_missing)
    cerr << "- Ignore missing data" << endl;

  if (!haplos.empty())
    cerr << "- Selected haplotypes { " << IntegerListString(haplos.begin(), haplos.end()) << " }" << endl;

  if (!diplos.empty())
    cerr << "- Selected diploid individuals { " << IntegerListString(diplos.begin(), diplos.end()) << " }" << endl;

  if (no_ends)
    cerr << "- No chromosome-end edges" << endl;

  if (no_gaussian)
    cerr << "- No Gaussian scoring" << endl;
}


void GetIntegerSpec(string s, vector<int> &result)
{
  int p = 0;
  while (p < s.length())
  {
    if (isdigit(s[p]))
    {
      int i;
      Assert(StringToInt(s, i, p, p));
      if ((p == s.length()) || (s[p] == ','))
      {
        result.push_back(i);
        if (p < s.length())
          ++p;
        continue;
      }

      Assert(s[p++] == '-');
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p == s.length()) || (s[p] == ','));
      if (i <= j)
      {
        for (int k = i; k <= j; ++k)
          result.push_back(k);
      }
      else
      {
        for (int k = i; k >= j; --k)
          result.push_back(k);
      }
      if (p < s.length());
      ++p;
    }
    else if (s[p] == '(')
    {
      int i;
      Assert(StringToInt(s, i, p, ++p));
      Assert((p < s.length()) && (s[p] == ','));
      ++p;
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p < s.length()) && (s[p] == ')'));
      Assert(j >= 0);
      for (int k = 0; k < j; ++k)
        result.push_back(i + k);
      ++p;
      Assert((p == s.length()) || (s[p] == ','));
      if (p < s.length())
        ++p;
    }
    else
      Assert(false);
  }
}


double sqr(double x)
{
  return x*x;
}


bool GaussianInRange(double Z)
{
  return (Z >= -3.0) && (Z <= 3.0);
}

double Gaussian(double Z)
{
  return exp(-0.5*Z*Z);
}


struct Edge
{
  int pos;
  int from, to;

  Edge() { }
  Edge(int pos_, int from_, int to_) : pos(pos_), from(from_), to(to_) { }

  bool SameType(const Edge &other) const
  {
    return (from == other.from) && (to == other.to);
  }

  bool SameTo(const Edge &other) const
  {
    return (to == other.to);
  }
  
  bool operator<(const Edge &other) const
  {
    if (pos != other.pos)
      return pos < other.pos;
    else if (from != other.from)
      return from < other.from;
    else
      return to < other.to;
  }
};


double EdgeScore(const SNPProfile &profile, double match_stddev, int i, int j)
{
  double genetic_sep = profile.loci[j].genetic_pos - profile.loci[i].genetic_pos;
  if (no_gaussian)
  {
    if (fabs(genetic_sep) < match_stddev)
      return 1.0;
    else
      return 0.0;
  }
  else
  {
    double Z = genetic_sep / match_stddev;
    double f = exp(-0.5*Z*Z);
    return f;
  }
}


void HaploEdgeSensSpec(const SNPProfile &profile, double match_stddev,
                       const vector<Edge> &true_edges,
                       const vector<Edge> &guess_edges,
                       SensSpec<double> &edge_sens_spec)
{
  vector< vector<double> > v(true_edges.size(), vector<double>(guess_edges.size(), 0.0));
  vector< vector<double> > m(true_edges.size() + 1, vector<double>(guess_edges.size() + 1, 0.0));
  vector< vector<int> > ptr(true_edges.size() + 1, vector<int>(guess_edges.size() + 1, -1));

  for (int i = 0; i < true_edges.size(); ++i)
    for (int j = 0; j < guess_edges.size(); ++j)
    {
      int max_dir = 0;
      double max_score = m[(i - 1) + 1][j + 1];

      {
        double score = m[i + 1][(j - 1) + 1];
        if (score > max_score)
        {
          max_score = score;
          max_dir = 1;
        }
      }

      if (true_edges[i].SameTo(guess_edges[j]))
        v[i][j] = EdgeScore(profile, match_stddev, true_edges[i].pos, guess_edges[j].pos);
      else
        v[i][j] = 0.0;
      
      {
        double score = m[(i - 1) + 1][(j - 1) + 1] + v[i][j];
        if (score > max_score)
        {
          max_score = score;
          max_dir = 2;
        }
      }

      m[i + 1][j + 1] = max_score;
      ptr[i + 1][j + 1] = max_dir;
    }

  vector<int> true_match(true_edges.size(), -1);
  vector<int> guess_match(guess_edges.size(), -1);
  for (int i = int(true_edges.size()) - 1, j = int(guess_edges.size()) - 1;
       (i >= 0) && (j >= 0); )
  {
    switch (ptr[i + 1][j + 1])
    {
    case 0:
      --i;
      break;

    case 1:
      --j;
      break;

    case 2:
      true_match[i] = j;
      guess_match[j] = i;
      --i;
      --j;
      break;

    default:
      Assert(false);
    }
  }

  for (int i = 0; i < true_edges.size(); ++i)
  {
    if (true_match[i] < 0)
      edge_sens_spec(1, 0) += 1.0;
    else
#if 0
      edge_sens_spec(1, 1) += v[i][true_match[i]];
#else
    {
      double score = v[i][true_match[i]];
      edge_sens_spec(1, 1) += score;
      edge_sens_spec(1, 0) += 1.0 - score;
      edge_sens_spec(0, 1) += 1.0 - score;
    }
#endif
  }

  for (int j = 0; j < guess_edges.size(); ++j)
    if (guess_match[j] < 0)
      edge_sens_spec(0, 1) += 1.0;
}


void DiploEdgeSensSpec(const SNPProfile &profile, double match_stddev,
                       const vector<Edge> &true_edges0,
                       const vector<Edge> &true_edges1,
                       const vector<Edge> &guess_edges0,
                       const vector<Edge> &guess_edges1,
                       SensSpec<double> &edge_sens_spec)
{
  vector< vector<double> > v00(true_edges0.size(), vector<double>(guess_edges0.size(), 0.0));
  vector< vector<double> > v01(true_edges0.size(), vector<double>(guess_edges1.size(), 0.0));
  vector< vector<double> > v10(true_edges1.size(), vector<double>(guess_edges0.size(), 0.0));
  vector< vector<double> > v11(true_edges1.size(), vector<double>(guess_edges1.size(), 0.0));
  vector< vector< vector< vector<double> > > > m(true_edges0.size() + 1, vector< vector< vector<double> > >(true_edges1.size() + 1, vector< vector<double> >(guess_edges0.size() + 1, vector<double>(guess_edges1.size() + 1, 0.0))));
  vector< vector< vector< vector<int> > > > ptr(true_edges0.size() + 1, vector< vector< vector<int> > >(true_edges1.size() + 1, vector< vector<int> >(guess_edges0.size() + 1, vector<int>(guess_edges1.size() + 1, -1))));

  for (int i0 = -1; i0 < int(true_edges0.size()); ++i0)
    for (int i1 = -1; i1 < int(true_edges1.size()); ++i1)
      for (int j0 = -1; j0 < int(guess_edges0.size()); ++j0)
        for (int j1 = -1; j1 < int(guess_edges1.size()); ++j1)
        {
          int max_dir = -1;
          double max_score = -1e100;
          double score;

          if ((i0 >= 0) && ((score = m[(i0 - 1) + 1][i1 + 1][j0 + 1][j1 + 1]) > max_score))
            max_score = score, max_dir = 0;

          if ((i1 >= 0) && ((score = m[i0 + 1][(i1 - 1) + 1][j0 + 1][j1 + 1]) > max_score))
            max_score = score, max_dir = 1;

          if ((j0 >= 0) && ((score = m[i0 + 1][i1 + 1][(j0 - 1) + 1][j1 + 1]) > max_score))
            max_score = score, max_dir = 2;

          if ((j1 >= 0) && ((score = m[i0 + 1][i1 + 1][j0 + 1][(j1 - 1) + 1]) > max_score))
            max_score = score, max_dir = 3;

          if ((i0 >= 0) && (j0 >= 0))
          {
            v00[i0][j0] = true_edges0[i0].SameTo(guess_edges0[j0]) ? EdgeScore(profile, match_stddev, true_edges0[i0].pos, guess_edges0[j0].pos) : 0.0;
            if ((score = m[(i0 - 1) + 1][i1 + 1][(j0 - 1) + 1][j1 + 1] + v00[i0][j0]) > max_score)
              max_score = score, max_dir = 4;
          }

          if ((i0 >= 0) && (j1 >= 0))
          {
            v01[i0][j1] = true_edges0[i0].SameTo(guess_edges1[j1]) ? EdgeScore(profile, match_stddev, true_edges0[i0].pos, guess_edges1[j1].pos) : 0.0;
            if ((score = m[(i0 - 1) + 1][i1 + 1][j0 + 1][(j1 - 1) + 1] + v01[i0][j1]) > max_score)
              max_score = score, max_dir = 5;
          }

          if ((i1 >= 0) && (j0 >= 0))
          {
            v10[i1][j0] = true_edges1[i1].SameTo(guess_edges0[j0]) ? EdgeScore(profile, match_stddev, true_edges1[i1].pos, guess_edges0[j0].pos) : 0.0;
            if ((score = m[i0 + 1][(i1 - 1) + 1][(j0 - 1) + 1][j1 + 1] + v10[i1][j0]) > max_score)
              max_score = score, max_dir = 6;
          }

          if ((i1 >= 0) && (j1 >= 0))
          {
              v11[i1][j1] = true_edges1[i1].SameTo(guess_edges1[j1]) ? EdgeScore(profile, match_stddev, true_edges1[i1].pos, guess_edges1[j1].pos) : 0.0;
              if ((score = m[i0 + 1][(i1 - 1) + 1][j0 + 1][(j1 - 1) + 1] + v11[i1][j1]) > max_score)
                max_score = score, max_dir = 7;
          }

          if (max_dir < 0)
          {
            m[i0 + 1][i1 + 1][j0 + 1][j1 + 1] = 0.0;
            ptr[i0 + 1][i1 + 1][j0 + 1][j1 + 1] = -1;
          }
          else
          {
            m[i0 + 1][i1 + 1][j0 + 1][j1 + 1] = max_score;
            ptr[i0 + 1][i1 + 1][j0 + 1][j1 + 1] = max_dir;
          }
        }

  vector<int> true_match_haplo0(true_edges0.size(), -1), true_match0(true_edges0.size(), -1);
  vector<int> true_match_haplo1(true_edges1.size(), -1), true_match1(true_edges1.size(), -1);
  vector<int> guess_match_haplo0(guess_edges0.size(), -1), guess_match0(guess_edges0.size(), -1);
  vector<int> guess_match_haplo1(guess_edges1.size(), -1), guess_match1(guess_edges1.size(), -1);

  for (int i0 = int(true_edges0.size()) - 1, i1 = int(true_edges1.size()) - 1,
         j0 = int(guess_edges0.size()) - 1, j1 = int(guess_edges1.size()) - 1;
       (i0 >= 0) || (i1 >= 0) || (j0 >= 0) || (j1 >= 0); )
  {
    switch (ptr[i0 + 1][i1 + 1][j0 + 1][j1 + 1])
    {
    case 0:
      --i0;
      break;

    case 1:
      --i1;
      break;

    case 2:
      --j0;
      break;

    case 3:
      --j1;
      break;

    case 4:
      true_match_haplo0[i0] = 0, true_match0[i0] = j0;
      guess_match_haplo0[j0] = 0, guess_match0[j0] = i0;
      --i0, --j0;
      break;

    case 5:
      true_match_haplo0[i0] = 1, true_match0[i0] = j1;
      guess_match_haplo1[j1] = 0, guess_match1[j1] = i0;
      --i0, --j1;
      break;

    case 6:
      true_match_haplo1[i1] = 0, true_match1[i1] = j0;
      guess_match_haplo0[j0] = 1, guess_match0[j0] = i1;
      --i1, --j0;
      break;

    case 7:
      true_match_haplo1[i1] = 1, true_match1[i1] = j1;
      guess_match_haplo1[j1] = 1, guess_match1[j1] = i1;
      --i1, --j1;
      break;

    default:
      AssertMsg(false, ToStr(ptr[i0 + 1][i1 + 1][j0 + 1][j1 + 1]) + ", " + ToStr(i0) + ", " + ToStr(i1) + ", " + ToStr(j0) + ", " + ToStr(j1));
    }
  }

#if 0
  cerr << endl;
  cerr << "true_edges0:" << endl;
  for (int i0 = 0; i0 < true_edges0.size(); ++i0)
    cerr << true_match_haplo0[i0] << ", " << true_match0[i0] << endl;
  cerr << "true_edges1:" << endl;
  for (int i1 = 0; i1 < true_edges1.size(); ++i1)
    cerr << true_match_haplo1[i1] << ", " << true_match1[i1] << endl;
  cerr << "guess_edges0:" << endl;
  for (int i0 = 0; i0 < guess_edges0.size(); ++i0)
    cerr << guess_match_haplo0[i0] << ", " << guess_match0[i0] << endl;
  cerr << "guess_edges1:" << endl;
  for (int i1 = 0; i1 < guess_edges1.size(); ++i1)
    cerr << guess_match_haplo1[i1] << ", " << guess_match1[i1] << endl;
#endif

  for (int i0 = 0; i0 < true_edges0.size(); ++i0)
    if (true_match0[i0] < 0)
      edge_sens_spec(1, 0) += 1.0;
    else
#if 0
      edge_sens_spec(1, 1) += ((true_match_haplo0[i0] == 0) ? v00 : v01)[i0][true_match0[i0]];
#else
    {
      double score = ((true_match_haplo0[i0] == 0) ? v00 : v01)[i0][true_match0[i0]];
      edge_sens_spec(1, 1) += score;
      edge_sens_spec(1, 0) += 1.0 - score;
      edge_sens_spec(0, 1) += 1.0 - score;
    }
#endif

  for (int i1 = 0; i1 < true_edges1.size(); ++i1)
    if (true_match1[i1] < 0)
      edge_sens_spec(1, 0) += 1.0;
    else
#if 0
      score edge_sens_spec(1, 1) += ((true_match_haplo1[i1] == 0) ? v10 : v11)[i1][true_match1[i1]];
#else
    {
      double score = ((true_match_haplo1[i1] == 0) ? v10 : v11)[i1][true_match1[i1]];
      edge_sens_spec(1, 1) += score;
      edge_sens_spec(1, 0) += 1.0 - score;
      edge_sens_spec(0, 1) += 1.0 - score;
    }
#endif

  for (int j0 = 0; j0 < guess_edges0.size(); ++j0)
    if (guess_match0[j0] < 0)
      edge_sens_spec(0, 1) += 1.0;

  for (int j1 = 0; j1 < guess_edges1.size(); ++j1)
    if (guess_match1[j1] < 0)
      edge_sens_spec(0, 1) += 1.0;
}


template <typename T>
void DecodeEdges(const SNPProfile &profile, const vector<T> &pops, 
                 vector<Edge> &edges)
{
  edges.clear();
  for (int chrom = 0; chrom < profile.num_chroms; ++chrom)
  {
    if (profile.chroms[chrom].range.len() == 0)
      continue;

    int left = profile.chroms[chrom].range.left();
    int right = profile.chroms[chrom].range.right();
    edges.push_back(Edge(left, -1, pops[left]));
    for (int i = left + 1; i < right; ++i)
    {
      if (pops[i] != pops[i - 1])
        edges.push_back(Edge(i, pops[i - 1], pops[i]));
    }
  }
}


template <typename T>
void DecodeEdgesNoEnds(const SNPProfile &profile, const vector<T> &pops, 
                       vector<Edge> &edges)
{
  edges.clear();
  //edges.push_back(Edge(0, -1, pops[0]));
  for (int i = 1; i < pops.size(); ++i)
  {
    if (pops[i] != pops[i - 1])
      edges.push_back(Edge(i, pops[i - 1], pops[i]));
  }
}



int MajorPop(const vector<int8> &pops)
{
  Assert(!pops.empty());
  int max_pop = -1;
  for (int i = 0; i < pops.size(); ++i)
    if (pops[i] > max_pop)
      max_pop = pops[i];

  vector<int> counts(max_pop + 1, 0);
  for (int i = 0; i < pops.size(); ++i)
    ++counts[pops[i]];
  
  int max_count = 0;
  for (int c = 1; c < counts.size(); ++c)
    if (counts[c] > counts[max_count])
      max_count = c;

  return max_count;
}


int MajorPop(const vector<int8> &pops0, const vector<int8> &pops1)
{
  Assert(!pops0.empty() && !pops1.empty());
  Assert(pops0.size() == pops1.size());
  int max_pop = -1;
  for (int i = 0; i < pops0.size(); ++i)
  {
    if (pops0[i] > max_pop)
      max_pop = pops0[i];
    if (pops1[i] > max_pop)
      max_pop = pops1[i];
  }

  vector<int> counts(max_pop + 1, 0);
  for (int i = 0; i < pops0.size(); ++i)
  {
    ++counts[pops0[i]];
    ++counts[pops1[i]];
  }
  
  int max_count = 0;
  for (int c = 1; c < counts.size(); ++c)
    if (counts[c] > counts[max_count])
      max_count = c;

  return max_count;
}



void HaploSNPAccuracy(const vector<int8> &truth, const vector<int8> &guess,
                      SensSpec<int> &snp_accuracy)
{
  Assert(truth.size() == guess.size());
  for (int i = 0; i < truth.size(); ++i)
  {
    if (truth[i] == guess[i])
      ++snp_accuracy(1, 1);
    else
      ++snp_accuracy(0, 1), ++snp_accuracy(1, 0);
  }
}


void HaploSNPSensSpec(const vector<int8> &truth, const vector<int8> &guess,
                      int major, SensSpec<int> &snp_sens_spec)
{
  Assert(truth.size() == guess.size());
  for (int i = 0; i < truth.size(); ++i)
  {
    if (truth[i] == major)
      ++snp_sens_spec(0, (guess[i] == major) ? 0 : 1);
    else
      ++snp_sens_spec(1, (guess[i] == truth[i]) ? 1 : 0);
  }
}


void DiploSNPAccuracy(const vector<int8> &truth0, const vector<int8> &truth1,
                      const vector<int8> &guess0, const vector<int8> &guess1,
                      SensSpec<int> &snp_accuracy)
{
  Assert(truth0.size() == truth1.size());
  Assert(guess0.size() == guess1.size());
  Assert(truth0.size() == guess0.size());
  for (int i = 0; i < truth0.size(); ++i)
  {
    int matches0 = ((truth0[i] == guess0[i]) ? 1 : 0) + ((truth1[i] == guess1[i]) ? 1 : 0);
    int matches1 = ((truth0[i] == guess1[i]) ? 1 : 0) + ((truth1[i] == guess0[i]) ? 1 : 0);
    int matches = max(matches0, matches1);
    snp_accuracy(1, 1) += matches;
    snp_accuracy(0, 1) += (2 - matches);
    snp_accuracy(1, 0) += (2 - matches);
  }
}


void DiploSNPSensSpec(const vector<int8> &truth0, const vector<int8> &truth1,
                      const vector<int8> &guess0, const vector<int8> &guess1,
                      int major, SensSpec<int> &snp_sens_spec)
{
  Assert(truth0.size() == truth1.size());
  Assert(guess0.size() == guess1.size());
  Assert(truth0.size() == guess0.size());
  for (int i = 0; i < truth0.size(); ++i)
  {
    int true_signal = ((truth0[i] != major) ? 1 : 0) + ((truth1[i] != major) ? 1 : 0);
    if ((true_signal == 0) || (true_signal == 2))
    {
      ++snp_sens_spec((true_signal == 2) ? 1 : 0, (guess0[i] != major) ? 1 : 0);
      ++snp_sens_spec((true_signal == 2) ? 1 : 0, (guess1[i] != major) ? 1 : 0);
    }
    else
    {
      ++snp_sens_spec(0, ((guess0[i] == major) || (guess1[i] == major)) ? 0 : 1);
      
      int signal = (truth0[i] == major) ? truth1[i] : truth0[i];
      ++snp_sens_spec(1, ((guess0[i] == signal) || (guess1[i] == signal)) ? 1 : 0);
    }
  }
}





int Main(vector<string> args)
{
  cerr << "ScoreHaploModelMSE" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;

  if (random_pause)
  {
    int pause = rand() % 30;
    cerr << "* Pausing for " << pause << " seconds" << endl;
    sleep(pause);
  }

  SNPProfile profile;
  {
    istream *in = InFileStream(RelPath(input_dir, snpprofile_file));
    if (!in)
    {
      cerr << "Could not open file \"" << snpprofile_file << "\"";
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"";
      cerr << endl;
      Exit(1);
    }

    cerr << "* Reading SNPProfile...";
    profile.ReadBin(*in);
    cerr << " done." << endl;
    delete in;

    profile.Validate();
  }

  vector< vector<int8> > true_pop;
  {
    istream *in = InFileStream(RelPath(input_dir, true_file));
    AssertMsg(in, RelPath(input_dir, true_file));
    ReadBin(*in, true_pop);
    delete in;
    
    for (int i = 0; i < true_pop.size(); ++i)
      Assert(true_pop[i].size() == profile.seq_len);
  }

  vector< vector< vector<float> > > guess_pop_ps;
  {
    if (GetExtension(RemoveCompressedExtension(guess_file)) == "bin")
    {
      istream *in = InFileStream(RelPath(input_dir, guess_file));
      AssertMsg(in, RelPath(input_dir, guess_file));
      ReadBin(*in, guess_pop_ps);
      for (int i = 0; i < guess_pop_ps.size(); ++i)
        for (int pop = 0; pop < guess_pop_ps[i].size(); ++pop)
          Assert(guess_pop_ps[i][pop].size() == profile.seq_len);
      delete in;
    }
    else if (GetExtension(RemoveCompressedExtension(guess_file)) == "lst")
    {
      string path, dummy;
      SplitPath(RelPath(input_dir, guess_file), path, dummy);
      vector<string> files;
      {
        istream *in = InFileStream(RelPath(input_dir, guess_file));
        AssertMsg(in, RelPath(input_dir, guess_file));
        string line;
        while (getline(*in, line))
          files.push_back(line);
        delete in;
      }
      guess_pop_ps.reserve(files.size());
      
      for (int file = 0; file < files.size(); ++file)
      {
        if (ignore_missing && files[file].empty())
          continue;
        
        istream *in = InFileStream(RelPath(path, files[file]));
        if (ignore_missing && !in)
          continue;
        AssertMsg(in, RelPath(path, files[file]));
        
        guess_pop_ps.resize(guess_pop_ps.size() + 1);
        guess_pop_ps[file].reserve(52);
        
        string line;
        vector<string> fields;
        vector<double> pop_p;
        int i = 0;
        while (getline(*in, line))
        {
          SplitFields(line, fields, " \t,;");
          if (fields.empty())
            continue;
          Assert(i < profile.seq_len);
          if (i == 0)
          {
            Assert(guess_pop_ps[file].empty());
            guess_pop_ps[file].resize(fields.size(), vector<float>(profile.seq_len, 0.0));
          }
          for (int pop = 0; pop < fields.size(); ++pop)
          {
            StrTo(fields[pop], guess_pop_ps[file][pop][i]);
            Assert((guess_pop_ps[file][pop][i] >= 0.0) && (guess_pop_ps[file][pop][i] <= 1.0));
          }
          ++i;
        }
        delete in;
        Assert(i == profile.seq_len);
      }

      if (diplo)
      {
        vector< vector< vector<float> > > temp(guess_pop_ps.size() * 2);
        for (int k = 0; k < guess_pop_ps.size(); ++k)
        {
          temp[k*2 + 0] = guess_pop_ps[k];
          temp[k*2 + 1] = guess_pop_ps[k];
        }
        guess_pop_ps.swap(temp);
      }
    }
    else
      Assert(false);
  }

  if (!ignore_missing)
    AssertMsg(true_pop.size() == guess_pop_ps.size(), ToStr(true_pop.size()) + ", " + ToStr(guess_pop_ps.size()));
  else
  {
    cerr << "* Resizing to " << min(true_pop.size(), guess_pop_ps.size()) << endl;
    true_pop.resize(min(true_pop.size(), guess_pop_ps.size()));
    guess_pop_ps.resize(min(true_pop.size(), guess_pop_ps.size()));
  }

  int num_pops;
  if (!diplo || do_haplos)
  {
    if (haplos.empty())
    {
      int num_samples = true_pop.size();
      for (int k = 0; k < num_samples; ++k)
        haplos.push_back(k);
    }
    num_pops = guess_pop_ps[0].size();
  }
  else
  {
    if (diplos.empty())
    {
      int num_samples = true_pop.size() / 2;
      Assert(true_pop.size() == (num_samples * 2));
      for (int d = 0; d < num_samples; ++d)
        diplos.push_back(d);
    }
    num_pops = guess_pop_ps[0].size();
  }
    
  double avg_mse = 0.0;
  vector<double> avg_pop_mse(num_pops, 0.0);

  if (tabulate)
  {
    cout << "index\tmse";
    for (int pop = 0; pop < num_pops; ++pop)
      cout << "\t" << "mse(" << pop << ")";
    cout << endl;
  }
  
  if (!diplo || do_haplos)
  {
    for (int index = 0; index < haplos.size(); ++index)
    {
      int k = haplos[index];
      Assert(k < true_pop.size());
      if (ignore_missing && guess_pop_ps[k].empty())
        continue;
      cerr << "* Scoring haplotype " << k << ":";
      
      double mse = 0.0;
      vector<double> pop_mse(num_pops, 0.0);
      vector<double> true_values(num_pops, 0.0);

      for (int i = 0; i < true_pop[k].size(); ++i)
      {
        for (int pop = 0; pop < num_pops; ++pop)
          true_values[pop] = 0.0;
        true_values[true_pop[k][i]] = 1.0;

        for (int pop = 0; pop < num_pops; ++pop)
        {
          double delta = sqr(true_values[pop] - guess_pop_ps[k][pop][i]);
          mse += delta;
          pop_mse[pop] += delta;
        }
      }
        
      mse /= double(true_pop[k].size());
      for (int pop = 0; pop < num_pops; ++pop)
        pop_mse[pop] /= double(true_pop[k].size());
      
      cerr << " MSE = " << mse;
      for (int pop = 0; pop < num_pops; ++pop)
        cerr << ", MSE(" << pop << ") = " << pop_mse[pop];
      cerr << endl;
      
      if (tabulate)
      {
        cout << "haplo(" << k << ")";
        cout << "\t" << mse;
        for (int pop = 0; pop < num_pops; ++pop)
          cout << "\t" << pop_mse[pop];
        cout << endl;
      }
      
      avg_mse += mse;
      for (int pop = 0; pop < num_pops; ++pop)
        avg_pop_mse[pop] += pop_mse[pop];
    }
    
    avg_mse /= double(haplos.size());
    for (int pop = 0; pop < num_pops; ++pop)
      avg_pop_mse[pop] /= double(haplos.size());

    cerr << "* Average";
    cerr << " MSE = " << avg_mse;
    for (int pop = 0; pop < num_pops; ++pop)
      cerr << ", MSE(" << pop << ") = " << avg_pop_mse[pop];
    cerr << endl;

    if (tabulate)
    {
      cout << "average\t" << avg_mse;
      for (int pop = 0; pop < num_pops; ++pop)
        cout << "\t" << avg_pop_mse[pop];
      cout << endl;
    }
  }
  else
  {
    for (int index = 0; index < diplos.size(); ++index)
    {
      int d = diplos[index];
      Assert((d*2 + 1) < true_pop.size());
      if (ignore_missing && guess_pop_ps[d*2 + 0].empty())
        continue;
      cerr << "* Scoring diploid individual " << d << ":";
      
      double mse = 0.0;
      vector<double> pop_mse(num_pops, 0.0);
      vector<double> true_values(num_pops, 0.0);

      for (int i = 0; i < true_pop[d*2 + 0].size(); ++i)
      {
        for (int pop = 0; pop < num_pops; ++pop)
          true_values[pop] = 0.0;
        true_values[true_pop[d*2 + 0][i]] += 0.5;
        true_values[true_pop[d*2 + 1][i]] += 0.5;
        
        for (int pop = 0; pop < num_pops; ++pop)
        {
          double delta = sqr(true_values[pop] - 0.5*(guess_pop_ps[d*2 + 0][pop][i] + guess_pop_ps[d*2 + 1][pop][i]));
          mse += delta;
          pop_mse[pop] += delta;
        }
      }
      mse /= double(true_pop[d*2 + 0].size());
      for (int pop = 0; pop < num_pops; ++pop)
        pop_mse[pop] /= double(true_pop[d*2 + 0].size());
      
      cerr << " MSE = " << mse;
      for (int pop = 0; pop < num_pops; ++pop)
        cerr << ", MSE(" << pop << ") = " << pop_mse[pop];
      cerr << endl;

      if (tabulate)
      {
        cout << "diplo(" << d << ")\t" << mse;
        for (int pop = 0; pop < num_pops; ++pop)
          cout << "\t" << pop_mse[pop];
        cout << endl;
      }
      
      avg_mse += mse;
      for (int pop = 0; pop < num_pops; ++pop)
        avg_pop_mse[pop] += pop_mse[pop];
    }
    
    avg_mse /= double(diplos.size());
    for (int pop = 0; pop < num_pops; ++pop)
      avg_pop_mse[pop] /= double(diplos.size());

    cerr << "* Average";
    cerr << " MSE = " << avg_mse;
    for (int pop = 0; pop < num_pops; ++pop)
      cerr << ", MSE(" << pop << ") = " << avg_pop_mse[pop];
    cerr << endl;

    if (tabulate)
    {
      cout << "average\t" << avg_mse;
      for (int pop = 0; pop < num_pops; ++pop)
        cout << "\t" << avg_pop_mse[pop];
      cout << endl;
    }
  }
  
  cerr << "Done!" << endl;
  return 0;
}
