
#ifndef SEGMENT_H
#define SEGMENT_H

#include "System.h"
#include "Utility.h"

template <typename T = int>
class TSegment
{
private:
  T _left, _right;
public:
  TSegment() : _left(0), _right(0) { }

  TSegment(T coord1, T coord2, bool leftright = false) 
    : _left(coord1), _right(leftright ? coord2 : (coord1 + coord2))
  {
  }

  void swap(TSegment<T> &other)
  {
    ::swap(_left, other._left);
    ::swap(_right, other._right);
  }

  T start() const { return _left; }
  T len() const { return _right - _left; }
  T left() const { return _left; }
  T right() const { return _right; }

  void SetLeft(T left_) { _left = left_; }

  void SetRight(T right_) { _right = right_; }

  void SetStart(T start_) 
  { 
    T delta = start_ - _left;
    _right += delta;
    _left = start_;
  }

  void SetLen(T len_) { _right = _left + len_; }

  void Set(T coord1, T coord2, bool leftright = false)
  {
    _left = coord1;
    _right = (leftright ? coord2 : (coord1 + coord2));
  }

  void Translate(T offset)
  {
    _left += offset;
    _right += offset;
  }

  void SetLeftRight(T left_, T right_)
  {
    _left = left_;
    _right = right_;
  }

  void SetStartLen(T start_, T len_)
  {
    _left = start_;
    _right = start_ + len_;
  }

  bool contains(T coord) const 
  { 
    return (_left <= coord) && (coord < _right); 
  }
};


template <typename T>
bool operator<(const TSegment<T> &a, const TSegment<T> &b)
{ 
  return a.left() < b.left();
}

template <typename T>
bool operator==(const TSegment<T> &a, const TSegment<T> &b)
{
  return (a.left() == b._left()) && (a.right() == b.right());
}

template <typename T>
bool operator!=(const TSegment<T> &a, const TSegment<T> &b)
{
  return !(a == b);
}

template <typename T>
bool Intersect(const TSegment<T> &a, const TSegment<T> &b, TSegment<T> &result)
{
  if ((a.right() > b.left()) && (a.left() < b.right()))
  {
    result.SetLeft(max(a.left(), b.left()));
    result.SetRight(min(a.right(), b.right()));
    return true;
  }
  else
    return false;
}

template <typename T>
bool Intersects(const TSegment<T> &a, const TSegment<T> &b)
{
  return (a.right() > b.left()) && (a.left() < b.right());
}
  

template <typename T>
void Union(const TSegment<T> &a, const TSegment<T> &b, TSegment<T> &result)
{
  result.SetLeft(min(a.left(), b.left()));
  result.SetRight(max(a.right(), b.right()));
}

template <typename T>
bool Union(const TSegment<T> &a, const TSegment<T> &b, TSegment<T> &result, TSegment<T> &gap)
{
  result.SetLeft(min(a.left(), b.left()));
  result.SetRight(max(a.right(), b.right()));

  gap.SetLeft(min(a.right(), b.right()));
  gap.SetRight(max(a.left(), b.left()));
  if ((a.right() > b.left()) & (a.left() < b.right()))
    return false;
  else
    return true;
}

template <typename T>
bool Contains(const TSegment<T> &a, const TSegment<T> &b)
{
  return (b.left() >= a.left()) && (b.right() <= a.right());
}

template <typename T>
bool StrictContains(const TSegment<T> &a, const TSegment<T> &b)
{
  return (b.left() > a.left()) && (b.right() < a.right());
}

template <typename T>
int Subtract(const TSegment<T> &a, const TSegment<T> &b,
              TSegment<T> &result, TSegment<T> &result2)
{
  TSegment<T> intersect;
  if (!Intersect(a, b, intersect))
  {
    result = a;
    return 1;
  }

  if (a == intersect)
    return 0;

  if (a.left() == intersect.left())
  {
    result.SetLeft(intersect.right());
    result.SetRight(a.right());
    return 1;
  }

  if (a.right() == intersect.right())
  {
    result.SetLeft(a.left());
    result.SetRight(intersect.left());
    return 1;
  }

  result.SetLeft(a.left());
  result.SetRight(intersect.left());
  result2.SetLeft(intersect.right());
  result2.SetRight(a.right());
  return 2;
}

template <typename T>
int Separation(const TSegment<T> &a, const TSegment<T> &b)
{
  if (a.left() >= b.right())
    return a.left() - b.right();
  else if (a.right() <= b.left())
    return b.left() - a.right();
  else
    return 0;
}

struct GSPOPoint
{
  vector<int> lefts, rights;
};

template <typename T>
struct GSPOReorder
{
  TSegment<T> seg;
  int index;
};

template <typename T>
struct GSPOReorderLess
{
  bool operator()(const GSPOReorder<T> &a, const GSPOReorder<T> &b) const
  {
    return a.seg.left() < b.seg.left();
  }
};

template <typename T>
void GetTSegmentPairOverlaps(const vector< TSegment<T> > &segments1,
                            const vector< TSegment<T> > &segments2,
                            vector< pair<int, int> > &overlaps)
{
  overlaps.clear();

  typedef map<T, GSPOPoint> GSPOExtrema;
  GSPOExtrema extrema;

  for (int i = 0; i < segments2.size(); ++i)
  {
    extrema[segments2[i].left()].lefts.push_back(i);
    extrema[segments2[i].right()].rights.push_back(i);
  }

  vector< GSPOReorder<T> > sorted1(segments1.size());
  for (int i = 0; i < segments1.size(); ++i)
  {
    sorted1[i].seg = segments1[i];
    sorted1[i].index = i;
  }
  sort(sorted1.begin(), sorted1.end(), GSPOReorderLess<T>());

  set<int> current;
  typename GSPOExtrema::iterator current_left = extrema.begin();
  typename GSPOExtrema::iterator current_right = extrema.begin();

  for (int i = 0; i < sorted1.size(); ++i)
  {
    while (current_right != extrema.begin())
    {
      typename GSPOExtrema::iterator new_right = current_right;
      --new_right;
      if (new_right->first >= sorted1[i].seg.right())
      {
        current_right = new_right;
        for (int j = 0; j < current_right->second.lefts.size(); ++j)
          current.erase(current_right->second.lefts[j]);
      }
      else
        break;
    }

    while ((current_right != extrema.end()) && 
           (current_right->first < sorted1[i].seg.right()))
    {
      current.insert(current_right->second.lefts.begin(),
                     current_right->second.lefts.end());
      ++current_right;
    }

    while ((current_left != extrema.end()) &&
           (current_left->first <= sorted1[i].seg.left()))
    {
      for (int j = 0; j < current_left->second.rights.size(); ++j)
        current.erase(current_left->second.rights[j]);
      ++current_left;
    }

    for (set<int>::iterator iter = current.begin();
         iter != current.end(); ++iter)
      overlaps.push_back(pair<int, int>(sorted1[i].index, *iter));
  }
}

template <typename T>
void Write(ostream &o, const TSegment<T> &x)
{
  WriteBin(o, x);
}

template <typename T>
void Read(istream &i, TSegment<T> &x)
{
  ReadBin(i, x);
}

template <typename T>
struct LessTSegmentLeft
{
  bool operator()(const TSegment<T> &a, const TSegment<T> &b) const
  {
    return a.left() < b.left();
  }
};

template <typename T>
struct MoreTSegmentLeft
{
  bool operator()(const TSegment<T> &a, const TSegment<T> &b) const
  {
    return a.left() > b.left();
  }
};

template <typename T>
struct LessTSegmentRight
{
  bool operator()(const TSegment<T> &a, const TSegment<T> &b) const
  {
    return a.right() < b.right();
  }
};

template <typename T>
struct MoreTSegmentRight
{
  bool operator()(const TSegment<T> &a, const TSegment<T> &b) const
  {
    return a.right() > b.right();
  }
};

template <typename T>
struct LessTSegmentMid
{
  bool operator()(const TSegment<T> &a, const TSegment<T> &b) const
  {
    return (a.left() + a.right()) < (b.left() + b.right());
  }
};

template <typename T>
struct MoreTSegmentMid
{
  bool operator()(const TSegment<T> &a, const TSegment<T> &b) const
  {
    return (a.left() + a.right()) > (b.left() + b.right());
  }
};

template <typename T>
struct LessTSegmentLen
{
  bool operator()(const TSegment<T> &a, const TSegment<T> &b) const
  {
    return a.len() < b.len();
  }
};

template <typename T>
struct MoreTSegmentLen
{
  bool operator()(const TSegment<T> &a, const TSegment<T> &b) const
  {
    return a.len() > b.len();
  }
};


typedef TSegment<int> Segment;

#endif
