
#include "Simulate.h"


namespace Simulate
{

  void Assign(const vector<Segment> &restrict, int source,
              vector< pair<Segment, int> > &sources)
  {
    sources.reserve(sources.size() + restrict.size());
    for (int i = 0; i < restrict.size(); ++i)
      if (restrict[i].len() > 0)
        sources.push_back(pair<Segment, int>(restrict[i], source));
  }


  bool Intersect(const vector<Segment> &x, 
                 const vector<Segment> &y,
                 vector<Segment> &result)
  {
    if (x.empty() || y.empty())
      return false;

    result.reserve(result.size() + max(x.size(), y.size()));
    
    int i = 0, j = 0;
    bool intersected = false;
    while (true)
    {
      if (x[i].len() == 0)
      {
        if (++i == x.size())
          break;
      }
      else if (y[j].len() == 0)
      {
        if (++j == y.size())
          break;
      }
      else if (x[i].right() <= y[j].left())
      {
        if (++i == x.size())
          break;
      }
      else if (y[j].right() <= x[i].left())
      {
        if (++j == y.size())
          break;
      }
      else
      {
        Segment intersection;
        if (Intersect(x[i], y[j], intersection))
        {
          result.push_back(intersection);
          intersected = true;
        }
        if (x[i].right() <= y[j].right())
        {
          if (++i == x.size())
            break;
        }
        else
        {
          if (++j == y.size())
            break;
        }
      }
    }
    
    return intersected;
  }
  

  bool Intersect(const vector< pair<Segment, int> > &sources, 
                 const vector<Segment> &mask,
                 vector< pair<Segment, int> > &result)
  {
    if (sources.empty() || mask.empty())
      return false;
    
    result.reserve(result.size() + max(sources.size(), mask.size()));

    int i = 0, j = 0;
    bool intersected = false;
    while (true)
    {
      if (sources[i].first.len() == 0)
      {
        if (++i == sources.size())
          break;
      }
      else if (mask[j].len() == 0)
      {
        if (++j == mask.size())
          break;
      }
      else if (sources[i].first.right() <= mask[j].left())
      {
        if (++i == sources.size())
          break;
      }
      else if (mask[j].right() <= sources[i].first.left())
      {
        if (++j == mask.size())
          break;
      }
      else
      {
        Segment intersection;
        if (Intersect(sources[i].first, mask[j], intersection))
        {
          result.push_back(pair<Segment, int>(intersection, sources[i].second));
          intersected = true;
        }
        if (sources[i].first.right() <= mask[j].right())
        {
          if (++i == sources.size())
            break;
        }
        else
        {
          if (++j == mask.size())
            break;
        }
      }
    }
    
    return intersected;
  }
  

  void Merge(const vector< pair<Segment, int> > &x,
             const vector< pair<Segment, int> > &y,
             vector< pair<Segment, int> > &result)
  {
    result.reserve(result.size() + x.size() + y.size());
    int i = 0, j = 0;
    if (!x.empty() && !y.empty())
    {
      while (true)
      {
        if (x[i].first.len() == 0)
        {
          if (++i == x.size())
            break;
        }
        else if (y[j].first.len() == 0)
        {
          if (++j == y.size())
            break;
        }
        else if (x[i].first.left() <= y[j].first.left())
        {
          result.push_back(x[i]);
          if (++i == x.size())
            break;
        }
        else
        {
          result.push_back(y[j]);
          if (++j == y.size())
            break;
        }
      }
    }
    while (i < x.size())
    {
      if (x[i].first.len() > 0)
        result.push_back(x[i]);
      ++i;
    }
    while (j < y.size())
    {
      if (y[j].first.len() > 0)
        result.push_back(y[j]);
      ++j;
    }
  }

  
  struct SegmentIntTouchLess
  {
    bool operator()(const pair<Segment, int> &x, const pair<Segment, int> &y) const
    {
      return x.first.right() <= y.first.left();
    }
  };
  

  void Haplo::Set(const SNPData &data, int k)
  {
    sources.clear();
    sources.push_back(pair<Segment, int>(Segment(0, data.seq_len), k));
  }

  void Haplo::Sample(const SNPData &data)
  {
    Set(data, irand() % data.num_haplos);
  }

  void Haplo::Sample(const SNPData &data, const vector<int> &select)
  {
    Assert(select.size() == data.num_haplos);
    int count = 0;
    for (int k = 0; k < data.num_haplos; ++k)
      if (select[k])
        ++count;
    Assert(count > 0);
    
    int index = irand() % count;
    for (int k = 0; k < data.num_haplos; ++k)
      if (select[k])
        if (--index < 0)
        {
          Set(data, k);
          break;
        }
  }

  void Haplo::SamplePop(const SNPData &data, int pop)
  {
    Assert((pop >= 0) && (pop <data.num_pops));
    vector<int> select(data.num_pops, 0);
    select[pop] = 1;
    SamplePops(data, select);
  }

  void Haplo::SamplePops(const SNPData &data, const vector<int> &select)
  {
    Assert(select.size() == data.num_pops);
    vector<int> select_haplo(data.num_haplos, 0);
    for (int k = 0; k < data.num_haplos; ++k)
      if (select[data.haplos[k].pop])
        select_haplo[k] = 1;
    Sample(data, select_haplo);
  }

  void Haplo::SamplePops(const SNPData &data, uint64 pop_mask)
  {
    vector<int> select(data.num_pops, 0);
    for (int pop = 0; pop < data.num_pops; ++pop)
      if (pop_mask & (uint64(1) << pop))
        select[pop] = 1;
    SamplePops(data, select);
  }

  void Haplo::SetFullTree(const SNPProfile &profile, int num_gens)
  {
    Assert(num_gens >= 1);
    vector<Segment> full_range(1, Segment(0, profile.seq_len));
    sources.clear();
    SetFullTreeRecurse(profile, num_gens, full_range, sources);
  }

  void Haplo::SetFullTreeRecurse(const SNPProfile &profile, int gens,
                                 const vector<Segment> &restrict,
                                 vector< pair<Segment, int> > &sources,
                                 int index)
  {
    Assert(!restrict.empty());

    if (gens == 0)
    {
      Assign(restrict, index, sources);
      return;
    }

    vector<Segment> r0, r1, restrict0, restrict1;
    vector< pair<Segment, int> > sources0, sources1;

    GenerateRecombs(profile, r0, r1);

    if (Intersect(restrict, r0, restrict0))
      SetFullTreeRecurse(profile, gens - 1, restrict0, sources0, index*2 + 0);
    if (Intersect(restrict, r1, restrict1))
      SetFullTreeRecurse(profile, gens - 1, restrict1, sources1, index*2 + 1);
    
    Merge(sources0, sources1, sources);
  }

  void Haplo::Composite(const HaploPop &pop, const Haplo &haplo)
  {
    sources.clear();
    for (int i = 0; i < haplo.sources.size(); ++i)
    {
      int k = haplo.sources[i].second;
      Assert((k >= 0) && (k < pop.haplos.size()));
      
      vector< pair<Segment, int> >::const_iterator begin = 
        lower_bound(pop.haplos[k].sources.begin(), pop.haplos[k].sources.end(),
                    haplo.sources[i], SegmentIntTouchLess());
      vector< pair<Segment, int> >::const_iterator end =
        upper_bound(pop.haplos[k].sources.begin(), pop.haplos[k].sources.end(),
                    haplo.sources[i], SegmentIntTouchLess());
      Assert(begin != end);

      Segment intersection;
      Assert(Intersect(begin->first, haplo.sources[i].first, intersection));
      sources.push_back(pair<Segment, int>(intersection, begin->second));

      if (++begin == end)
        continue;

      for (--end; begin != end; ++begin)
        sources.push_back(*begin);

      Assert(Intersect(begin->first, haplo.sources[i].first, intersection));
      sources.push_back(pair<Segment, int>(intersection, begin->second));
    }
  }

  void Haplo::Composite(const vector<HaploPop> &pops, const Haplo &haplo)
  {
    sources.clear();
    for (int i = 0; i < haplo.sources.size(); ++i)
    {
      int p = (haplo.sources[i].second >> 16);
      int k = (haplo.sources[i].second & 0xffff);
      Assert((k >= 0) && (k < pops[p].haplos.size()));
      
      vector< pair<Segment, int> >::const_iterator begin = 
        lower_bound(pops[p].haplos[k].sources.begin(), 
                    pops[p].haplos[k].sources.end(),
                    haplo.sources[i], SegmentIntTouchLess());
      vector< pair<Segment, int> >::const_iterator end =
        upper_bound(pops[p].haplos[k].sources.begin(), 
                    pops[p].haplos[k].sources.end(),
                    haplo.sources[i], SegmentIntTouchLess());
      Assert(begin != end);

      Segment intersection;
      Assert(Intersect(begin->first, haplo.sources[i].first, intersection));
      sources.push_back(pair<Segment, int>(intersection, begin->second));

      if (++begin == end)
        continue;

      for (--end; begin != end; ++begin)
        sources.push_back(pair<Segment, int>(*begin));

      Assert(Intersect(begin->first, haplo.sources[i].first, intersection));
      sources.push_back(pair<Segment, int>(intersection, begin->second));
    }
  }

  void Haplo::MapSources(const vector<int> &mapping)
  {
    for (int i = 0; i < sources.size(); ++i)
    {
      int k = sources[i].second;
      Assert((k >= 0) && (k < mapping.size()));
      sources[i].second = mapping[k];
    }
  }

  void Haplo::MapSources(const map<int, int> &mapping)
  {
    for (int i = 0; i < sources.size(); ++i)
    {
      int k = sources[i].second;
      map<int, int>::const_iterator find = mapping.find(k);
      Assert(find != mapping.end());
      sources[i].second = find->second;
    }
  }

  void Haplo::WriteBin(ostream &o) const
  {
    ::WriteBin(o, sources);
  }

  void Haplo::ReadBin(istream &i)
  {
    ::ReadBin(i, sources);
  }

  void Haplo::DebugDisplay(ostream &o) const
  {
    o << "Sources (" << sources.size() << "):" << endl;
    for (int i = 0; i < sources.size(); ++i)
    {
      o << "  [" << sources[i].first.left() << ", " << sources[i].first.right() << ")" << ((((i + 1) < sources.size()) && (sources[i].first.right() != sources[i + 1].first.left())) ? "*" : "") << " => " << sources[i].second << endl;
    }
  }

  
  void Diplo::Set(const SNPData &data, int d)
  {
    Assert((d >= 0) && (d < data.num_diplos));
    haplos[0].Set(data, data.diplos[d].haplos[0]);
    haplos[1].Set(data, data.diplos[d].haplos[1]);
    haplo_sry[0] = data.haplos[data.diplos[d].haplos[0]].sry;
    haplo_sry[1] = data.haplos[data.diplos[d].haplos[1]].sry;
    Assert((haplo_sry[0] == 0) || (haplo_sry[1] == 0));
    SetGender();
  }

  void Diplo::SetGender()
  {
    Assert((haplo_sry[0] == 0) || (haplo_sry[1] == 0));
    gender = ((haplo_sry[0] == 0) && (haplo_sry[1] == 0)) ? 0 : 1;
  }

  void Diplo::Sample(const SNPData &data)
  {
    Set(data, irand() % data.num_diplos);
  }

  void Diplo::Sample(const SNPData &data, const vector<int> &select)
  {
    Assert(select.size() == data.num_diplos);
    int count = 0;
    for (int d = 0; d < data.num_diplos; ++d)
      if (select[d])
        ++count;
    Assert(count > 0);
    
    int index = irand() % count;
    for (int d = 0; d < data.num_diplos; ++d)
      if (select[d])
        if (--index < 0)
        {
          Set(data, d);
          break;
        }
  }

  void Diplo::SamplePop(const SNPData &data, int pop)
  {
    Assert((pop >= 0) && (pop <data.num_pops));
    vector<int> select(data.num_pops, 0);
    select[pop] = 1;
    SamplePops(data, select);
  }

  void Diplo::SamplePops(const SNPData &data, const vector<int> &select)
  {
    Assert(select.size() == data.num_pops);
    vector<int> select_diplo(data.num_diplos, 0);
    for (int d = 0; d < data.num_diplos; ++d)
      if (select[data.diplos[d].pop])
        select_diplo[d] = 1;
    Sample(data, select_diplo);
  }

  void Diplo::SamplePops(const SNPData &data, uint64 pop_mask)
  {
    vector<int> select(data.num_pops, 0);
    for (int pop = 0; pop < data.num_pops; ++pop)
      if (pop_mask & (uint64(1) << pop))
        select[pop] = 1;
    SamplePops(data, select);
  }

  void Diplo::clear()
  {
    haplos[0].clear();
    haplos[1].clear();
    haplo_sry[0] = haplo_sry[1] = 0;
    gender = -1;
  }

  void Diplo::Free()
  {
    haplos[0].Free();
    haplos[1].Free();
    haplo_sry[0] = haplo_sry[1] = 0;
    gender = -1;
  }

  void Diplo::swap(Diplo &other)
  { 
    haplos[0].swap(other.haplos[0]); 
    haplos[1].swap(other.haplos[1]);
    ::swap(haplo_sry[0], other.haplo_sry[0]);
    ::swap(haplo_sry[1], other.haplo_sry[1]);
    SetGender();
    other.SetGender();
  }

  void Diplo::SetFullTree(const SNPProfile &profile, int num_gens, int gender)
  {
    Assert(num_gens >= 1);
    vector<Segment> full_range(1, Segment(0, profile.seq_len));
    haplos[0].clear();
    haplos[1].clear();
    haplo_sry[0] = 0;
    haplo_sry[1] = rand() & 1;
    SetFullTreeRecurse(profile, num_gens - 1,  full_range, 
                       haplos[0].sources, 0, 0);
    SetFullTreeRecurse(profile, num_gens - 1, full_range,
                       haplos[1].sources, 1, 1);
    SetGender();
  }


  void Diplo::SetFullTreeRecurse(const SNPProfile &profile, int gens,
                                 const vector<Segment> &restrict, 
                                 vector< pair<Segment, int> > &sources,
                                 int gender, int index)
  {
    Assert(!restrict.empty());

    vector<Segment> r0, r1, restrict0, restrict1;
    vector< pair<Segment, int> > sources0, sources1;

    int which_sry = GenerateRecombs(profile, r0, r1, gender);
    if (which_sry < 0)
    {
      if (rand() & 1)
        r0.swap(r1);
    }
    else if (which_sry == 0)
      r0.swap(r1);

    if (Intersect(restrict, r0, restrict0))
    {
      if (gens == 0)
        Assign(restrict0, index*2 + 0, sources0);
      else
        SetFullTreeRecurse(profile, gens - 1, restrict0, sources0, 
                           0, index*2 + 0);
    }

    if (Intersect(restrict, r1, restrict1))
    {
      if (gens == 0)
        Assign(restrict1, index*2 + 1, sources1);
      else
        SetFullTreeRecurse(profile, gens - 1, restrict1,
                           sources1, 1, index*2 + 1);
    }
    
    Merge(sources0, sources1, sources);
  }

  void Diplo::Composite(const DiploPop &pop, const Diplo &diplo)
  {
    for (int a = 0; a < 2; ++a)
    {
      haplos[a].sources.clear();
      for (int i = 0; i < diplo.haplos[a].sources.size(); ++i)
      {
        int k = diplo.haplos[a].sources[i].second;
        int d = k >> 1;
        int b = k & 1;
        Assert((d >= 0) && (d < pop.diplos.size()));
        
        vector< pair<Segment, int> >::const_iterator begin = 
          lower_bound(pop.diplos[d].haplos[b].sources.begin(), 
                      pop.diplos[d].haplos[b].sources.end(), 
                      diplo.haplos[a].sources[i], SegmentIntTouchLess());
        vector< pair<Segment, int> >::const_iterator end =
          upper_bound(pop.diplos[d].haplos[b].sources.begin(), 
                      pop.diplos[d].haplos[b].sources.end(),
                      diplo.haplos[a].sources[i], SegmentIntTouchLess());
        Assert(begin != end);

        Segment intersection;
        Assert(Intersect(begin->first, diplo.haplos[a].sources[i].first, intersection));
        haplos[a].sources.push_back(pair<Segment, int>(intersection, begin->second));

        if (++begin == end)
          continue;
        
        for (--end; begin != end; ++begin)
          haplos[a].sources.push_back(*begin);

        Assert(Intersect(begin->first, diplo.haplos[a].sources[i].first, intersection));
        haplos[a].sources.push_back(pair<Segment, int>(intersection, begin->second));
      }

      haplo_sry[a] = diplo.haplo_sry[a];
    }
    SetGender();
  }

  void Diplo::Composite(const vector<DiploPop> &pops, const Diplo &diplo)
  {
    for (int a = 0; a < 2; ++a)
    {
      haplos[a].sources.clear();
      for (int i = 0; i < diplo.haplos[a].sources.size(); ++i)
      {
        if (diplo.haplos[a].sources[i].first.len() == 0)
          continue;
        int p = (diplo.haplos[a].sources[i].second >> 16);
        int k = (diplo.haplos[a].sources[i].second & 0xffff);
        int d = k >> 1;
        int b = k & 1;
        Assert((p >= 0) && (p < pops.size()));
        Assert((d >= 0) && (d < pops[p].diplos.size()));
        
        vector< pair<Segment, int> >::const_iterator begin = 
          lower_bound(pops[p].diplos[d].haplos[b].sources.begin(), 
                      pops[p].diplos[d].haplos[b].sources.end(), 
                      diplo.haplos[a].sources[i], SegmentIntTouchLess());
        vector< pair<Segment, int> >::const_iterator end =
          upper_bound(pops[p].diplos[d].haplos[b].sources.begin(), 
                      pops[p].diplos[d].haplos[b].sources.end(),
                      diplo.haplos[a].sources[i], SegmentIntTouchLess());
        if (begin == end)
        {
          cerr << "Failure:" << endl;
          cerr << "[" << diplo.haplos[a].sources[i].first.left() << ", " << diplo.haplos[a].sources[i].first.right() << ") => " << p << ", " << d << ", " << b << endl;
          cerr << "num_pops = " << pops.size() << endl;
          cerr << "num_diplos = " << pops[p].diplos.size() << endl;
          cerr << "Haplotype 0:" << endl;
          pops[p].diplos[d].haplos[0].DebugDisplay(cerr);
          cerr << "Haplotype 1:" << endl;
          pops[p].diplos[d].haplos[1].DebugDisplay(cerr);
          cerr << "-1 Haplotype 0:" << endl;
          pops[p].diplos[d-1].haplos[0].DebugDisplay(cerr);
          cerr << "-1 Haplotype 1:" << endl;
          pops[p].diplos[d-1].haplos[1].DebugDisplay(cerr);
          cerr << "+1 Haplotype 0:" << endl;
          pops[p].diplos[d+1].haplos[0].DebugDisplay(cerr);
          cerr << "+1 Haplotype 1:" << endl;
          pops[p].diplos[d+1].haplos[1].DebugDisplay(cerr);
        }
        Assert(begin != end);

        Segment intersection;
        Assert(Intersect(begin->first, diplo.haplos[a].sources[i].first, intersection));
        haplos[a].sources.push_back(pair<Segment, int>(intersection, begin->second));

        if (++begin == end)
          continue;
        
        for (--end; begin != end; ++begin)
          haplos[a].sources.push_back(*begin);

        Assert(Intersect(begin->first, diplo.haplos[a].sources[i].first, intersection));
        haplos[a].sources.push_back(pair<Segment, int>(intersection, begin->second));
      }

      haplo_sry[a] = diplo.haplo_sry[a];
    }
    SetGender();
  }

  void Diplo::MapSources(const vector<int> &mapping)
  {
    haplos[0].MapSources(mapping);
    haplos[1].MapSources(mapping);
  }

  void Diplo::MapSources(const map<int, int> &mapping)
  {
    haplos[0].MapSources(mapping);
    haplos[1].MapSources(mapping);
  }

  void Diplo::WriteBin(ostream &o) const
  {
    haplos[0].WriteBin(o);
    haplos[1].WriteBin(o);
    ::WriteBin(o, haplo_sry[0]);
    ::WriteBin(o, haplo_sry[1]);
    ::WriteBin(o, gender);
  }

  void Diplo::ReadBin(istream &i)
  {
    haplos[0].ReadBin(i);
    haplos[1].ReadBin(i);
    ::ReadBin(i, haplo_sry[0]);
    ::ReadBin(i, haplo_sry[1]);
    ::ReadBin(i, gender);
  }
    

  void GenerateRecombs(const SNPProfile &profile, vector<int> &recombs, 
                       int gender)
  {
    const vector<double> &pos = (gender < 0) ? profile.locus_genetic_pos : profile.gender_locus_genetic_pos[gender];

    vector<int> spots;
    for (int chrom = 0; chrom < profile.num_chroms; ++chrom)
    {
      if (rand() & 1)
        spots.push_back(profile.chroms[chrom].range.left());

      if (profile.chroms[chrom].range.len() == 0)
        continue;
      
      double genetic_size = pos[profile.chroms[chrom].range.right() - 1] - pos[profile.chroms[chrom].range.left()];
      if (genetic_size == 0.0)
        continue;
      
      int num_recombs = int(floor(grand() * sqrt(genetic_size) + genetic_size + 0.5));
      if (num_recombs < 1)
        num_recombs = 1;
    
      vector<double>::const_iterator begin = pos.begin() + profile.chroms[chrom].range.left();
      vector<double>::const_iterator end = pos.begin() + profile.chroms[chrom].range.right();
      for (int recomb = 0; recomb < num_recombs; ++recomb)
      {
        double genetic_pos = drand() * genetic_size + *begin;
        vector<double>::const_iterator find = lower_bound(begin, end, genetic_pos);
        Assert(find != end);
        spots.push_back(find - pos.begin());
      }
    }
  
    sort(spots.begin(), spots.end());
    recombs.clear();
    recombs.reserve(spots.size());
    for (int i = 0; i < spots.size(); )
    {
      if (((i + 1) < spots.size()) && (spots[i] == spots[i + 1]))
        i += 2;
      else
        recombs.push_back(spots[i++]);
    }
  }

  
  int GenerateRecombs(const SNPProfile &profile, 
                      vector<Segment> &x, vector<Segment> &y, int gender)
  {
    vector<int> recombs;
    GenerateRecombs(profile, recombs, gender);
    x.reserve(recombs.size() / 2 + 1);
    y.reserve((recombs.size() + 1) / 2);

    vector<int> start_stop(recombs.size() + 2);
    start_stop[0] = 0;
    for (int i = 0; i < recombs.size(); ++i)
      start_stop[i + 1] = recombs[i];
    start_stop.back() = profile.seq_len;

    int which_sry = -1;
    for (int i = 1; i < start_stop.size(); ++i)
    {
      if ((start_stop[i - 1] <= profile.sry_pos) && (profile.sry_pos < start_stop[i]))
        which_sry = (i & 1) ? 0 : 1;

      if (start_stop[i - 1] != start_stop[i])
      {
        if (i & 1)
          x.push_back(Segment(start_stop[i - 1], start_stop[i], true));
        else
          y.push_back(Segment(start_stop[i - 1], start_stop[i], true));
      }
    }

    if ((profile.loci[0].genetic_pos != 0.0) && (rand() & 1))
    {
      x.swap(y);
      if (which_sry >= 0)
        which_sry = 1 - which_sry;
    }

    return which_sry;
  }


  void RecombineHaplos(const SNPProfile &profile, 
                       const Haplo &x, const Haplo &y, Haplo &result)
  {
    vector<Segment> x_mask, y_mask;
    GenerateRecombs(profile, x_mask, y_mask);

    vector< pair<Segment, int> > x_sources, y_sources;
    Intersect(x.sources, x_mask, x_sources);
    Intersect(y.sources, y_mask, y_sources);

    result.clear();
    Merge(x_sources, y_sources, result.sources);
  }


  void MateDiplos(const SNPProfile &profile, const Diplo &x, const Diplo &y, 
                  Diplo &result, int gender)
  {
    if ((x.gender == 0) && (y.gender == 0))
      Assert(gender != 1);

    while (true)
    {
      for (int s = 0; s < 2; ++s)
      {
        const Diplo &parent = (s == 0) ? x : y;
        Assert((parent.haplo_sry[0] == 0) || (parent.haplo_sry[1] == 0));
        
        vector< vector<Segment> > mask(2);
        int which_sry = GenerateRecombs(profile, mask[0], mask[1], parent.gender);
        
        vector< vector< pair<Segment, int> > > sources(2);
        Intersect(parent[0].sources, mask[0], sources[0]);
        Intersect(parent[1].sources, mask[1], sources[1]);

        result[s].clear();
        Merge(sources[0], sources[1], result[s].sources);
        if (which_sry >= 0)
          result.haplo_sry[s] = parent.haplo_sry[which_sry];
        else
          result.haplo_sry[s] = parent.haplo_sry[rand() & 1];
      }

      if ((result.haplo_sry[0] != 0) && (result.haplo_sry[1] != 0))
        continue;
      result.SetGender();
      if (gender < 0)
        break;
      else if (gender == result.gender)
        break;
    }
  }

  
  void HaploPop::swap(HaploPop &other)
  {
    ::swap(data, other.data);
    haplos.swap(other.haplos);
  }

  void HaploPop::SetSNPData(const SNPData &data_)
  {
    data = &data_;
    haplos.clear();
  }

  void HaploPop::SetAll()
  {
    Assert(data);
    vector<int> select(data->num_haplos, 1);
    Set(select);
  }

  void HaploPop::SetPop(int pop)
  {
    Assert(data);
    Assert((pop >= 0) && (pop < data->num_pops));
    vector<int> select(data->num_pops, 0);
    select[pop] = 1;
    SetPops(select);
  }

  void HaploPop::SetPops(const vector<int> &select)
  {
    Assert(data);
    Assert(select.size() == data->num_pops);
    vector<int> select_haplo(data->num_haplos, 0);
    for (int k = 0; k < data->num_haplos; ++k)
      if (select[data->haplos[k].pop])
        select_haplo[k] = 1;
    Set(select_haplo);
  }
  
  void HaploPop::SetPops(uint64 pop_mask)
  {
    Assert(data);
    vector<int> select(data->num_pops, 0);
    for (int pop = 0; pop < data->num_pops; ++pop)
      if (pop_mask & (uint64(1) << pop))
        select[pop] = 1;
    SetPops(select);
  }

  void HaploPop::Set(const vector<int> &select)
  {
    Assert(data);
    Assert(select.size() == data->num_haplos);
    int count = 0;
    for (int k = 0; k < data->num_haplos; ++k)
      if (select[k])
        ++count;

    haplos.clear();
    haplos.resize(count, Haplo(*data, -1));
    for (int k = 0, i = 0; k < data->num_haplos; ++k)
      if (select[k])
        haplos[i++][0].second = k;
  }


  void HaploPop::Derive(Haplo &result) const
  {
    int x = irand() % haplos.size();
    int y = irand() % haplos.size();
    RecombineHaplos(*data->profile, haplos[x], haplos[y], result);
  }
    
  void HaploPop::DerivePop(const HaploPop &sources, int num_samples)
  {
    Assert(data);
    SetSNPData(*sources.data);
    haplos.clear();
    haplos.resize(num_samples);
    for (int haplo = 0; haplo < num_samples; ++haplo)
      sources.Derive(haplos[haplo]);
  }

  void HaploPop::MapSources(const vector<int> &mapping)
  {
    for (int k = 0; k < haplos.size(); ++k)
      haplos[k].MapSources(mapping);
  }

  void HaploPop::MapSources(const map<int, int> &mapping)
  {
    for (int k = 0; k < haplos.size(); ++k)
      haplos[k].MapSources(mapping);
  }

  int HaploPop::CompactSources()
  {
    GraphColoring coloring;

    map<int, int> mapping;
    coloring.Set(haplos, mapping);
    int num_sources = coloring.GreedyColor();

    for (map<int, int>::iterator iter = mapping.begin();
         iter != mapping.end(); ++iter)
      iter->second = coloring.vertices[iter->second].color;
    MapSources(mapping);

    return num_sources;
  }

  void HaploPop::CreateSNPData(SNPData &result)
  {
    Assert(data);
    Assert(data->profile);

    result.SetProfile(*data->profile);
    result.Clear();

    result.pops.push_back(SNPData::Pop(""));

    vector<int8> haplo;
    for (int k = 0; k < haplos.size(); ++k)
    {
      haplos[k].Generate(*data, haplo);
      result.AddHaplo(haplo, 0);
    }

    result.ComputeDerived();
  }

  void HaploPop::WriteBin(ostream &o) const
  {
    Assert(data);
    Assert(data->profile);

    uint32 profile_hash = data->profile->ComputeHash();
    ::WriteBin(o, profile_hash);

    int32 size = haplos.size();
    ::WriteBin(o, size);
    for (int k = 0; k < size; ++k)
      haplos[k].WriteBin(o);
  }

  void HaploPop::WriteBin(const string &filename) const
  {
    ostream *out = OutFileStream(filename);
    AssertMsg(out, filename);
    WriteBin(*out);
    delete out;
  }

  void HaploPop::ReadBin(istream &i)
  {
    Assert(data);
    Assert(data->profile);

    uint32 profile_hash;
    ::ReadBin(i, profile_hash);
    Assert(profile_hash == data->profile->ComputeHash());

    int32 size;
    ::ReadBin(i, size);
    haplos.resize(size);
    for (int k = 0; k < size; ++k)
      haplos[k].ReadBin(i);
  }

  void HaploPop::ReadBin(const string &filename)
  {
    istream *in = InFileStream(filename);
    AssertMsg(in, filename);
    ReadBin(*in);
    delete in;
  }

  void DiploPop::swap(DiploPop &other)
  {
    ::swap(data, other.data);
    diplos.swap(other.diplos);
  }

  void DiploPop::SetSNPData(const SNPData &data_)
  {
    data = &data_;
    diplos.clear();
  }

  void DiploPop::SetAll()
  {
    Assert(data);
    vector<int> select(data->num_diplos, 1);
    Set(select);
  }

  void DiploPop::SetPop(int pop)
  {
    Assert(data);
    Assert((pop >= 0) && (pop < data->num_pops));
    vector<int> select(data->num_pops, 0);
    select[pop] = 1;
    SetPops(select);
  }

  void DiploPop::SetPops(const vector<int> &select)
  {
    Assert(data);
    Assert(select.size() == data->num_pops);
    vector<int> select_diplo(data->num_diplos, 0);
    for (int d = 0; d < data->num_diplos; ++d)
      if (select[data->diplos[d].pop])
        select_diplo[d] = 1;
    Set(select_diplo);
  }

  void DiploPop::SetPops(uint64 pop_mask)
  {
    Assert(data);
    vector<int> select(data->num_pops, 0);
    for (int pop = 0; pop < data->num_pops; ++pop)
      if (pop_mask & (uint64(1) << pop))
        select[pop] = 1;
    SetPops(select);
  }

  void DiploPop::Set(const vector<int> &select)
  {
    Assert(data);
    Assert(select.size() == data->num_diplos);
    int count = 0;
    for (int d = 0; d < data->num_diplos; ++d)
      if (select[d])
        ++count;

    diplos.clear();
    diplos.reserve(count);
    for (int d = 0; d < data->num_diplos; ++d)
      if (select[d])
        diplos.push_back(Diplo(*data, d));
  }

  void DiploPop::Derive(Diplo &result) const
  {
    int female, male;
    do
    {
      female = irand() % diplos.size();
    } while (diplos[female].gender != 0);
    do
    {
      male = irand() % diplos.size();
    } while (diplos[male].gender != 1);
    
    MateDiplos(*data->profile, diplos[female], diplos[male], result);
  }

  void DiploPop::DerivePop(const DiploPop &sources, int num_samples)
  {
    Assert(sources.data);
    SetSNPData(*sources.data);
    diplos.clear();
    diplos.resize(num_samples);
    for (int diplo = 0; diplo < num_samples; ++diplo)
      sources.Derive(diplos[diplo]);
  }

  void DiploPop::MapSources(const vector<int> &mapping)
  {
    for (int d = 0; d < diplos.size(); ++d)
      diplos[d].MapSources(mapping);
  }

  void DiploPop::MapSources(const map<int, int> &mapping)
  {
    for (int d = 0; d < diplos.size(); ++d)
      diplos[d].MapSources(mapping);
  }

  int DiploPop::CompactSources()
  {
    GraphColoring coloring;

    vector<Haplo> haplos(diplos.size() * 2);
    for (int d = 0; d < diplos.size(); ++d)
    {
      haplos[d*2 + 0] = diplos[d].haplos[0];
      haplos[d*2 + 1] = diplos[d].haplos[1];
    }

    map<int, int> mapping;
    coloring.Set(haplos, mapping);
    int num_sources = coloring.GreedyColor();

    for (map<int, int>::iterator iter = mapping.begin();
         iter != mapping.end(); ++iter)
      iter->second = coloring.vertices[iter->second].color;
    MapSources(mapping);

    return num_sources;
  }

  void DiploPop::CreateSNPData(SNPData &result)
  {
    Assert(data);
    Assert(data->profile);
    result.SetProfile(*data->profile);

    result.Clear();
    result.pops.push_back(SNPData::Pop(""));

    vector<int8> haplo0, haplo1;
    for (int d = 0; d < diplos.size(); ++d)
    {
      diplos[d].Generate(*data, haplo0, haplo1);
      result.AddDiplo(haplo0, haplo1, 0.0, 0, "", 
                      diplos[d].haplo_sry[0], diplos[d].haplo_sry[1]);
    }

    result.ComputeDerived();
  }

  void DiploPop::CreateHaploPop(HaploPop &result)
  {
    result.SetSNPData(*data);
    result.clear();
    for (int d = 0; d < diplos.size(); ++d)
      for (int a = 0; a < 2; ++a)
        result.haplos.push_back(diplos[d].haplos[a]);
  }

  void DiploPop::WriteBin(ostream &o) const
  {
    Assert(data);
    Assert(data->profile);

    uint32 profile_hash = data->profile->ComputeHash();
    ::WriteBin(o, profile_hash);

    int32 size = diplos.size();
    ::WriteBin(o, size);
    for (int k = 0; k < size; ++k)
      diplos[k].WriteBin(o);
  }

  void DiploPop::WriteBin(const string &filename) const
  {
    ostream *out = OutFileStream(filename);
    AssertMsg(out, filename);
    WriteBin(*out);
    delete out;
  }

  void DiploPop::ReadBin(istream &i)
  {
    Assert(data);
    Assert(data->profile);

    uint32 profile_hash;
    ::ReadBin(i, profile_hash);
    Assert(profile_hash == data->profile->ComputeHash());

    int32 size;
    ::ReadBin(i, size);
    diplos.resize(size);
    for (int k = 0; k < size; ++k)
      diplos[k].ReadBin(i);
  }

  void DiploPop::ReadBin(const string &filename)
  {
    istream *in = InFileStream(filename);
    AssertMsg(in, filename);
    ReadBin(*in);
    delete in;
  }


  void GraphColoring::Set(const vector<Haplo> &haplos,
                          map<int, int> &source_to_vertex)
  {
    source_to_vertex.clear();
    for (int k = 0; k < haplos.size(); ++k)
      for (int i = 0; i < haplos[k].size(); ++i)
        source_to_vertex[haplos[k][i].second] = -1;

    int num_vertices = 0;
    for (map<int, int>::iterator iter = source_to_vertex.begin();
         iter != source_to_vertex.end(); ++iter)
      iter->second = num_vertices++;
    
    vertices.clear();
    vertices.resize(num_vertices);

    vector< vector<Segment> > segments(haplos.size());
    for (int k = 0; k < haplos.size(); ++k)
    {
      segments[k].reserve(haplos[k].size());
      for (int i = 0; i < haplos[k].size(); ++i)
        segments[k].push_back(haplos[k][i].first);
    }

    vector< pair<int, int> > overlaps;
    for (int k = 0; k < haplos.size(); ++k)
    {
      for (int i = 1; i < haplos[k].size(); ++i)
      {
        int vertex1 = source_to_vertex[haplos[k][i - 1].second];
        int vertex2 = source_to_vertex[haplos[k][i].second];
        if (vertex1 != vertex2)
          vertices[min(vertex1, vertex2)].adj.push_back(max(vertex1, vertex2));
      }

      for (int l = k + 1; l < haplos.size(); ++l)
      {
        overlaps.clear();
        GetTSegmentPairOverlaps<int>(segments[k], segments[l], overlaps);
        
        for (int i = 0; i < overlaps.size(); ++i)
        {
          int vertex1 = source_to_vertex[haplos[k][overlaps[i].first].second];
          int vertex2 = source_to_vertex[haplos[l][overlaps[i].second].second];
          if (vertex1 != vertex2)
            vertices[min(vertex1, vertex2)].adj.push_back(max(vertex1, vertex2));
        }
      }
    }

    for (int i = 0; i < vertices.size(); ++i)
      for (int j = 0; j < vertices[i].adj.size(); ++j)
        vertices[vertices[i].adj[j]].adj.push_back(i);

    for (int i = 0; i < vertices.size(); ++i)
    {
      sort(vertices[i].adj.begin(), vertices[i].adj.end());
      vertices[i].adj.resize(unique(vertices[i].adj.begin(), vertices[i].adj.end()) - vertices[i].adj.begin());
    }
  }


  int GraphColoring::GreedyColor()
  {
    for (int i = 0; i < vertices.size(); ++i)
      vertices[i].color = -1;

    vector<int> order;
    SortOrder(vertices.begin(), vertices.end(), order, VertexDegreeDecreasing());

    int remaining = vertices.size();
    int color;
    for (color = 0; remaining; ++color)
    {
      for (int o = 0; o < order.size(); ++o)
      {
        int i = order[o];
        if (vertices[i].color < 0)
        {
          bool okay = true;
          for (int j = 0; j < vertices[i].adj.size(); ++j)
            if (vertices[vertices[i].adj[j]].color == color)
            {
              okay = false;
              break;
            }

          if (okay)
          {
            vertices[i].color = color;
            --remaining;
          }
        }
      }
    }

    return color;
  }
};


template <>
void Write<Simulate::Haplo>(ostream &o, const Simulate::Haplo &x)
{
  x.WriteBin(o);
}

template <>
void Read<Simulate::Haplo>(istream &i, Simulate::Haplo &x)
{
  x.ReadBin(i);
}

template <>
void Write<Simulate::Diplo>(ostream &o, const Simulate::Diplo &x)
{
  x.WriteBin(o);
}

template <>
void Read<Simulate::Diplo>(istream &i, Simulate::Diplo &x)
{
  x.ReadBin(i);
}

template <>
void Write<Simulate::HaploPop>(ostream &o, const Simulate::HaploPop &x)
{
  x.WriteBin(o);
}

template <>
void Read<Simulate::HaploPop>(istream &i, Simulate::HaploPop &x)
{
  x.ReadBin(i);
}

template <>
void Write<Simulate::DiploPop>(ostream &o, const Simulate::DiploPop &x)
{
  x.WriteBin(o);
}

template <>
void Read<Simulate::DiploPop>(istream &i, Simulate::DiploPop &x)
{
  x.ReadBin(i);
}
