
#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"
#include "Simulate.h"


Params params(false);

int debug = 0;

string input_dir;
string input_file;
string snpprofiledata_file;
string output_dir;
string output_file;
bool output_profile;
string label = "simulated";
vector<string> composite_files;
double switch_error = 0.0;

bool do_diplo;
vector<int> selected;

string required_param[] =
{
  "input_file",
  "snpprofiledata_file",
  "output_file"
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

void GetIntegerSpec(string s, vector<int> &result);

void InitParams()
{
  params.ExpandImports();

  uint32 seed = RandomSeed();
  if (params.Contains("rand_seed"))
    seed = params["rand_seed"].GetInt();
  srand(seed);
  cerr << "- Random seed = " << seed << endl;

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("input_file"))
  {
    input_file = params["input_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(input_file)) == "bin");
    if (GetExtension(RemoveExtension(RemoveCompressedExtension(input_file))) == "haplopop")
      do_diplo = false;
    else if (GetExtension(RemoveExtension(RemoveCompressedExtension(input_file))) == "diplopop")
      do_diplo = true;
    else
      AssertMsg(false, "Neither haplopop nor diplopop!");
  }

  if (params.Contains("snpprofiledata_file"))
  {
    snpprofiledata_file = params["snpprofiledata_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(snpprofiledata_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(snpprofiledata_file))) == "snpprofiledata");
  }

  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (params.Contains("output_file"))
  {
    output_file = params["output_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(output_file)) == "bin");
    string file_type = GetExtension(RemoveExtension(RemoveCompressedExtension(output_file)));
    if (file_type == "snpdata")
      output_profile = false;
    else if (file_type == "snpprofiledata")
      output_profile = true;
    else
      AssertMsg(false, "Neither snpdata nor snpprofiledata!");
  }

  if (params.Contains("select"))
    GetIntegerSpec(params["select"].GetString(), selected);

  if (params.Contains("label"))
    label = params["label"].GetString();

  if (params.Contains("composite_files"))
  {
    Assert(!input_file.empty());
    vector<ParamValue> values = params["composite_files"].GetList();
    composite_files.resize(values.size());
    bool has_haplo = false;
    bool has_diplo = false;
    for (int i = 0; i < values.size(); ++i)
    {
      string filename = values[i].GetString();
      Assert(GetExtension(RemoveCompressedExtension(filename)) == "bin");
      if (GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "diplopop")
        has_diplo = true;
      else if (GetExtension(RemoveExtension(RemoveCompressedExtension(filename))) == "haplopop")
        has_haplo = true;
      else
        AssertMsg(false, filename);
      composite_files[i] = filename;
    }
    Assert(!has_haplo || !has_diplo);
  }

  if (params.Contains("switch_error"))
    switch_error = params["switch_error"].GetDouble();

  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;
  cerr << "- Input file = " << input_file << endl;
  cerr << "- SNPProfileData file = " << snpprofiledata_file << endl;
  if (!output_dir.empty())
    cerr << "- Output directory = " << output_dir << endl;
  cerr << "- Output file = " << output_file << endl;
  if (output_profile)
    cerr << "- Outputing profile & data" << endl;
  else
    cerr << "- Outputing data only" << endl; 
  if (!selected.empty())
  {
    cerr << "- Selected samples = {";
    for (int i = 0; i < selected.size(); ++i)
      cerr << " " << selected[i];
    cerr << " }" << endl;
  }
  if (!do_diplo)
    cerr << "- Haplotypes" << endl;
  else
    cerr << "- Diploid individuals" << endl;
  cerr << "- Label = \"" << label << "\"" << endl;

  if (!composite_files.empty())
  {
    cerr << "- Composite files {";
    for (int i = 0; i < composite_files.size(); ++i)
      cerr << " " << composite_files[i];
    cerr << " }" << endl;
  }

  cerr << "- Switch error rate = " << switch_error << endl;
}

void GetIntegerSpec(string s, vector<int> &result)
{
  int p = 0;
  while (p < s.length())
  {
    if (isdigit(s[p]))
    {
      int i;
      Assert(StringToInt(s, i, p, p));
      if ((p == s.length()) || (s[p] == ','))
      {
        result.push_back(i);
        if (p < s.length())
          ++p;
        continue;
      }

      Assert(s[p++] == '-');
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p == s.length()) || (s[p] == ','));
      if (i <= j)
      {
        for (int k = i; k <= j; ++k)
          result.push_back(k);
      }
      else
      {
        for (int k = i; k >= j; --k)
          result.push_back(k);
      }
      if (p < s.length());
      ++p;
    }
    else if (s[p] == '(')
    {
      int i;
      Assert(StringToInt(s, i, p, ++p));
      Assert((p < s.length()) && (s[p] == ','));
      ++p;
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p < s.length()) && (s[p] == ')'));
      Assert(j >= 0);
      for (int k = 0; k < j; ++k)
        result.push_back(i + k);
      ++p;
      Assert((p == s.length()) || (s[p] == ','));
      if (p < s.length())
        ++p;
    }
    else
      Assert(false);
  }
}

int Main(vector<string> args)
{
  cerr << "SimulatePopToSNPData" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;
  
  SNPProfile profile;
  SNPData data;
  {
    istream *in = NULL;
    if (!input_dir.empty())
      in = InFileStream(RelPath(input_dir, snpprofiledata_file));
    if (!in)
      in = InFileStream(snpprofiledata_file);

    if (!in)
    {
      cerr << "Could not open file \"" << snpprofiledata_file << "\" in directory";
      if (!input_dir.empty())
        cerr << " \"" << input_dir << "\"";
      Exit(1);
    }
    
    cerr << "* Reading sample data...";
    ReadBinSNPProfileData(*in, profile, data);
    cerr << " done." << endl;
    delete in;
    
    profile.Validate();
    data.Validate();
  }

  Simulate::HaploPop haplo_pop(data);
  Simulate::DiploPop diplo_pop(data);

  {
    istream *in = InFileStream(RelPath(input_dir, input_file));
    if (!in)
    {
      cerr << "Could not open file \"" << input_file << "\"";
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"" << endl;
      Exit(1);
    }

    cerr << "* Reading input data...";
    if (!do_diplo)
      haplo_pop.ReadBin(*in);
    else
      diplo_pop.ReadBin(*in);
    cerr << " done." << endl;
    delete in;
  }

  if (selected.empty())
  {
    if (!do_diplo)
    {
      for (int k = 0; k < haplo_pop.size(); ++k)
        selected.push_back(k);
    }
    else
    {
      for (int d = 0; d < diplo_pop.size(); ++d)
        selected.push_back(d);
    }
  }

  vector<Simulate::HaploPop> haplo_composite;
  vector<Simulate::DiploPop> diplo_composite;
  if (!composite_files.empty())
  {
    for (int i = 0; i < composite_files.size(); ++i)
    {
      cerr << "* Reading simulated population \"" << composite_files[i] << "\"...";
      if (GetExtension(RemoveExtension(RemoveCompressedExtension(composite_files[i]))) == "diplopop")
      {
        diplo_composite.push_back(Simulate::DiploPop(data));
        diplo_composite.back().ReadBin(RelPath(input_dir, composite_files[i]));
      }
      else
      {
        haplo_composite.push_back(Simulate::HaploPop(data));
        haplo_composite.back().ReadBin(RelPath(input_dir, composite_files[i]));
      }
      cerr << " done.";
      if (GetExtension(RemoveExtension(RemoveCompressedExtension(composite_files[i]))) == "diplopop")
        cerr << " # diploid individuals = " << diplo_composite.back().size() << endl;
      else
        cerr << " # haplotypes = " << haplo_composite.back().size() << endl;
    }
  }

  if (!do_diplo)
  {
    if (!diplo_composite.empty())
    {
      Assert(haplo_composite.empty());
      haplo_composite.resize(diplo_composite.size());
      for (int pop = 0; pop < diplo_composite.size(); ++pop)
      {
        for (int d = 0; d < diplo_composite[pop].size(); ++d)
        {
          haplo_composite[pop].Add(diplo_composite[pop][d].haplos[0]);
          haplo_composite[pop].Add(diplo_composite[pop][d].haplos[1]);
        }
      }
      diplo_composite.clear();
    }
  }
  else
  {
    Assert(haplo_composite.empty());
  }

  cerr << "* Creating SNP data...";
  SNPData new_data(profile);
  new_data.pops.push_back(SNPData::Pop(label));
  
  vector<int8> haplo0, haplo1;
  for (int i = 0; i < selected.size(); ++i)
  {
    int k = selected[i];
    Assert(k >= 0);
    if (!do_diplo)
    {
      Simulate::Haplo temp_haplo;
      Assert(k < haplo_pop.haplos.size());
      if (haplo_composite.empty())
        temp_haplo = haplo_pop.haplos[k];
      else
        temp_haplo.Composite(haplo_composite, haplo_pop.haplos[k]);
      temp_haplo.Generate(data, haplo0);
      new_data.AddHaplo(haplo0, 0, label + "_" + ToStr(k));
    }
    else
    {
      Simulate::Diplo temp_diplo;
      Assert(k < diplo_pop.diplos.size());
      if (diplo_composite.empty())
        temp_diplo = diplo_pop.diplos[k];
      else
        temp_diplo.Composite(diplo_composite, diplo_pop.diplos[k]);
      temp_diplo.Generate(data, haplo0, haplo1);
      new_data.AddDiplo(haplo0, haplo1, switch_error, 0, label + "_" + ToStr(k),
                        temp_diplo.haplo_sry[0], temp_diplo.haplo_sry[1]);
    }
  }
  new_data.ComputeDerived();
  cerr << " done." << endl;

  new_data.Validate();

  {
    ostream *out = OutFileStream(RelPath(output_dir, output_file));
    AssertMsg(out, output_file);

    if (output_profile)
      WriteBinSNPProfileData(*out, new_data);
    else
      new_data.WriteBin(*out);

    delete out;
  }

  cerr << "Done!" << endl;
  return 0;
}
