
#include "System.h"
#include "Utility.h"
#include "Params.h"
#include "SNPData.h"
#include "Simulate.h"
#include "HaploModel.h"


Params params(false);

int debug = 0;

string input_dir;
string output_dir;
string model_file;
bool default_model = false;
string sample_file;
string training_file;
string simulate_training_file;
bool extrapolate = true;
bool unsupervised = false;
vector<int> diplos;
vector<int> haplos;


string required_param[] =
{
  "model_file",
  "sample_file",
};
int num_required_param = sizeof(required_param) / sizeof(required_param[0]);

void GetIntegerSpec(string s, vector<int> &result);

void InitParams()
{
  params.ExpandImports();

  uint32 seed = RandomSeed();
  if (params.Contains("rand_seed"))
    seed = params["rand_seed"].GetInt();
  srand(seed);
  cerr << "- Random seed = " << seed << endl;

  params.Require(vector<string>(&required_param[0],
                                &required_param[num_required_param]));

  if (params.Contains("work_dir"))
    input_dir = output_dir = params["work_dir"].GetString();

  if (params.Contains("input_dir"))
    input_dir = params["input_dir"].GetString();

  if (params.Contains("output_dir"))
    output_dir = params["output_dir"].GetString();

  if (params.Contains("model_file"))
  {
    model_file = params["model_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(model_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(model_file))) == "haplomodel");
  }

  if (params.Contains("default_model"))
    default_model = true;

  if (params.Contains("sample_file"))
  {
    sample_file = params["sample_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(sample_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(sample_file))) == "snpprofiledata");
  }

  if (params.Contains("training_file"))
  {
    training_file = params["training_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(training_file)) == "bin");
    Assert(GetExtension(RemoveExtension(RemoveCompressedExtension(training_file))) == "snpprofiledata");
  }

  if (params.Contains("simulate_training_file"))
  {
    simulate_training_file = params["simulate_training_file"].GetString();
    Assert(GetExtension(RemoveCompressedExtension(simulate_training_file)) == "bin");
    Assert((GetExtension(RemoveExtension(RemoveCompressedExtension(simulate_training_file))) == "diplopop") || (GetExtension(RemoveExtension(RemoveCompressedExtension(simulate_training_file))) == "haplopop"));
  }

  Assert(!training_file.empty() || !simulate_training_file.empty());

  if (params.Contains("no_extrapolate"))
    extrapolate = false;

  if (params.Contains("unsupervised"))
    unsupervised = true;

  if (params.Contains("diplos"))
  {
    Assert(!params.Contains("haplos"));
    GetIntegerSpec(params["diplos"].GetString(), diplos);
  }

  if (params.Contains("haplos"))
  {
    Assert(!params.Contains("diplos"));
    GetIntegerSpec(params["haplos"].GetString(), haplos);
  }

  if (!input_dir.empty())
    cerr << "- Input directory = " << input_dir << endl;

  cerr << "- Sample file = " << sample_file << endl;
  cerr << "- Training file = " << training_file << endl;

  if (!simulate_training_file.empty())
    cerr << "- Simulated training file = " << simulate_training_file << endl;\

  if (!output_dir.empty())
    cerr << "- Output directory = " << output_dir << endl;

  cerr << "- Model file = " << model_file << endl;

  if (default_model)
    cerr << "- Default initial model" << endl;

  if (extrapolate)
    cerr << "- Extrapolate training" << endl;

  if (unsupervised)
    cerr << "- Unsupervised training" << endl;

  if (!diplos.empty())
    cerr << "- Selected diploid individuals { " << IntegerListString(diplos.begin(), diplos.end()) << " }" << endl;

  if (!haplos.empty())
    cerr << "- Selected haplotypes { " << IntegerListString(haplos.begin(), haplos.end()) << " }" << endl;
}


void GetIntegerSpec(string s, vector<int> &result)
{
  int p = 0;
  while (p < s.length())
  {
    if (isdigit(s[p]))
    {
      int i;
      Assert(StringToInt(s, i, p, p));
      if ((p == s.length()) || (s[p] == ','))
      {
        result.push_back(i);
        if (p < s.length())
          ++p;
        continue;
      }

      Assert(s[p++] == '-');
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p == s.length()) || (s[p] == ','));
      if (i <= j)
      {
        for (int k = i; k <= j; ++k)
          result.push_back(k);
      }
      else
      {
        for (int k = i; k >= j; --k)
          result.push_back(k);
      }
      if (p < s.length());
      ++p;
    }
    else if (s[p] == '(')
    {
      int i;
      Assert(StringToInt(s, i, p, ++p));
      Assert((p < s.length()) && (s[p] == ','));
      ++p;
      int j;
      Assert(StringToInt(s, j, p, p));
      Assert((p < s.length()) && (s[p] == ')'));
      Assert(j >= 0);
      for (int k = 0; k < j; ++k)
        result.push_back(i + k);
      ++p;
      Assert((p == s.length()) || (s[p] == ','));
      if (p < s.length())
        ++p;
    }
    else
      Assert(false);
  }
}


void SaveModel(const HaploModel::Model &model)
{
  cerr << "* Writing model to file " << RelPath(output_dir, model_file) << "...";
  ostream *out = OutFileStream(RelPath(output_dir, model_file));
  AssertMsg(out, RelPath(output_dir, model_file));
  model.Write(*out);
  delete out;
  cerr << " done." << endl;
}

void CreateErrors(vector<int8> &haplos, double error_rate, double missing_rate)
{
  for (int i = 0; i < haplos.size(); ++i)
  {
    if (haplos[i] == 4)
      continue;
    if (drand() < missing_rate)
      haplos[i] = 4;
    else
    {
      if (drand() < error_rate)
        haplos[i] = (haplos[i] + 1 + (irand() % 3)) & 3;
    }
  }
}

int Main(vector<string> args)
{
  cerr << "TrainHaploModel" << endl;
  InitBaseTables();
  params.ImportArgs(vector<string>(args.begin() + 1, args.end()));
  InitParams();
  cerr << endl;

  SNPProfile profile;
  SNPData data;
  {
    string filename;
    istream *in;
    if (!input_dir.empty())
      in = InFileStream(RelPath(input_dir, sample_file));
    else
      in = InFileStream(sample_file);

    if (!in)
    {
      cerr << "Could not open file \"" << sample_file << "\"";
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"";
      cerr << endl;
      Exit(1);
    }

    cerr << "* Reading sample data...";
    ReadBinSNPProfileData(*in, profile, data);
    cerr << " done." << endl;
    delete in;

    profile.Validate();
    data.Validate();
  }

  SNPProfile training_profile;
  SNPData training_data;
  if (!training_file.empty())
  {
    string filename;
    istream *in;
    if (!input_dir.empty())
      in = InFileStream(RelPath(input_dir, training_file));
    else
      in = InFileStream(training_file);

    if (!in)
    {
      cerr << "Could not open file \"" << training_file << "\"";
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"";
      cerr << endl;
      Exit(1);
    }

    cerr << "* Reading training data...";
    ReadBinSNPProfileData(*in, training_profile, training_data);
    cerr << " done." << endl;
    delete in;

    training_profile.Validate();
    training_data.Validate();
  }

  Simulate::HaploPop training;
  if (!training_file.empty())
    training.SetSNPData(training_data);
  else
    training.SetSNPData(data);
  if (!simulate_training_file.empty())
  {
    istream *in = InFileStream(RelPath(input_dir, simulate_training_file));
    if (!in)
    {
      cerr << "Could not open file \"" << simulate_training_file << "\"";
      if (!input_dir.empty())
        cerr << " in directory \"" << input_dir << "\"";
      cerr << endl;
      Exit(1);
    }

    cerr << "* Reading simulated population \"" << simulate_training_file << "\"...";
    if (GetExtension(RemoveExtension(RemoveCompressedExtension(simulate_training_file))) == "diplopop")
    {
      Simulate::DiploPop temp_training;
      if (!training_file.empty())
        temp_training.SetSNPData(training_data);
      else
        temp_training.SetSNPData(data);
      temp_training.ReadBin(*in);

      training.haplos.resize(temp_training.diplos.size() * 2);
      for (int d = 0; d < temp_training.diplos.size(); ++d)
        for (int a = 0; a < 2; ++a)
          training.haplos[d*2 + a].swap(temp_training.diplos[d].haplos[a]);

      if (!diplos.empty())
        for (int i = 0; i < diplos.size(); ++i)
        {
          haplos.push_back(diplos[i]*2 + 0);
          haplos.push_back(diplos[i]*2 + 1);
        }
    }
    else
      training.ReadBin(*in);
      
    delete in;
    cerr << " done." << endl;
  }
  else
    training.SetAll();

  HaploModel::Model model(data);
  model.SetDefault();
  if (!default_model && !model_file.empty())
  {
    istream *in = InFileStream(RelPath(output_dir, model_file));
    if (in)
    {
      cerr << "* Reading initial model from " << RelPath(output_dir, model_file) << "...";
      model.Read(*in);
      delete in;
      cerr << " done." << endl;
    }
  }
  
  cerr << "* Initial model:" << endl;
  model.DisplayEmission(cerr);
  model.DisplayPopInit(cerr);
  model.DisplayPopRecomb(cerr);
  model.DisplayPopAdmix(cerr);

  if (haplos.empty())
  {
    haplos.resize(training.haplos.size());
    for (int i = 0; i < haplos.size(); ++i)
      haplos[i] = i;
  }
  
  HaploModel::BaumWelch bw(data, model);
  int max_iterations = 100;
  cerr << "* # training examples = " << haplos.size() << endl;
  int64 start_time = TimerSeconds();
  int rand_seed = irand();
  for (int iteration = 0; iteration < max_iterations; ++iteration)
  {
    cerr << "* Training iteration " << iteration << ":";
    srand(rand_seed);
    vector<int8> haplo, pops;
    vector<uint64> pop_mask;
    vector<int> haplo_exclude;
    for (int training_index = 0; training_index < haplos.size(); ++training_index)
    {
      int k = haplos[training_index];
      Assert((k >= 0) && (k < training.haplos.size()));

      if (!training_file.empty())
      {
        training.haplos[k].Generate(training_data, haplo);
        pop_mask.clear();
        haplo_exclude.clear();
      }
      else
      {
#if 0
        cerr << endl;
        for (int i = 0; i < training.haplos[k].sources.size(); ++i)
          cerr << "[" << training.haplos[k].sources[i].first.left() << ", " << training.haplos[k].sources[i].first.right() << ") = " << training.haplos[k].sources[i].second << endl;
#endif
        training.haplos[k].Generate(data, haplo);
        if (!unsupervised)
        {
          training.haplos[k].LabelPop(data, pops);
          HaploModel::GetPopMask(data, pops, pop_mask);
        }
        training.haplos[k].Label(haplo_exclude);
        //cerr << pop_mask[0] << endl;
        //cerr << haplo_exclude[0] << endl;
      }

      CreateErrors(haplo, 0.01, 0.05);
      bw.AccumulateTraining(haplo, pop_mask, vector<int>(), haplo_exclude);
      cerr << ".";
      //cerr << " time=" << (TimerSeconds() - start_time);
      start_time = TimerSeconds();
    }

    cerr << " Updating...";
    double partition_rel_diff;
    bool changed = bw.UpdateModel(partition_rel_diff);

    if (extrapolate && (iteration & 1))
    {
      cerr << " Extrapolating...";
      for (int pop = 0; pop < data.num_pops; ++pop)
        bw.model.pop_recomb[pop] = model.pop_recomb[pop] + 2.0*(bw.model.pop_recomb[pop] - model.pop_recomb[pop]);
    }
    model = bw.model;

    cerr << " done. New model:" << endl;
    model.DisplayEmission(cerr);
    model.DisplayPopInit(cerr);
    model.DisplayPopRecomb(cerr);
    model.DisplayPopAdmix(cerr);
    SaveModel(model);

    if (!changed)
      break;
  }

  cerr << "* Final model:" << endl;
  model.DisplayEmission(cerr);
  model.DisplayPopInit(cerr);
  model.DisplayPopRecomb(cerr);
  model.DisplayPopAdmix(cerr);
  SaveModel(model);

  cerr << "Done!" << endl;
  return 0;
}
